(* Case Tactic from Software Fondations http://www.cis.upenn.edu/~bcpierce/sf/ *)
(* --------------------------------------------- Fletch common definitions --------------------------------------------- *)

Require Import Arith.
Require Import NPeano.
Require Export List.
Require String. Open Scope string_scope.

Module Commons.

Ltac move_to_top x :=
  match reverse goal with
  | H : _ |- _ => try move x after H
  end.

(* TACTICS *)

(* Apply inversion and simplify the subgoals replacing all the substitutions coming from inversion. *)
Ltac sinversion H := inversion H;subst.
(* Destruct a subgoal V into l and H *)
Ltac vdestruct V l H := destruct V as [ l H ].
(* Apply destruct to an environment well formedness proposition *)
Ltac wfdestruct WFEnvs :=
  destruct WFEnvs as [ WFRvenv [ WFLenv WFEnv ] ].

Tactic Notation "assert_eq" ident(x) constr(v) :=
  let H := fresh in
  assert (x = v) as H by reflexivity;
  clear H.

Tactic Notation "Case_aux" ident(x) constr(name) :=
  first [
    set (x := name); move_to_top x
  | assert_eq x name; move_to_top x
  | fail 1 "because we are working on a different case" ].

(* Destruct used for the excluded middle. 
 * Use emdestruct formula Y Z to destruct a proposition "formula" in the form 
 * A \/ ~A into A (named as the argument Y) and B (named as the argment Z).
 *)
Tactic Notation "emdestruct" ident(Y) ident(Z) constr(L) :=
  let X := fresh in
  assert(X := L); destruct X as [ Y | Z ].

(* Add the result of IH H to the assumptions with a fresh variable. *)
Ltac hfree IH H :=
  let X := fresh in
    assert(X := IH H).

(* 
 * The following tactics are used to provide a textual description of the proof cases in an induction or case analysis proof. 
 * For example,
 * - Case "exp = val"
 * describes that the current case is the following: the expression is a value.
 *
 * Note that "exp = val" is just a label used to describe the current induction case.
 *)
Tactic Notation "Case" constr(name) := Case_aux Case name.
Tactic Notation "SCase" constr(name) := Case_aux SCase name.
Tactic Notation "SSCase" constr(name) := Case_aux SSCase name.
Tactic Notation "SSSCase" constr(name) := Case_aux SSSCase name.
Tactic Notation "SSSSCase" constr(name) := Case_aux SSSSCase name.
Tactic Notation "SSSSSCase" constr(name) := Case_aux SSSSSCase name.
Tactic Notation "SSSSSSCase" constr(name) := Case_aux SSSSSSCase name.
Tactic Notation "SSSSSSSCase" constr(name) := Case_aux SSSSSSSCase name.

(* Syntactic notation for lists and sets *)
Notation "[ ]" := nil.
Notation "[ x ; .. ; y ]" := (cons x .. (cons y nil) ..).
Notation "v '_in' S" := (In v S) (at level 1).
Notation "v 'not' '_in' S" := (~ (In v S)) (at level 1).

(* Coq uses an intiutionistic logic, but we need to assume the excluded middle. *)
Axiom excluded_middle : forall P, P \/ ~P.

(* ENVIRONMENT GENERIC DEFINITIONS *)

(* Generic definition of an environment as a mappint from a elements of type KEY to elements of type VALUE. *)
Definition fletch_env {KEY : Set}{VALUE : Set} : Set := (list (KEY * VALUE)).

(* This predicate asserts that there an encironment does not contain any duplicate keys. *)
Inductive NoDupKeys {KEY : Set}{VALUE : Set} : (@fletch_env KEY VALUE) -> Prop :=
  | NoDupKeys_nil : NoDupKeys []
  | NoDupKeys_cons : forall k v env,
                       (forall v', (k,v') not _in env) ->
                                         NoDupKeys env ->
                                         NoDupKeys ((k,v)::env).
(* 
 * Just Coq machinery to show that if an environment does not contain any duplicate keys, 
 * then there exists at most one (k, v) in the environment for each k of type KEY.
 *)
Lemma NoDupKeysLemma {KEY : Set}{VALUE : Set} : forall (k : KEY) (v1 : VALUE) v2 env,
                                     NoDupKeys env ->
                                     (k,v1) _in env ->
                                     (k,v2) _in env ->
                                     v1 = v2.
Proof.
  intros k v1 v2.
  induction env.
  inversion 2.
  intros. 
  destruct H0 ; destruct H1.
  rewrite -> H1 in H0 ; inversion H0 ; auto.
  inversion H. rewrite H0 in H2. inversion H2 ; subst.
  assert (C := H4 _ H1) ; contradiction.
  inversion H. rewrite H1 in H2. inversion H2 ; subst.
  assert (C := H4 _ H0) ; contradiction.
  inversion H ; subst.
  exact (IHenv H5 H0 H1).
Qed.

(* 
 * An environment is well formed if it does not contain any duplicate keys (clearly it applies to any
 * kind of environment in our formalism).
 *)
Definition well_formed_env {KEY : Set}
                           {VALUE : Set}
                           (env : @fletch_env KEY VALUE) : Prop :=
  NoDupKeys env. 

(* 
 * The following definitions are generic definitions of 
 * insert env k v env': insert a pair (k, v) in the environment env resulting in the environment env'. The key
 *                      k is required to be a fresh name.
 * update env k v env': update a pair (k, v) in the environment.
 *)

Inductive insert {KEY : Set}{VALUE : Set} : (@fletch_env KEY VALUE) ->
                                          KEY -> VALUE ->
                                         (@fletch_env KEY VALUE) -> Prop :=
  | insert_case : forall env k v,
                    (forall v', (k,v') not _in env) ->
                    insert env k v ((k,v)::env).

Inductive update {KEY : Set}{VALUE : Set} : @fletch_env KEY VALUE -> KEY -> VALUE
                                         -> @fletch_env KEY VALUE -> Prop :=
  | update_base : forall k v,
                    update [] k v [(k,v)]
  | update_ind_eq : forall k v1 v2 env,
                      update ((k,v1)::env) k v2 ((k,v2)::env)
  | update_ind_neq : forall k1 k2 v1 v2 env env',
                       k1 <> k2 ->
                       update env k1 v1 env' ->
                       update ((k2,v2)::env) k1 v1 ((k2,v2)::env').

Hint Constructors NoDupKeys insert update.

(* LEMMAS *)
(* The following lemmas states trivial properties about the environments *)

Lemma fol_logic_impl : forall P Q : Prop, (P -> Q) -> (~ Q -> ~ P).
Proof.
  auto.
Qed.

Lemma update_not_add {KEY:Set}{VALUE:Set} :
  forall k v env env' k' v',
    (k,v) not _in env ->
    (@update KEY VALUE) env k' v' env' ->
    k' <> k ->
    (k,v) not _in env'.
Proof.
  intros k v env env' k' v' H1 H2 H3.
  induction H2.
  Case "update [] (k,v) [(k,v)]".
    intro contr. simpl in contr. destruct contr. inversion H. contradiction. auto.
  Case "update (k,v)::env k0 v1 (k,v)::env'".
    intro contr. simpl in contr. destruct contr. inversion H. contradiction.
    destruct H1. assert(X := in_cons (k0,v1) (k,v) env H). exact X.
  Case "update (k,v)::env k v' (k,v')".
    assert(X := fol_logic_impl _ _ (in_cons _ _ _) H1).
    assert(IHupdate' := IHupdate X H3). intro contr. destruct contr.
    cut ((k2, v2) _in ((k2, v2) :: env)).
    rewrite <- H0 in H1. auto. econstructor. reflexivity.
    exact(IHupdate X H3 H0).
Qed.
    
Lemma insert_well_formed {KEY:Set}{VALUE:Set} :
  forall (k:KEY) (v:VALUE) env env',
         well_formed_env env ->
         insert env k v env' ->
         well_formed_env env'.
Proof.
    intros k v env env' WFenv Insert.
    induction Insert.
    Case "insert case".
      econstructor.
        exact H.
        inversion WFenv. trivial. econstructor. exact H0. exact H1.
Qed.
        
Lemma update_well_formed {KEY:Set}{VALUE:Set} :
  forall (k:KEY) (v:VALUE) env env',
         well_formed_env env ->
         update env k v env' ->
         well_formed_env env'.
Proof.
  intros k v env env' WFenv Update.
  unfold well_formed_env.
  unfold well_formed_env in WFenv.

  induction Update.
  Case "env=[], env'=[(k,v)]".
    econstructor. auto. auto.
  Case "env=(k,v1)::xs,env'=(k,v2)::xs".
    inversion WFenv. econstructor. exact H1. exact H3.
  Case "env=(k2,v2)::xs,env'=(k2,v2)::xs', update to (k1,v1)".
    inversion WFenv. assert(NoDupEnv' := IHUpdate H4). econstructor.
    intros. assert(H2' := H2 v').
    assert(UNA := (@update_not_add KEY VALUE) k2 v' env env' k1 v1 H2' Update H).
    exact UNA. auto.
Qed.

Lemma insert_correctness {KEY : Set}{VALUE : Set} :
  forall (k : KEY) (v : VALUE) env env',
    insert env k v env' ->
    (k,v) _in env'.
Proof.
  intros k v env env' Insert.
  inversion Insert.
  econstructor. trivial.
Qed.

Lemma insert_add {KEY : Set}{VALUE : Set} :
  forall (k : KEY) k' (v : VALUE) v' env env',
    (k,v) _in env ->
    (@insert KEY VALUE) env k' v' env' ->
    (k,v) _in env'.
Proof.
  intros k k' v v' env env' KVIn Insert.
  sinversion Insert.
  assert(X := in_cons (k', v') (k,v) env KVIn). auto.
Qed.

Lemma update_correctness {KEY : Set}{VALUE : Set} :
  forall (k : KEY) (v : VALUE) env env',
    update env k v env' ->
    (k,v) _in env'.
Proof.
  intros k v env env' Update.
  induction Update.
    econstructor. trivial.
    econstructor. trivial.
    exact (in_cons (k2,v2) (k1,v1) env' IHUpdate).
Qed.    

Lemma update_not_noise {KEY : Set}{VALUE : Set} : forall (k1 : KEY) k2 v1 (v2 : VALUE) env env',
                           well_formed_env env ->
                           k1 <> k2 ->
                           (k2, v1) _in env ->
                           update env k1 v2 env' ->
                           (k2, v1) _in env'.
Proof.
  intros. unfold well_formed_env in H.
  induction H2.
    inversion H1.
    sinversion H1. inversion H2. contradiction. right. auto.
    sinversion H1. sinversion H4. left. auto. 
    sinversion H.
    assert(X := IHupdate H9 H0 H4). right. auto.
Qed.

Lemma pair_em {KEY : Set} {VALUE : Set} : forall x (k : KEY) (v : VALUE),
                                            (k,v) = x \/ ~ (k,v) = x.
Proof.
  intros. assert(EM := excluded_middle ((k,v) = x)). auto.
Qed.

Lemma key_em {KEY : Set} : forall (k1 : KEY) k2,
                             k1 = k2 \/ k1 <> k2.
Proof.
  intros. assert(EM := excluded_middle (k1 = k2)). auto.
Qed.


Lemma env_excluded_middle {KEY : Set}{VALUE : Set} : forall (env : @fletch_env KEY VALUE) k v,
                                                       (k,v) _in env \/
                                                       (k,v) not _in env.
Proof.
  intros. induction env.
  eauto. destruct IHenv. left. assert(X := in_cons a (k,v) env H). trivial.
  assert(X := pair_em a k v). 
  destruct X. left. rewrite <- H0. assert(X := in_eq (k,v) env). trivial.
  right. intros F. inversion F. contradiction H0. rewrite <- H1. trivial.
  contradiction H1.
Qed.

Lemma insertable {KEY : Set}{VALUE : Set} :
  forall (k : KEY) (v : VALUE) env,
    (forall v, (k, v) not _in env) ->
    exists env', insert env k v env'.
Proof.
  intros k v env NotIn.
  induction env.
  - exists [(k, v)]. constructor. eauto.
  - assert (Clean : (forall v : VALUE, (k, v) not _in env)). intuition.
    assert(Contradiction : (k, v0) _in (a :: env)). constructor 2. trivial.
    assert(Contradiction' := NotIn v0 Contradiction). contradiction.
    assert(IH := IHenv Clean). destruct IH as [ env' Insert ]. exists ((k, v) :: a :: env).
    constructor. auto.
Qed.    

Lemma updatable {KEY : Set}{VALUE : Set} :
  forall (k : KEY) (v1 v2 : VALUE) env,
    (k, v1) _in env ->
    exists env', update env k v2 env'.
Proof.
  intros k v1 v2 env env'.
  induction env. 
  - sinversion env'.
  - destruct a. emdestruct In NotIn (key_em k k0).
    * sinversion env'.  sinversion H. subst. exists ((k0, v2) :: env). constructor.
      assert(IH := IHenv H). destruct IH. exists ((k0, v2) :: env). constructor.
    * sinversion env'. intuition. sinversion H. assert(E : k = k). trivial. 
      assert(Contradiction := NotIn E). contradiction.
      assert(IH := IHenv H). destruct IH as [ env'' Update ].
      exists ((k0, v) :: env''). constructor. auto. auto.
Qed.


Inductive fold_pred {KEY : Set}{VALUE : Set}{KEY' : Set}{VALUE' : Set} :
  ((@fletch_env KEY VALUE) ->
   KEY' ->
   VALUE' ->
   (@fletch_env KEY VALUE) -> Prop) ->
  (@fletch_env KEY VALUE) ->
  list (KEY'*VALUE') ->
  (@fletch_env KEY VALUE) -> Prop :=
| fold_pred_base : forall P (env:@fletch_env KEY VALUE),
                     fold_pred P env [] env
| fold_pred_in : forall (P:((@fletch_env KEY VALUE) ->
                            KEY' -> VALUE' ->
                            (@fletch_env KEY VALUE) -> Prop)) env
                        env' env''
                        k v el,
                   P env k v env' ->
                   fold_pred P env' el env'' ->
                   fold_pred P env ((k,v)::el) env''.

Lemma fold_preservation {KEY : Set}{VALUE : Set} :
  forall (P:(@fletch_env KEY VALUE) -> KEY -> VALUE -> ((@fletch_env KEY VALUE)) -> Prop)
         (Q : _ -> Prop) env el env',
    Q env ->
    (forall k v env env', P env k v env' -> Q env') ->
    fold_pred P env el env' ->
    Q env'.
Proof.
  intros P Q env el env' lemma_env lemma_env' fold_holds.
  induction fold_holds.
  Case "[]".
    exact lemma_env.
  Case "x::xs".
    assert(X := lemma_env' k v env env' H).
    assert(Y := IHfold_holds X lemma_env').
    exact Y.
Qed.

(* 
 * Abstract representation of idendifiers.
 * The id_kind type provides a way to define a separate namespace
 * for each identifier (variable identifier, type parameter identifier,
 * method identifier, etc.)
 *)
Inductive id_kind : Set :=
  | var_id : id_kind
  | type_var_id : id_kind
  | location_id : id_kind
  | var_location_id : id_kind
  | method_id : id_kind
  | field_id : id_kind
  | class_id : id_kind.

(* Generates an identifier. *)
Record generate_id {T : id_kind} :=
  | id : nat -> @generate_id T.

(* It represents the "Object" identifier, used to refer to the Object class. *)
Definition builtin_class_object := @id class_id 0.

End Commons.
