(* ------------------------------------------  Fletch class table definitions ----------------------------------------- *)

Require Import fletch_syntax.
Require Import fletch_commons.
Require Import fletch_types.

Module Type ClassTable.
Import FletchSyntax.
Import Commons. 
Import FletchTypes.

Definition ltype_nv (GL : list fletch_nvtype) : list fletch_type :=
  map (fun G => type_nv G) GL.

(* 
 * Fletch class table definition.
 * A class table is a list of pairs (class identifier, class definition).
 *)
Definition fletch_CT := list ((@generate_id class_id) * fletch_class).
Parameter CT : fletch_CT.

(* This predicate states that "there is no cycle in the class table definition" *)
Inductive nocycles : list (@generate_id class_id) -> (@generate_id class_id) -> Prop :=
  | cycles_base : forall E, nocycles E builtin_class_object
  | cycles_inductive : forall E C GL D HL FL ML,
                         (C, class C GL (atype_class D HL) FL ML) _in CT ->
                         nocycles (C :: E) D ->
                         C not _in E ->
                         nocycles E C.

(* Definition of "well formed class table" *)
Parameter well_formed_CT :
  NoDupKeys CT /\
  (forall CD, (builtin_class_object, CD) not _in CT) /\
  (forall C CD, (C,CD) _in CT -> nocycles [] C).
Inductive fully_annotated_type : fletch_type -> Prop :=
  | fa_type_var : forall X,
                     fully_annotated_type (atype_nv (atype_var X))
  | fa_type_classtype : forall C GL,
                           fully_annotated_type_list (ltype_nv GL) ->
                           fully_annotated_type
                             (type_nv
                                (type_classtype (type_class C GL)))
  | fa_type_closure : forall GL T,
                        fully_annotated_type_list (ltype_nv GL) ->
                        fully_annotated_type T ->
                        fully_annotated_type (type_nv (GL --> T))
  | fa_type_void : fully_annotated_type atype_void
  | fa_type_bottom : fully_annotated_type (type_nv type_bottom)
with
 fully_annotated_type_list : list fletch_type -> Prop :=
  | fat_base : fully_annotated_type_list []
  | fat_ind : forall x xs,
                fully_annotated_type x ->
                fully_annotated_type_list xs ->
                fully_annotated_type_list (x :: xs).

(* Mutual induction principle for the fully annotated type predicate. *)
Scheme fully_annotated_type_mut := Induction for fully_annotated_type Sort Prop
with fully_annotated_type_list_mut := Induction for fully_annotated_type_list Sort Prop.

(* 
 * Definition of fully annotated expression predicate.
 * An expression "e" is fully annotated IFF "e" does not contain any occurrences of
 * the type "dynamic" in any of declared types occurring in the expression.
 *)
Inductive fully_annotated_expression : fletch_exp -> Prop :=
  | fa_exp_var : forall y,
                   fully_annotated_expression (exp_var y)
  | fa_exp_getter : forall e p,
                      fully_annotated_expression e ->
                      fully_annotated_expression (exp_getter e p)
  | fa_exp_setter : forall e1 p e2,
                      fully_annotated_expression e1 ->
                      fully_annotated_expression e2 ->
                      fully_annotated_expression (exp_setter e1 p e2)
  | fa_exp_assignment : forall x e,
                          fully_annotated_expression e ->
                          fully_annotated_expression (exp_assignment x e)
  | fa_call : forall e eL,
                fully_annotated_expression e ->
                fully_annotated_expression_list eL ->
                fully_annotated_expression (exp_call e eL)
  | fa_new : forall C GL,
               fully_annotated_type (atype_nv (atype_classtype (atype_class C GL))) ->
               fully_annotated_expression (exp_new (atype_class C GL))
  | fa_function : forall e GxL T,
                    (forall G x,
                      (G,x) _in GxL -> fully_annotated_type (atype_nv G)) ->
                    fully_annotated_type T ->
                    fully_annotated_expression e ->
                    fully_annotated_expression (exp_function GxL T e)

  | fa_location : forall l,
                    fully_annotated_expression (exp_location l)
  | fa_var_location : forall tau,
                      fully_annotated_expression (exp_var_location tau)
  | fa_runtime_assignment : forall tau e,
                              fully_annotated_expression e ->
                              fully_annotated_expression (exp_runtime_assignment tau e)
  | fa_frame_expression : forall T e,
                         fully_annotated_type T ->
                         fully_annotated_expression e ->
                         fully_annotated_expression (exp_frame_expression T e)
with
 fully_annotated_expression_list : list fletch_exp -> Prop :=
  | fal_base : fully_annotated_expression_list []
  | fal_ind : forall x xs,
                fully_annotated_expression x ->
                fully_annotated_expression_list xs ->
                fully_annotated_expression_list (x::xs).

(* Mutual induction principle for the fully annotated expression predicate. *)
Scheme fully_annotated_expression_mut := Induction for fully_annotated_expression Sort Prop
with fully_annotated_expression_list_mut :=
  Induction for fully_annotated_expression_list Sort Prop.

(* 
 * Definition of fully annotated field predicate.
 * An field is fully annotated IFF its declared type is fully annotated.
 *)
Inductive fully_annotated_field : fletch_field -> Prop :=
  | faf_base : forall G f,
                 fully_annotated_type (atype_nv G) ->
                 fully_annotated_field (field G f).
(* 
 * Definition of fully annotated method predicate.
 * An method is fully annotated IFF its declared (paramter and return) types are fully annotated.
 *)
Inductive fully_annotated_method : fletch_method -> Prop :=
  | fam_base :
    forall (T : fletch_atype) m GxL e,
      fully_annotated_type T ->
      (forall G x,
         (G,x) _in GxL -> fully_annotated_type (atype_nv G)) ->
      fully_annotated_expression e ->
      fully_annotated_method (method T m GxL e).

(* Definition of fully annotated list of fields. *)
Definition fully_annotated_fields (FL : list fletch_field) :=
  forall F, F _in FL ->
            fully_annotated_field F.
(* Definition of fully annotated list of methods. *)
Definition fully_annotated_methods (ML : list fletch_method) :=
  forall M, M _in ML ->
            fully_annotated_method M.

(* 
 * Definition of fully annotated class table.
 * A class table is fully annotated IFF all the type annotations do not contain
 * the type dynamic.
 *)
Parameter fully_annotated_CT :
  forall C PL N FL ML,
    (C, class C PL N FL ML) _in CT ->
    fully_annotated_type (type_nv (type_classtype N)) /\
    (forall X N,
       (X,N) _in PL -> fully_annotated_type (type_nv (type_classtype N))) /\
    fully_annotated_fields FL /\
    fully_annotated_methods ML.

End ClassTable.
