(* ------------------------------------------  Fletch environments ----------------------------------------- *)

Require Import fletch_commons.
Require Import fletch_types.
Require Import fletch_syntax.

Module FletchEnv.
Export FletchSyntax.

(* ENVIRONMENT DEFINITIONS *)
(* 1. Typing environments *)

(* Variable environment mapping variables to types *)
Definition fletch_venv := @fletch_env fletch_var fletch_nvtype.
(* Type environment mapping type variables to their upper bound *)
Definition fletch_tenv := @fletch_env (@generate_id type_var_id) fletch_classtype.

(* Environment update definitions for variable and type environment *)
Definition update_venv := @update fletch_var fletch_nvtype.
Definition insert_venv := @insert fletch_var fletch_nvtype.

Definition update_tenv := @update (@generate_id type_var_id) fletch_nvtype. 
Definition insert_tenv := @insert (@generate_id type_var_id) fletch_classtype.

(* Initial variable and type environment *)
Definition Gamma0 : fletch_venv := [(var_null, type_bottom)].
Definition Delta0 : fletch_tenv := [].

(* 2. Runtime environments *)

(* Fresh-generating environments  *)

Definition fletch_fresh_env {KEY : Set}{VALUE : Set} : Set :=
  ((@fletch_env KEY VALUE) * KEY)%type.

Inductive well_formed_fresh_env {KEY : Set}{VALUE : Set} : @fletch_fresh_env KEY VALUE ->
                                                           Prop :=
| well_formed_lenv_case : forall wenv fresh,
                            well_formed_env wenv ->
                            well_formed_fresh_env (wenv,fresh).

(* Environmnt insert and update definitions *)

Inductive insert_fresh {KEY : Set}{VALUE : Set}{NFRESH : KEY -> KEY} :
  (@fletch_fresh_env KEY VALUE) ->
  VALUE ->
  (KEY * (@fletch_fresh_env KEY VALUE))%type ->
  Prop :=

  | insert_fresh_case : forall env env' k v,
                    insert env (NFRESH k) v env' ->
                    insert_fresh (env,k) v (NFRESH k,(env',NFRESH k)).

Inductive update_fresh {KEY : Set}{VALUE : Set} : @fletch_fresh_env KEY VALUE -> KEY -> VALUE
                                         -> @fletch_fresh_env KEY VALUE -> Prop :=
  | update_fresh_case : forall env k env' v fresh,
                          update env k v env' ->
                          update_fresh (env,fresh) k v (env',fresh).

Hint Constructors insert_fresh update_fresh.

Definition BLIn {KEY : Set}{VALUE : Set}
           (entry : (KEY * VALUE)%type)
           (env : @fletch_fresh_env KEY VALUE) :=
  match env with
    | (env',_) => entry _in env'
  end.

Notation "v '__in' S" := (BLIn v S) (at level 41).
Notation "v 'not' '__in' S" := (~ (BLIn v S)) (at level 41).

(* ENVIRONMENT DEFINITIONS *)

(* The null location *)
Definition loc_null := (@id location_id 0).
(* 
 * The first fresh location. This is used as a starting value for the counter 
 * that generates fresh locations
 *)
Definition first_fresh := loc_null.
(* The first fresh variable location. *)
Definition first_vfresh := (@id var_location_id 0).

(* 
 * Variable location to heap location mapping. Since this environment is necessary
 * in the syntax definitions, it has to be defined here, i.e., before the fletch syntax
 * in the file "fletch_syntax.v", which imports this file.
 *)
Definition fletch_rvenv := @fletch_fresh_env (@generate_id var_location_id)
                                             (fletch_anvtype * @generate_id location_id)%type.

(* Environment insert and update definitions *)

Definition generate_fresh_key {KEY : id_kind} (fresh : @generate_id KEY) :
  @generate_id KEY :=
  match fresh with
      | id k => id (S k)
  end.

Definition generate_fresh_variable_location :
  @generate_id var_location_id -> @generate_id var_location_id := @generate_fresh_key var_location_id.

Definition generate_fresh_location :
  @generate_id location_id -> @generate_id location_id := @generate_fresh_key location_id.

Definition insert_rvenv := @insert_fresh (@generate_id var_location_id)
                                         (fletch_anvtype * @generate_id location_id)%type
                                         generate_fresh_variable_location.
Definition update_rvenv := @update_fresh (@generate_id var_location_id)
                                         (fletch_anvtype * @generate_id location_id)%type.

Notation runtime_field := ((@generate_id field_id) *
                           (fletch_nvtype *
                           (@generate_id location_id)))%type.
Notation runtime_method := ((@generate_id method_id) * (@generate_id location_id))%type.

Definition runtime_fields := (@fletch_env (@generate_id field_id)
                                          (fletch_nvtype * @generate_id location_id)%type).
Definition runtime_methods := (@fletch_env (@generate_id method_id)
                                           (@generate_id location_id)).

Inductive fletch_runtime_object : Set :=
  | class_instance_object : fletch_aclasstype -> runtime_fields -> runtime_methods -> fletch_runtime_object
  | closure_object : list param -> fletch_type -> fletch_exp -> fletch_runtime_object
  | empty_location : fletch_runtime_object.

Hint Constructors fletch_runtime_object.

Definition fletch_lenv := @fletch_fresh_env (@generate_id location_id) fletch_runtime_object.

Definition insert_lenv := @insert_fresh (@generate_id location_id)
                                        fletch_runtime_object
                                        generate_fresh_location.
Definition update_lenv := @update_fresh (@generate_id location_id)
                                        fletch_runtime_object.

(* Base runtime environments and runtime environments auxiliary definitions *)

Definition nu0 : fletch_rvenv := ([], first_vfresh).
Definition sigma0 : fletch_lenv := ([(loc_null, empty_location)], first_fresh).

Definition well_formed_rvenv (nu : fletch_rvenv) := well_formed_fresh_env nu.              
Definition well_formed_lenv (sigma : fletch_lenv) := well_formed_fresh_env sigma.
Definition lenv_base_superset (sigma : fletch_lenv) :=
  (forall l o, (l,o) __in sigma0 -> (l,o) __in sigma).
Fixpoint runtime_types (GxL : params) : list fletch_nvtype :=
  match GxL with
    | [] => []
    | (G,_) :: GL => (convert_nvtype G) :: (runtime_types GL)
  end.
Definition typeof o : fletch_type :=
  match o with
      | class_instance_object N _ _ => type_nv (type_classtype N)
      | closure_object GxL T _ => type_nv ((runtime_types GxL) --> T)
      | empty_object => type_nv type_bottom
  end.

Fixpoint update_runtime_field (FL : runtime_fields) (entry : runtime_field) : runtime_fields :=
  match (FL, entry) with
    | ([],_) => []
    | ((f,(_,_)) as entry' :: xs, (f',(T,l))) => if field_id_compare f f' then
                                            (f,(T,l)) :: xs
                                          else
                                            entry' :: (update_runtime_field
                                                         xs entry)
  end.

End FletchEnv.
