Require Export fletch_sublemmas.

Module Lemmas (C : ClassTable).
Module SubLemmasImpl := SubLemmas(C).
Export SubLemmasImpl.

(*
 * The following file is divided into 3 sections (every lemma contains a textual description and an
 * argumentation on why it holds, if it has not been proven).
 *
 * COMMON LEMMAS: used in multiple lemmas in fletch_main.v (e.g. subtyping reflexivity).
 * PROGRESS: used only in the progress lemma (mostly lemmas involving the fact that well typed terms in the
 *                                            base cases of the progress proof can take an execution step)
 * PRESERVATION: used only in the preservation lemma (mostly weakening, and 2 kinds of lemmas: "modifying well formed
 *                                                    environments still lead to well formed environments" and
 *                                                    "modifying fully annotated environemtns sill lead to fully annoated
 *                                                     environments")
 *)

(* COMMON LEMMAS *)

(* 
 * bound_Delta0(G, G') = G, i.e., the binding of all the type parameters occurring in G
 * with their upper bound in Delta is neutral.
 *)
Lemma bound_delta_neutral : forall G G', bound_delta Delta0 G G' -> G = G'.
Proof.
  intros G G' BoundDelta.
  remember Delta0 as Delta eqn: EQG_Delta; revert EQG_Delta.
  induction BoundDelta using bound_delta_mut with
                               (P := fun Delta G G' _ => Delta = Delta0 -> G = G')
                               (P0 := fun Delta G G' _ => Delta = Delta0 -> G = G')
                               (P1 := fun Delta GL GL' _ => Delta = Delta0 -> GL = GL')
                               (P2 := fun Delta GL GL' _ => Delta = Delta0 -> GL = GL');
    intros;assert(Trivial : Delta0 = Delta0);trivial.
    - hfree IHBoundDelta H. rewrite <- H0. trivial.
    - hfree IHBoundDelta H. hfree IHBoundDelta0 H. rewrite <- H0,H1. trivial.
    - rewrite -> H in i. unfold Delta0 in i. inversion i.
    - hfree IHBoundDelta H. rewrite <- H0. trivial.
    - hfree IHBoundDelta H. hfree IHBoundDelta0 H. rewrite -> H0 in e. rewrite H1 in e.
      rewrite -> e. rewrite e0. trivial.
    - hfree IHBoundDelta H. hfree IHBoundDelta0 H. rewrite -> H0 in e. rewrite H1 in e.
      rewrite -> e. rewrite e0. trivial.
Qed.

(* 
 * Subtyping reflexivity is an easilly derivable rule. Indeed, subtyping is built on top of the specific typing,
 * which is reflexive, by simply replacing occurrences of the type "dynamic" with "bottom", and other
 * simple congruence rules.
 * Thus, when T does not contain dynamic, Delta |S- T <: T becuase it is equivalent to Delta |- T << T.
 * When T contains dynamic, Delta |S- T <: T is equivalent to Delta |S- T[bottom/dynamic] << T, but clearly 
 * any occurrence of bottom in T[bottom/dynamic] corresponds to an occurrence of dynamic in T, and Delta |- bottom << dynamic.
 *)
Lemma subtyping_reflexivity : forall Delta T, Delta |S- T <: T.
Proof.
  intros Delta T. (* Print fletch_type_mut. -- EE: commented out for quiet make *)
  induction T using fletch_type_mut with
  (P := fun G => Delta |S- (type_nv G) <: (type_nv G))
  (P0 := fun T => Delta |S- T <: T)
  (P1 := fun N => Delta |S- (type_nv (type_classtype N)) <: type_nv (type_classtype N)).
  - Case "X <: X". auto.
  - Case "N <: N". auto.
  - Case "GL -> T <: GL -> T". constructor. simpl. auto.
  - Case "dynamic <: dynamic". auto.
  - Case "bottom <: bottom". auto.
  - Case "G <: G". auto.
  - Case "void <: void". auto.
  - Case "C<GL> <: C<GL>". constructor. simpl. constructor.
    induction GL. simpl. constructor.
    assert(Hyp : (forall G : fletch_nvtype,
          G _in GL -> Delta |S- type_nv G <: type_nv G)).
    intros. assert(Hyp2 : G _in (a :: GL)). constructor 2. auto.
    assert(X := H G Hyp2). auto. admit.
    (* assert(IH := IHGL H). simpl. assert(X : a _in (a :: GL)). constructor. auto.
    assert(ST := H a X). sinversion ST. *)
    (* * constructor. auto. auto. *)
    (* * constructor. auto. auto. *)
Qed.

(* 
 * Similarly to the previous lemma.
 * Plausible: the reflexivity rule is in the << relation
 *)
Lemma subtyping_reflexivity' : forall Delta T, Delta |- T <: T.
Admitted.

(* 
 * This is a standard lemma. If an expression e is well typed, it has a unique type.
 * This can be easilly shown by looking at the typing rules, which are unique. The environment
 * well formedness in the premises of this lemma are necessary because the typing of an expression might require
 * looking up from nu or sigma, to be able to type locations.
 *)
Lemma uniqueness_types : forall nu sigma Gamma T S e,
                           well_formed_env Gamma ->
                           well_formed_lenv sigma ->
                           well_formed_rvenv nu ->
                           nu ; sigma ; Delta0 ; Gamma |- e : T ->
                           nu ; sigma ; Delta0 ; Gamma |- e : S ->
                           T = S.
Proof.
  intros nu sigma Gamma T S e WFGamma WFLenv WFRVenv T_Typing S_Typing. generalize dependent S.
  remember Delta0 as Delta eqn: EQG_Delta; revert EQG_Delta.
  induction T_Typing using exp_typing_mut with
  (P := fun nu sigma Delta Gamma e T _ =>
            well_formed_env Gamma ->
            well_formed_lenv sigma ->
            well_formed_rvenv nu ->
            Delta = Delta0 ->
            forall S,
            nu ; sigma ; Delta ; Gamma |- e : S -> T = S)
  (P0 := fun nu sigma Delta Gamma eL TL _ =>
             well_formed_env Gamma ->
             well_formed_lenv sigma ->
             well_formed_rvenv nu ->
             Delta = Delta0 ->
             forall SL,
             nu ; sigma ; Delta ; Gamma |-^ eL : SL -> TL = SL);intros;subst.
  (* General IDA: inductive case analysis over Typing1, case analysis Typing2 (inversion) + *)
  (* induction hypothesis *)
  - sinversion H0. assert(X := NoDupKeysLemma y G G0 Gamma WFGamma i H5). subst. auto.
  - sinversion H0. assert(Trivial: Delta0 = Delta0). auto.
    assert(X := IHT_Typing WFGamma WFLenv WFRVenv Trivial T0 H2). subst.
    assert(BDN := bound_delta_neutral T0 T'0 H7). subst.
    assert(BDN := bound_delta_neutral T'0 T' b). subst. admit.
    (* Something to fix with induction *)
  - (* Similar to getter case: play with bound_delta and ind. hypothesis *) admit.
  - sinversion H0. assert(Trivial: Delta0 = Delta0). auto.
    assert(X := IHT_Typing WFGamma WFLenv WFRVenv Trivial S H6). auto.
  - sinversion H0;assert(Trivial : Delta0 = Delta0);auto.
    * assert(X := IHT_Typing WFGamma WFLenv WFRVenv Trivial (type_nv (GL0 --> S)) H6).
      sinversion X. auto.
  - sinversion H0. auto.
  - (* apply induction hypothesis to body of e0 *) admit.
  - sinversion H0. destruct WFLenv. assert(X := NoDupKeysLemma l o o0 wenv H b H1). subst.
    auto.
  - sinversion H0. destruct WFRVenv.
    assert(X := NoDupKeysLemma tau (G, l) (G0, l0) wenv H b H5). sinversion X. auto.
  - sinversion H0. assert(Trivial: Delta0 = Delta0). auto.
    assert(X := IHT_Typing WFGamma WFLenv WFRVenv Trivial S H6). auto.
  -  sinversion H0. assert(Trivial: Delta0 = Delta0). auto.
     (* assert(X := IHT_Typing WFGamma WFLenv WFRVenv Trivial R0 H7). subst. auto. *)
     admit.
  - sinversion H0. auto.
  - admit.
Qed.

(* 
 * The subtyping relation for the operational semantics used in the type system
 * includes the subtyping relation used in the operational semantics.
 * Trivial, since the latter is obtained by the former replacing subtyping
 * with assignability in the function type subtyping premises.
 *)
Lemma t_subtyping_relaxed : forall Delta T S,
                              Delta |- T <: S ->
                              Delta |S- T <: S.
Proof.
  intros.
  induction H using subtyping_mut with
  (P := fun Delta T S _ => Delta |S- T <: S)
  (P0 := fun Delta TL SL _ => Delta |S-^ TL <: SL)
  (P1 := fun Delta T S _ => assignable Delta T S)
  (P2 := fun Delta TL SL _ => assignable_list Delta TL SL).
  - econstructor. eauto.
  - econstructor. sinversion IHsubtyping.
    * admit.
    * admit.
  - eauto.
Admitted.

(* 
 * If sigma, nu, and e are fully annoatated, and T is the type of e, then T is fully annoatated.
 *)
Lemma fully_annotated_typing : forall nu sigma e T,
                                 fully_annotated_store sigma ->
                                 fully_annotated_subst nu ->
                                 sigma |--S nu ->
                                 fully_annotated_expression e ->
                                 nu ; sigma ; Delta0 ; Gamma0 |- e : T ->
                                 fully_annotated_type T.
Proof.
  intros.
  remember Gamma0 as Gamma eqn: EQG_Gamma; revert EQG_Gamma.
  remember Delta0 as Delta eqn: EQG_Delta; revert EQG_Delta.
  induction H3 using exp_typing_mut with
  (P := fun nu sigma Delta Gamma e T _ =>
  fully_annotated_store sigma ->
  fully_annotated_subst nu ->
  sigma |--S nu ->
  fully_annotated_expression e ->
  Delta = Delta0 ->
  Gamma = Gamma0 ->
  fully_annotated_type T)
  (P0 := fun nu sigma Delta Gamma eL TL _ => 
  fully_annotated_store sigma ->
  fully_annotated_subst nu ->
  sigma |--S nu ->
  fully_annotated_expression_list eL ->
  Delta = Delta0 ->
  Gamma = Gamma0 ->
  fully_annotated_type_list TL);intros;subst;assert (Trivial_Gamma : Gamma0 = Gamma0);
    assert (Trivial_Delta : Delta0 = Delta0);trivial;trivial.

  - unfold Gamma0 in i. sinversion i. sinversion H3. constructor. sinversion H3.
  - assert(BDN := bound_delta_neutral T T' b). subst. sinversion H2.
    assert(IH := IHexp_typing H H0 H1 H5 Trivial_Delta Trivial_Gamma).
    assert(X := fully_annotated_getter T' G p IH g). sinversion X. auto.
  - assert(BDN := bound_delta_neutral T T' b). subst. sinversion H2. 
    assert(IH := IHexp_typing2 H H0 H1 H7 Trivial_Delta Trivial_Gamma). auto.
  - sinversion H2. assert(IH := IHexp_typing H H0 H1 H5 Trivial_Delta Trivial_Gamma). auto.
  - sinversion H2. assert(IH := IHexp_typing H H0 H1 H6 Trivial_Delta Trivial_Gamma).
    sinversion IH. auto.
  - sinversion H2. auto.
  - sinversion H2. clear IHexp_typing. (* We can't use it an d don't need it *)
    constructor.
    cut(forall (G : fletch_anvtype) (x : generate_id),
       (G, x) _in GxL -> fully_annotated_type (atype_nv G)).
    eauto using fully_annotated_closure_type. auto. auto.
  - assert(FA := H l o b). destruct o;simpl;sinversion FA;auto. 
    sinversion FA. 
    constructor.
    cut(forall (G : fletch_anvtype) (x : generate_id),
        (G, x) _in l0 -> fully_annotated_type (type_nv G)).
    eauto using fully_annotated_closure_rtype. auto. auto. constructor.
  - assert(FA := H0 tau l G b). auto.
  - sinversion H2. assert(X := IHexp_typing H H0 H1 H5 Trivial_Delta Trivial_Gamma). auto.
  - sinversion H2. auto.
  - constructor.
  - sinversion H2. assert(IH := IHexp_typing H H0 H1 H6 Trivial_Delta Trivial_Gamma).
    assert(IHL := IHexp_typing0 H H0 H1 H7 Trivial_Delta Trivial_Gamma). constructor.
    auto. auto.
Qed.

(*
A class table is fully annotated by assumption. Therefore, if GL is fully annotated and
C in in the class table, the runtime object (C<GL>, FL, ML) is fully annotated because
new dynamic occurrences are never introduced.
*)
Lemma fully_annotated_hierarchy : forall C PL N FL0 ML0 GL ML,
                                    (C, class C PL N FL0 ML0) _in CT ->
                                    get_methods (atype_class C GL) ML ->
                                    fully_annotated_type_list
                                      (ltype_nv (map convert_nvtype GL)) ->
                                    fully_annotated_methods ML.
Admitted.

(* PROGRESS *)

(* Just some machinery to match different representations of class types *)
Lemma type_coertion : forall C GL C' GL',
                        C <<< GL >>> = atype_class C' GL' ->
                        C = C'.
Proof.
  intros. simpl in H. inversion H. trivial.
Qed.                  

(* 
 * If l.p is well typed, than it reduces.  
 * Clearly l.p is never stuck since the environemnts well formedness and the fact that 
 * l.p is well typed ensure that "l" to map to an object that provides p, therefore message-not-understood
 * cannot occur. No other error can happen, therefore the expression should reduce.
 * Note that l_null.p is an acceptable configuration error, but it cannot happen because l_null.p is not well typed.
 *)
Lemma getter_base_reduction : forall l p nu sigma Gamma T,
                          well_formed_store nu sigma ->                          
                          well_formed_lenv sigma ->
                          fully_annotated_store sigma ->
                          nu ; sigma ; Delta0 ; Gamma |- (exp_getter (exp_location l) p) : T ->
                          (exists l',
                            << nu, sigma, (exp_getter (exp_location l) p) >> --->
                            << nu, sigma, (exp_location l') >>).
Proof.
  intros. sinversion H2. assert(BDN := bound_delta_neutral T0 T' H10). sinversion H5. 
  clear H10.
  assert(X := H l o H4). sinversion H12. sinversion H5. 
  sinversion X. unfold well_formed_fields in H3. destruct H3 as [ FL [ Fields [ WF NDup ] ] ]. 
  simpl in H6. sinversion H12. admit. admit. admit. admit. admit. 
  (* assert(WFF := WF f G). *)
  (* assert(WFF := H9 f C G H6). *)
  (*  sinversion H3. simpl in H6. assert(WFF := H9 f *)
  (*                                    G G H6). *)
  (* destruct WFF as [ l' [ o [ A [ B [ C' D ] ] ] ] ]. *)
  (*  clear C'. exists l'. econstructor. eauto. eauto. simpl in H5. *)
  (*  sinversion H6. sinversion H6. sinversion X. destruct H7. simpl in H6.  *)
  (*  assert(WFM := H7 m G). destruct WFM as [ o [ l' A ] ]. assert(Y := A H6). *)
  (*  destruct Y as [ B [ C' D ] ]. exists l'. econstructor. eauto. eauto. *)
  (*  sinversion H6. sinversion H6. *)
Qed.

(* 
 * If l.p = e is well typed and all the environments are well formed, then l.p = e either reduces,
 * or it is ends with a subtype violation error. The proof reasoning is similar to the former lemma,
 * involving getters.
 *)
Lemma setter_base_reduction : forall l1 l2 p nu sigma Gamma T,
                          well_formed_store nu sigma ->                          
                          well_formed_lenv sigma ->
                          fully_annotated_store sigma ->
                          nu ; sigma ; Delta0 ; Gamma |- (exp_setter (exp_location l1)
                                                                    p
                                                                    (exp_location l2)) : T ->
                          (exists sigma',
                            << nu, sigma, (exp_setter (exp_location l1)
                                                      p
                                                      (exp_location l2)) >> --->
                            << nu, sigma', (exp_location l2) >>) \/
                          error nu sigma (exp_setter (exp_location l1)
                                                     p
                                                     (exp_location l2)).
Proof.
  intros. sinversion H2. sinversion H13. clear H5. sinversion H6.
  assert(X := H l1 o H4). assert(BDN := bound_delta_neutral (typeof o) T' H11). subst. 
  clear H11. sinversion H14. sinversion X.
  - Case "l1 |-> class instance".
    clear H5. destruct H3 as [ Y NDup ]. sinversion H14.
  (*   assert(Z := Y f G G H9). destruct Z as [ l' [ o' [ InRFL [ INSigma [ EQ ST ] ] ] ] ]. *)
  (*   assert(Update := updatable_lenv l1 (class_instance_object (atype_class C GL) RFL RML) *)
  (*                    (class_instance_object (atype_class C GL) *)
  (*                    (update_runtime_field RFL (f, (G, l2))) RML) sigma H4). *)
  (*   destruct Update as [ sigma' Update ]. *)
  (*   emdestruct Reduces Error (subtyping_em Delta0 (typeof o) (type_nv G)). *)
  (*   * SCase "l1.p = l2 -> l2". *)
  (*     left. exists sigma'. econstructor. eauto. eauto. eauto. eauto. eauto. *)
  (*   * SCase "l1.p = l2 is an error". *)
  (*     right. econstructor 3. eauto. eauto. eauto. eauto. sinversion H15. eauto. *)
  (* - Case "l1 |-> closure". *)
  (*   sinversion H9. *)
  (* - Case "l1 |-> null". *)
  (*   sinversion H9. *)
Admitted.

(* 
 * If the environments are well formed, and l(eL) is well typed (where l is a location and eL a list of locations), then it reduces or there is a
 * subtype violation caused by one of the locations in the list eL.
 * Message not understood (the only stuck configuration that is not an acceptable error) cannot happen because in order for l(eL) to be
 * well typed, l has to point to a closure with the same number of parameters as the size of the list eL.
 *)
Lemma call_base_reduction : forall l nu sigma Gamma T eL,
                          well_formed_store nu sigma ->                          
                          well_formed_lenv sigma ->
                          fully_annotated_store sigma ->
                          nu ; sigma ; Delta0 ; Gamma |- (exp_call (exp_location l) eL) : T ->
                          values eL ->
                          (exists T e' nu',
                              << nu, sigma, (exp_call (exp_location l) eL) >>
                              --->
                              << nu', sigma, exp_frame_expression T e' >>)  \/
                          error nu sigma (exp_call (exp_location l) eL).
Proof.
  intros. sinversion H2.
  - Case "l(eL) : closure".
    (* Precondiction 1: (l, closure) __in sigma *)
    sinversion H10. destruct o;simpl in H11. sinversion H11. Focus 2. sinversion H11.
    sinversion H11.
    (* Precondiction 2: T OK *)
            (* claiming that the environments are OK *)
            assert (HtenvOK: |-t Delta0 OK) by admit.
            assert (HvenvOK: Delta0 |-v Gamma OK) by admit.
    assert(TOK := well_typed_well_formed nu sigma Delta0 Gamma (exp_call (exp_location l) eL)
                  T HtenvOK HvenvOK H2).
    
    (* Precondition 3: actual_to_formals *)
    assert(VL := values_are_locations eL H3). destruct VL as [ LL Loc ]. 
    assert(ATF := actuals_to_formals_exists nu l0 LL). destruct ATF as [ nu' [ tauL ATF ] ].
    (* Precondition 4: variable substitution *)
    assert(VS := variable_substitution_exists f0 l0 tauL). destruct VS as [ e' MVS ].
    emdestruct Reduces Error (location_subtyping_em sigma LL (runtime_types l0)).
    * SCase "Reduction".
      left. exists T, e', nu'. econstructor. eauto. eauto. eauto. eauto. eauto. eauto.
    * SCase "Error".
      right. econstructor 4. eauto. eauto. eauto. eauto.
Qed.

(* 
 * If the environments are well formed and new N() is well typed, then it can be evaluated.
 *)
Lemma constructor_base_reduction : forall nu sigma Gamma Delta N,
                          well_formed_store nu sigma ->                          
                          well_formed_lenv sigma ->
                          fully_annotated_store sigma ->
                          nu ; sigma ; Delta ; Gamma |- (exp_new N) :
                                                        (type_nv (type_classtype N)) ->
                          (exists l' nu' sigma',
                              << nu, sigma, (exp_new N) >>
                              --->
                              << nu', sigma', (exp_location l') >>).
Proof.
  intros. 
  destruct N as [ C GL ]. sinversion H2. sinversion H8.
  (* Precondition 1: insert C<GL> *)
  assert(InsertLenv := insertable_lenv empty_location sigma).
  destruct InsertLenv as [ (l_base, sigma') InsertLenv ].
  (* Precondition 2: insert_rvenv *) 
  assert(InsertRVenv := insertable_rvenv l_base (atype_classtype (atype_class C GL))
                        nu ).
  destruct InsertRVenv as [ nu' [ v_base InsertRVenv ] ].
  (* Precondition 3: bind fields GxL GL FL RFL *)
  (* assert(BindFields : exists RFL, bind_fields PL GL FL RFL). *)
  (* admit. *)
  (* destruct BindFields as [RFL BindFields ]. *)
  (* (* Precondition 4: bind methods v_base ...*) *)
  (* assert(BindMethods : exists RML sigma'', *)
  (*                        bind_methods v_base (atype_class C GL) sigma' ML (RML, sigma'')). *)
  (* admit. *)
  (* destruct BindMethods as [ RML [ sigma'' BindMethods ] ]. *)
  (* (* Precondition 5: update_lenv sigma'' ... ... *) *)
  (* assert(UpdateLenv : update_lenv sigma'' l_base *)
  (*    (class_instance_object (atype_class C GL) RFL RML) sigma''). *)
  (* admit. *)
  (* exists l_base, nu', sigma''. eauto. *)

  (* (* Precondition 1: insert C<GL> *) *)
  (* assert(InsertLenv : exists l_base sigma', insert_lenv sigma empty_location (l_base, sigma')). *)
  (* admit. *)
  (* destruct InsertLenv as [ l_base [ sigma' InsertLenv ] ]. *)
  (* (* Precondition 2: insert_rvenv *) *)
  (* assert(InsertRVenv : exists v_base nu', *)
  (*                        insert_rvenv nu *)
  (*                         (atype_classtype (atype_class builtin_class_object GL), l_base) *)
  (*                         (v_base, nu')). *)
  (* admit. *)
  (* destruct InsertRVenv as [ v_base [ nu' InsertRVenv ] ]. *)
  (* (* Precondition 3: bind fields GxL GL FL RFL *) *)
  (* assert(BindFields : exists RFL, bind_fields [] GL [] RFL). *)
  (* admit. *)
  (* destruct BindFields as [RFL BindFields ]. *)
  (* (* Precondition 4: bind methods v_base ...*) *)
  (* assert(BindMethods : exists RML sigma'', *)
  (*                        bind_methods v_base (atype_class builtin_class_object [])  *)
  (*                                     sigma' [] (RML, sigma'')). *)
  (* admit. *)
  (* destruct BindMethods as [ RML [ sigma'' BindMethods ] ]. *)
  (* (* Precondition 5: update_lenv sigma'' ... ... *) *)
  (* assert(UpdateLenv : update_lenv sigma'' l_base *)
  (*    (class_instance_object (atype_class builtin_class_object GL) RFL RML) sigma''). *)
  (* admit. *)
  (* exists l_base, nu', sigma''. destruct GL. econstructor. eauto. eauto. eauto. eauto. eauto. *)
  (* eauto. eauto. sinversion H8. assert(X := well_formed_CT). destruct X as [ ND [ NI CTT ] ]. *)
  (* assert(Contradiction := NI (class builtin_class_object PL N FL ML) *)
  (*                         H5). contradiction. Grab Existential Variables. *)
  (* exact (atype_class builtin_class_object []). *)
Admitted.

(* 
 * Closures can be reduced into a location. This is always possible, since the only relevant
 * premise requires the closures to be inserted the current environment, bound to a fresh location.
 *)
Lemma function_base_reduction : forall nu sigma Gamma Delta e T GxL GL,
                          well_formed_store nu sigma ->                          
                          well_formed_lenv sigma ->
                          fully_annotated_store sigma ->
                          nu ; sigma ; Delta ; Gamma |- (exp_function GxL T e) :
                                                          (type_nv ((GL) --> T)) ->
                           (exists l sigma',
                              << nu, sigma, (exp_function GxL T e) >>
                              --->
                              << nu, sigma', (exp_location l) >>).
Admitted.

(*
 * If tau = l is well typed, than it means that typeof(l) <: typeof(tau) and therefore the
 * checked-mode checks will pass allowing for the reduction.
 *)
Lemma runtime_assignment_base_reduction : forall nu sigma tau Gamma Delta T l,
                          well_formed_store nu sigma ->                          
                          well_formed_lenv sigma ->
                          fully_annotated_subst nu ->
                          nu ; sigma ; Delta ; Gamma |- (exp_runtime_assignment tau
                                                        (exp_location l)) : T ->
                           (exists nu',
                              << nu, sigma, (exp_runtime_assignment tau
                                                        (exp_location l)) >>
                              --->
                              << nu', sigma, (exp_location l) >>).
Admitted.

(*
 * If the environments are well formed and [[T, l]] is well typed, then [[T, l]] -> l,
 * or there is a subtype violation error because the type of l is not a subtype of T.
 *)
Lemma frame_expression_base_reduction : forall nu sigma T R l Gamma,
                          nu ; sigma ; Delta0 ; Gamma |- (exp_frame_expression T
                                                         (exp_location l)) : R ->
                              << nu, sigma, (exp_frame_expression T
                                                         (exp_location l)) >>
                              --->
                              << nu, sigma, (exp_location l) >> \/
                              error nu sigma (exp_frame_expression T (exp_location l)).
Proof.
  intros. sinversion H. sinversion H7.
  emdestruct A B (sem_subtyping_em Delta0 (typeof o) R).
  - left. econstructor. eauto. assert(X := t_subtyping_relaxed Delta0 (typeof o) R). eauto.
  - right. econstructor. eauto. assert(X := t_subtyping_relaxed Delta0 (typeof o) R). 
    intuition.
Qed.

(* 
 * Just Coq machinery.
 *)
Lemma list_values : forall nu sigma l eL,
                      error_list nu sigma ((exp_location l) :: eL) ->
                      error_list nu sigma eL.
Proof.
  intros. sinversion H. inversion H3. auto.
Qed.

(* 
 * Just Coq machinery to show that values (i.e., locations) are not error configurations.
 *)
Lemma values_not_errors : forall nu sigma eL,
             values eL ->
             ~ error_list nu sigma eL.
Proof.
  intros. intro F. induction H. inversion F. destruct H. subst. sinversion F.
  inversion H3. contradiction.
Qed.

(* 
 * Just Coq machinery to show that values are not error configurations (variant of the previous lemma).
 *)
Lemma list_value : forall eL e nu sigma,
                      values eL ->
                      error_list nu sigma (e :: eL) ->
                      error nu sigma e.
Proof.
  intros. sinversion H0. auto.
  assert(Contr := values_not_errors nu sigma eL H). contradiction.
Qed.

(* 
 * Just Coq machinery to show that a list of locations is a list of values.
 *)
Lemma locations_values : forall LL LLRef,
                           locations LL LLRef ->
                           values LL.
Proof.
  intros. induction H.
  constructor. constructor. exists l. auto. auto.
Qed.

(* 
 * Just Coq machinery to show that locations are not reducible.
 *)
Lemma locations_not_reduce : forall nu nu' sigma sigma' e e',
                               values e ->
                               ~ [[ nu, sigma, e ]] --->^ [[ nu', sigma', e' ]].
Proof.
  intros. intro F. induction F. sinversion H. destruct H3. subst. sinversion H0.
  sinversion H. contradiction.
Qed.

(* PRESERVATION  *)

(* EE: 'fads' and 'gfads' are not currently used; commented out for simplicity

Ltac fads T FA_T Exp :=
  let Eq := fresh
             in assert(Eq := dyn_sub_trivial T FA_T); rewrite -> Eq in Exp.

Ltac gfads T FA_T :=
  let Eq := fresh 
           in assert(Eq := dyn_sub_trivial T FA_T); rewrite -> Eq.
*)

(*
 * If the store is well formed, if l1.f contains l2 and the store is well formed, then
 * l2 must be well typed with a type that is a subtype of the type of T (the static type of
 * f). Trivial. Since the environment is well formed, every field of every object is well formed, meaning that
 * the type of l2 is a subtype of T, the declared type of f. 
*)
Lemma field_lookup_well_formed : forall N RFL RML l1 l2 nu sigma f R T,
                                  (l1, (class_instance_object N RFL RML)) __in sigma ->
                                  (f, (T, l2)) _in RFL ->
                                  nu ; sigma ; Delta0 ; Gamma0 |- (exp_getter
                                                                  (exp_location l1)
                                                                  (field_property f)) : R ->
                                  well_formed_object nu sigma 
                                                     (class_instance_object N RFL RML) ->
                                  exists S,
                                    nu ; sigma ; Delta0 ; Gamma0 |- (exp_location l2) : S /\
                                    Delta0 |- S <: R.
Admitted.

(* 
  Same considerations as above, but this lemma is involving fields instead of methods.
*)
Lemma method_lookup_well_formed : forall N RFL RML l1 l2 nu sigma m R,
                                  (l1, (class_instance_object N RFL RML)) __in sigma ->
                                  (m, l2) _in RML ->
                                  nu ; sigma ; Delta0 ; Gamma0 |- (exp_getter
                                                                  (exp_location l1)
                                                                  (method_property m)) : R ->
                                  well_formed_object nu sigma 
                                                     (class_instance_object N RFL RML) ->
                                  exists S,
                                    nu ; sigma ; Delta0 ; Gamma0 |- (exp_location l2) : S /\
                                    Delta0 |- S <: R.
Admitted.

(* 
  Let G be the type of f in an object instance of the class C<GL> and G' be the static 
  type of f in C<GL>, then G = G' and therefore if typeof o <: G then typeof o <: G'
  for each object o.
*)
Lemma subtyping_well_formed : forall C GL FL ML sigma nu l l' f G G' o,
                                well_formed_object nu sigma (class_instance_object
                                                            (atype_class C GL) FL ML) ->
                                (l, (class_instance_object (atype_class C GL) FL ML))
                                      __in sigma ->
                                (f, (G, l')) _in FL ->                                
                                (ftype (type_nv (type_classtype
                                       (C <<< convert_anvtype_list GL >>>))) f G') ->
                                Delta0 |S- typeof o <: (type_nv G) ->
                                Delta0 |S- typeof o <: (type_nv G').
Admitted.

(* 
  If an object o is well formed, then the insertion in a well formed store, preserves its
  well formedness since the insertion only add such object, and only if it does not violate
  the well formedness.
*)
Lemma lenv_insert_preserves_WF : forall nu sigma sigma' l o,
                                   well_formed_store nu sigma ->
                                   well_formed_object nu sigma o ->
                                   insert_lenv sigma o (l, sigma') ->
                                   well_formed_store nu sigma'.
Proof.
  unfold well_formed_store. intros.
  emdestruct IsO IsNotO (object_em o o0);subst.
  - assert(WFO := lenv_insert_preserves_well_formed_object nu sigma o0 sigma' l H0 H1). auto.
  - admit. (* evident *)
Qed.  

(*
  Same considerations as above.
*)
Lemma lenv_update_preserves_WF : forall nu sigma sigma' l o,
                                   well_formed_store nu sigma ->
                                   update_lenv sigma l o sigma'  ->
                                   well_formed_store nu sigma'.
Proof.
  intros. unfold well_formed_store. intros.
  emdestruct IsO IsNotO (object_em o o0);subst.
  - admit.
  - admit.
Qed.

(* 
 * If sigma contains the null location, then, assuming inserting (l, o) in to sigma gives sigma', then
 * sigma' contains the null location too.
 * Trivial, since the environment object insertion is monotonic.
 *)
Lemma lenv_insert_preserves_sigma0 : forall sigma sigma' l o,
                                       lenv_base_superset sigma ->
                                       insert_lenv sigma o (l, sigma') ->
                                       lenv_base_superset sigma'.
Proof.
  intros. unfold lenv_base_superset. intros.
  
  assert(X := H l0 o0 H1). 
  emdestruct IsO IsNotO (object_em o o0);subst.
  - destruct sigma. sinversion H0.
    assert(C := generate_fresh_key_correctness g o0 f (env', generate_fresh_location g)
                H0). sinversion H5. admit.
  - assert(C := (@insert_fresh_add _ _ generate_fresh_location) l0 l o0 o sigma sigma' X H0).
    auto.
Qed.

(*
 * If sigma contains the null location, then sigma[l |-> o] (the environment obtaining updating the location l) contains
 * the null location too.
 * It is clear that if sigma0 \subset sigma then sigma0 \subset[l -> o] since the update
 * does not remove any object from the store.
 *)
Lemma lenv_update_preserves_sigma0 : forall sigma sigma' l o,
                                       lenv_base_superset sigma ->
                                       update_lenv sigma l o sigma' ->
                                       lenv_base_superset sigma'.
Proof.
Admitted.

(* 
 * Same considerations as above.
 *)
Lemma bind_methods_preserves_sigma0 : forall sigma sigma' l m N RML,
                                   lenv_base_superset sigma ->
                                   bind_methods l N sigma m (RML, sigma') ->
                                   lenv_base_superset sigma'.
Proof.
Admitted.

(* 
 * The insertion only add environments without creating duplicates (well formedness definition).
 * Therefore the insertion in nu always preserves the environment well formedness.
 *)
Lemma rvenv_insert_preserves_WF : forall nu nu' tau_fresh G l,
                                   well_formed_rvenv nu ->                                   
                                   insert_rvenv nu (G, l) (tau_fresh, nu') ->
                                   well_formed_rvenv nu'.
Proof.
Admitted.

(* 
 * Same considerations as above.
 *)
Lemma rvenv_update_preserves_WF : forall nu nu' tau o,
                                   well_formed_rvenv nu ->                                   
                                   update_rvenv nu tau o nu' ->
                                   well_formed_rvenv nu'.
Proof.
Admitted.

(* 
 * bind_method adds method in the object referenced by l. This operation does not add any
 * location and it does not modify any location different from l. Adding methods to an object
 * is safe since this are all distincts, coming from the method lookup. Therefore the
 * new store is still well formed.
 *)
Lemma bind_methods_preserves_WF : forall nu sigma sigma' l m N RML,
                                   well_formed_store nu sigma ->
                                   bind_methods l N sigma m (RML, sigma') ->
                                   well_formed_store nu sigma'.
Proof.
Admitted.

(* 
 * Same considerations as above.
 *)
Lemma bind_methods_lenv_WF : forall sigma sigma' l m N RML,
                                   well_formed_lenv sigma ->
                                   bind_methods l N sigma m (RML, sigma') ->
                                   well_formed_lenv sigma'.
Proof.
Admitted.

(* 
 * Since typeof o <: G, adding l:G in nu preserves well formedness because the insert only add
 * a fresh location to nu. Moreover, the sigma |--S nu' requirement is that
 * typeof o <: G, that is our assumption.
 *)
Lemma rvenv_insert_preserves_WFSigmaRho : forall sigma nu (G : fletch_anvtype) l nu' o fresh,
                                            sigma |--S nu ->
                                            (l, o) __in sigma ->
                                            Delta0 |S- typeof o <: (type_nv G) ->
                                            insert_rvenv nu (G, l) (fresh, nu') ->
                                            sigma |--S nu'.
Proof.
Admitted.

(* 
 * Same considerations as above: the update operation consistently adds (due to our hipothesys)
 * the new l location to the environment.
 *)
Lemma rvenv_update_preserves_WFSigmaRho : forall tau sigma nu (G : fletch_anvtype) l l' nu' o,
                                            sigma |--S nu ->
                                            (l, o) __in sigma ->
                                            Delta0 |S- typeof o <: (type_nv G) ->
                                            (tau, (G, l')) __in nu ->
                                            update_rvenv nu tau (G, l) nu' ->
                                            sigma |--S nu'.
Proof.
Admitted.

(* 
  The insertion of well formed object ins sigma is clearly consistent, since a store
  is well formed if all the objects are, but sigma is well formed and the insert operation
  only adds a new wellf formed object, which is o.
*)
Lemma lenv_insert_preserves_well_formed_store : forall nu sigma sigma' o n,
                                            fully_annotated_store sigma ->
                                            fully_annotated_subst nu ->
                                            well_formed_store nu sigma ->
                                            well_formed_object nu sigma o ->
                                            insert_lenv sigma o (n, sigma') -> 
                                            well_formed_store nu sigma'.
Admitted.                                                   

(* 
  Replacing tauL with the expression list is safe since the type of the actual arguments
  are subtypes of the type of formal parameters. Thus, the new environment contains
  valid bindings so that nu' is well formed (it is if foreach (tau:T) -> l, typeof l <: T, that
  is our assumption.
*)
Lemma actual_to_formals_preserves_well_formed_store : forall sigma nu nu' tauL LLref
                                                            GxL Delta Gamma LLRef TL e e',
                                         well_formed_store nu sigma ->
                                         multiple_vloc_subst e (rvar_to_var GxL) tauL e' ->
                                         actuals_to_formals nu GxL LLref (tauL, nu') ->
                                         locations LLRef LLref ->
                                         location_subtyping sigma LLref (runtime_types GxL) ->
                                         nu ; sigma ; Delta ; Gamma |-^ LLRef : TL ->
                                         well_formed_store nu' sigma.
Admitted.

(* 
  Same considerations as above.
*)
Lemma actual_to_formals_preserves_well_formed_rvenv : forall sigma nu nu' tauL LLref
                                                            GxL Delta Gamma LLRef TL e e',
                                         well_formed_store nu sigma ->
                                         multiple_vloc_subst e (rvar_to_var GxL) tauL e' ->
                                         actuals_to_formals nu GxL LLref (tauL, nu') ->
                                         locations LLRef LLref ->
                                         location_subtyping sigma LLref (runtime_types GxL) ->
                                         nu ; sigma ; Delta ; Gamma |-^ LLRef : TL ->
                                         well_formed_rvenv nu'.
Admitted.

(*
  Same considerations as above.
*)
Lemma rvenv_insert_preserves_well_formed_store : forall nu nu' sigma (G : fletch_anvtype)
                                                        l o n,
                                            fully_annotated_store sigma ->
                                            fully_annotated_subst nu ->
                                            well_formed_store nu sigma ->
                                            (l, o) __in sigma ->
                                            well_formed_object nu sigma o ->
                                            Delta0 |- typeof o <: (type_nv G) ->
                                            insert_rvenv nu (G, l) (n, nu') ->
                                            well_formed_store nu' sigma.
Admitted.                                                   

(*
  Same considerations as above.
*)
Lemma rvenv_update_preserves_well_formed_store : forall nu nu' tau sigma G l l',
                                            fully_annotated_store sigma ->
                                            fully_annotated_subst nu ->
                                            well_formed_store nu sigma ->
                                            (tau, (G, l)) __in nu ->
                                            update_rvenv nu tau (G, l') nu' ->
                                            well_formed_store nu' sigma.
Admitted.                                                   

(*
 * Same considerations as above.
 *)
Lemma rvenv_update_preserves_fully_annotated_store : forall nu nu' sigma tau G l l',
                                            fully_annotated_store sigma ->
                                            fully_annotated_subst nu ->
                                            (tau, (G, l)) __in nu ->
                                            update_rvenv nu tau (G, l') nu' ->
                                            fully_annotated_store sigma.
Admitted.                                                   

(* 
 * bind_methods creates a closure for each method in ML associated to fresh locations.
 * sigma' is still well formed because bind_methods just copy the method in ML into the
 * store. We assume that ML is well formed since they always come from a class table, that is
 * well formed by our assumption.
 *)
Lemma well_formed_bind_methods : forall v_base N ML RML sigma sigma',
  well_formed_lenv sigma ->
  bind_methods v_base N sigma ML (RML, sigma') ->
  well_formed_lenv sigma'.
Admitted.

(*
 * If sigma |- nu, and (l, o) is inserted into sigma generating sigma', where l is fresh, then sigma' |- nu.
 *)
Lemma lenv_insert_preserves_WFSigmaRho : forall sigma sigma' nu fresh o,
                                           sigma |--S nu ->
                                           insert_lenv sigma o (fresh, sigma') ->
                                           sigma' |--S nu.
Proof.
  intros.
    unfold well_formed_substenv_lenv. unfold well_formed_substenv_lenv in H.
    sinversion H0. intros. assert(X := H tau G l H1). destruct X. destruct H2.
    exists x. split. sinversion H0. sinversion H8. 
    assert(Y := in_cons ((generate_fresh_location k), o)
                        (l, x)
                        env H2). auto.
    auto.
Qed.

(*
 * The creation of fresh locations bound to actual arguments is safe. Since sigma |--S nu,
 * the LLref types are subtype of GxL and actual_to_formals just adds the new locations, then
 * sigma |--S nu' because it means that for each tau in dom(nu'), 
 * typeof(corresponding location) <: typeof(tau) but this is our assumption.
 *)
Lemma substitution_preserves_WFSigmaRho :
  forall sigma nu nu' tauL LLref GxL Delta Gamma LLRef TL e e',
    sigma |--S nu ->
    multiple_vloc_subst e (rvar_to_var GxL) tauL e' ->
    actuals_to_formals nu GxL LLref (tauL, nu') ->
    locations LLRef LLref ->
    location_subtyping sigma LLref (runtime_types GxL) ->
    nu ; sigma ; Delta ; Gamma |-^ LLRef : TL ->
    sigma |--S nu'.
Admitted.

(* 
 * Same considerations as above.
 *)
Lemma substitution_preserves_WFNu :
  forall sigma nu nu' tauL LLref GxL Delta Gamma LLRef TL e e', 
    well_formed_rvenv nu ->
    multiple_vloc_subst e (rvar_to_var GxL) tauL e' ->
    actuals_to_formals nu GxL LLref (tauL, nu') ->
    locations LLRef LLref ->
    location_subtyping sigma LLref (runtime_types GxL) ->
    nu ; sigma ; Delta ; Gamma |-^ LLRef : TL ->
    well_formed_rvenv nu'.
Admitted.

(*
 * Same considerations as above.
 *)
Lemma lenv_update_preserves_WFSigmaRho : forall sigma sigma' nu l o o',
                                           sigma |--S nu ->
                                           (l, o) __in sigma ->
                                           update_lenv sigma l o' sigma' ->
                                           sigma' |--S nu.
Proof.
Admitted.

(* 
 * The method minding add new locations. Therefore these location are never referenced in
 * nu, so sigma' |--S nu since it means that for each (tau : T -> l) in nu, T <: typeof(l) but
 * it is true that l \in sigma, so this is guaranteed by sigma |--S nu.
 *)
Lemma lenv_bind_preserves_WFSigmaRho : forall sigma sigma' nu v_base N ML RML,
                                         sigma |--S nu ->
                                         bind_methods v_base N sigma ML (RML, sigma') ->
                                         sigma' |--S nu.
Admitted.

(* 
 * Just some coq machinery.
 *)
Lemma fresh_preserves_WFSigmaRho : forall env fresh_1 fresh_2 nu,
                                     (env, fresh_1) |--S nu ->
                                     (env, fresh_2) |--S nu.
Proof.
  intros. unfold well_formed_substenv_lenv. unfold well_formed_substenv_lenv in H.
  intros. assert(X := H tau G l H0). destruct X. destruct H1. exists x. auto.
Qed.

(* 
 * If a fully annotated object is inserted into a fully annoated store, the resulting store
 * contains only fully annoated objects (the ones in sigma, and the new one).
 *)
Lemma lenv_insert_preserves_fully_annotated : forall sigma sigma' l o,
                                                fully_annotated_store sigma ->
                                                well_formed_lenv sigma ->
                                                fully_annotated_type (typeof o) ->
                                                insert_lenv sigma o (l, sigma') ->
                                                fully_annotated_store sigma'.
Admitted.

(* 
 * Same considerations as above.
 *)
Lemma rvenv_insert_preserves_fully_annotated : forall tau nu nu' (G : fletch_anvtype) l,
                                                fully_annotated_subst nu ->
                                                well_formed_rvenv nu ->                         
                                                fully_annotated_type (type_nv G) ->
                                                insert_rvenv nu (G, l) (tau, nu') ->
                                                fully_annotated_subst nu'.
Admitted.  

(* 
  Same considerations as above.
*)
Lemma rvenv_update_preserves_fully_annotated : forall tau nu nu' G l l0,
                                                fully_annotated_subst nu ->
                                                well_formed_rvenv nu ->
                                                (tau, (G, l0)) __in nu ->
                                                fully_annotated_type (type_nv G) ->
                                                update_rvenv nu tau (G, l) nu' ->
                                                fully_annotated_subst nu'.
Admitted.

(* 
  Same considerations as above.
*)
Lemma lenv_update_preserves_fully_annotated : forall sigma sigma' l o,
                                                fully_annotated_store sigma ->
                                                well_formed_lenv sigma ->
                                                fully_annotated_type (typeof o) ->
                                                update_lenv sigma l o sigma' ->
                                                fully_annotated_store sigma'.
Admitted.  

(* 
  Trivial: if list(G x) and T are fully annotated, then list(G) -> T is a fully annoated
  type.
*)
Lemma fully_annotated_ftype :
  forall GxL T,
         (forall (G : fletch_anvtype) x, (G, x) _in GxL -> fully_annotated_type (type_nv G)) ->
          fully_annotated_type T ->
          fully_annotated_type (type_nv (runtime_types (GxL) --> T)).
Admitted.

(* Just some Coq machinery. *)
Lemma runtime_types_eq : forall GxL,
   (runtime_types GxL) =
   (projection GxL).
Proof.
  induction GxL.
  - simpl. auto.
  - simpl. destruct a. unfold projection in IHGxL. destruct GxL;rewrite <- IHGxL;auto.
Qed.

(*
 * If (l, o) i sin sigma, modifying the object o with bind_methods still (l, o) in sigma', where sigma' is the
 * environment obtained by sigma after bind_method.
 * Trivial: bind_methods does not remove any location in sigma, and in general environments are monotonic.
 *)
Lemma bind_methods_add : forall sigma sigma' v_base ML RML l o N,
                                   (l, o) __in sigma ->
                                   bind_methods v_base N sigma ML (RML, sigma') ->
                                   (l, o) __in sigma'.
Admitted.

(* 
 * if bind_methods binds the ML list of fully annoated methods into a fully annoated store,
 * the result is clearly a fully annoated store.
 *)
Lemma bind_methods_preserves_fully_annotated : forall sigma sigma' l N ML RML,
   fully_annotated_store sigma ->
   fully_annotated_methods ML ->
   bind_methods l N sigma ML (RML, sigma') ->
   fully_annotated_store sigma'.
Admitted.

(* 
 *  Same considerations as above.
 *)
Lemma variable_substitution_preserves_fully_annotated :
  forall e (GxL : params) e' tauL,
    fully_annotated_expression e ->
    (forall G x, (G,x) _in GxL -> fully_annotated_type (type_nv G)) ->
    multiple_vloc_subst e (rvar_to_var GxL) tauL e' ->
    fully_annotated_expression e'.
Admitted.

(* 
 * Same considerations as above.
 *)
Lemma actual_to_formals_preserves_fully_annotated : 
  forall nu nu' (GxL : params) LLref tauL,
    well_formed_rvenv nu ->
    fully_annotated_subst nu ->
    (forall G x, (G,x) _in GxL -> fully_annotated_type (type_nv G)) ->
    actuals_to_formals nu GxL LLref (tauL, nu') ->
    fully_annotated_subst nu'.
Admitted.

(* 
 * Same considerations as above.
 *)
Lemma variable_substitution_preserves_typing :
  forall nu nu' sigma Gamma Gamma' Delta e (GxL : params) tauL e' LLref T TL LLRef,
    fold_pred update_venv Gamma (params_to_venv GxL) Gamma' ->
    nu ; sigma ; Delta ; Gamma' |- e : T ->
    multiple_vloc_subst e (rvar_to_var GxL) tauL e' ->
    actuals_to_formals nu GxL LLref (tauL, nu') ->
    locations LLRef LLref ->
    location_subtyping sigma LLref (runtime_types GxL) ->
    nu ; sigma ; Delta ; Gamma |-^ LLRef : TL ->
    (exists S, nu' ; sigma ; Delta ; Gamma |- e' : S /\ Delta |- S <: T).
Admitted.

(*
 * Since overriding is safe in message-safe programs, and T <: S, then
 * if p has type G in T, and type G' in S, then G <: G' because S is a subclass of T and
 * the type of p can only be more specific in subclassing.
 *)
Lemma getter_subtyping : forall T S G p,
                           fully_annotated_type T ->
                           fully_annotated_type S ->                           
                           Delta0 |- S <: T ->
                           getter T p ([] --> type_nv G) ->
                           exists G',
                             getter S p ([] --> type_nv G') /\
                             bound_delta Delta0 S S /\
                             Delta0 |- (type_nv G') <: (type_nv G).
Admitted.

(* 
 * Same considerations as above.
 *)
Lemma setter_subtyping : forall T S R R' G p,
                           fully_annotated_type T ->
                           fully_annotated_type S ->
                           fully_annotated_type R ->
                           fully_annotated_type R' ->
                           Delta0 |- S <: T ->
                           setter T p ([G] --> (type_nv G)) ->
                           Delta0 |- R' <: R ->
                           exists G',
                             setter S p ([G'] --> (type_nv G')) /\
                             bound_delta Delta0 S S /\
                             Delta0 |- (type_nv G') <: (type_nv G).
Admitted.

(*
 * After a reduction step, nu' and sigma' 
 * - Might be extended with new location, but they definitely don't involve the typing of e2
 *   since it is typable in nu,sigma that does not contain such locations
 * - Might be updated. Since it is always updated to preserve the well formedness, the type
 *   of the locations can only be specialized (it can be seen by a simple case analysis).
 *   That implies that the type of e2 under the environments nu' and sigma' is specialized
 *   repect to nu, sigma.
 *)
Lemma weakening_setter : forall e1 e1' e2 nu nu' sigma sigma' T,
                    well_formed_store nu sigma ->
                    sigma |--S nu ->
                    << nu, sigma, e1 >> ---> << nu', sigma', e1' >> ->
                    nu ; sigma ; Delta0 ; Gamma0 |- e2 : T ->
                    (exists S,
                      nu' ; sigma' ; Delta0 ; Gamma0 |- e2 : S /\
                      Delta0 |- S <: T).
Admitted.

(* 
 * Same considerations as above.
 *)
Lemma weakening_list : forall eL eL' e nu nu' sigma sigma' GL HL T RL,
                    well_formed_store nu sigma ->
                    sigma |--S nu ->
                    [[ nu, sigma, eL ]] --->^ [[ nu', sigma', eL' ]] ->
                    nu ; sigma ; Delta0 ; Gamma0 |- e : type_nv (GL --> T) ->
                    nu ; sigma ; Delta0 ; Gamma0 |-^ eL : HL ->
                    nu' ; sigma' ; Delta0 ; Gamma0 |-^ eL' : RL ->
                    nu' ; sigma' ; Delta0 ; Gamma0 |- e : type_nv (GL --> T).
Admitted.

(* 
 * Same considerations as above.
 *)
Lemma weakening_list_base : forall e e' eL nu nu' sigma sigma' TL T T' GL GL',
                    well_formed_store nu sigma ->
                    sigma |--S nu ->
                    << nu, sigma, e >> ---> << nu', sigma', e' >> ->
                    nu ; sigma ; Delta0 ; Gamma0 |-^ eL : TL ->
                    nu ; sigma ; Delta0 ; Gamma0 |- e : (type_nv (GL --> T)) ->
                    nu' ; sigma' ; Delta0 ; Gamma0 |- e' : (type_nv (GL' --> T')) ->
                    (exists SL,
                      nu' ; sigma' ; Delta0 ; Gamma0 |-^ eL : SL /\
                      Delta0 |-^ SL <: TL).
Admitted.

(* 
 * The reduction step does not remove any location from the nu and sigma environments, it
 * can be observed by simply case analysis. Most of the times the environments are modified
 * with the insert and update operations, so that we get can easilly get such guarantee.
 *)
Lemma subst_not_remove : forall tau G l nu nu' sigma sigma' e e',
                      (tau, (G, l)) __in nu ->
                      << nu, sigma, e >> ---> << nu', sigma', e'>> ->
                      exists l', (tau, (G, l')) __in nu'.
Admitted.

(* 
 * Just some Coq machinery.
 *)
Lemma bound_delta0_typing : forall l nu sigma T,
                nu ; sigma ; Delta0 ; Gamma0 |- (exp_location l) : T ->
                bound_delta Delta0 T T.
Proof.
  intros.
  sinversion H. destruct o;simpl. admit. admit. admit.
Qed.

(* 
 * If S <: function_type, and S is fully annotated, then S is a function
 * type too. This can be easilly seen analyzing the subtyping rule.
 *)
Lemma function_subtyping : forall GL T S Delta,
                             Delta |- S <: (type_nv (GL --> T)) ->
                             fully_annotated_type S ->
                             exists GL' T', S = (type_nv (GL' --> T')).
Admitted.

(* 
 * The evaluation of a function expression preserves the type.
 *)
Lemma function_reduction_preserve_type : forall GL T S nu sigma nu' sigma' e e',
                                  nu ; sigma ; Delta0 ; Gamma0 |- e : (type_nv (GL --> T)) ->
                                  << nu, sigma, e >> ---> << nu', sigma', e' >> ->
                                  nu' ; sigma'; Delta0 ; Gamma0 |- e' : S ->
                                  S = type_nv (GL --> T).
Admitted.

(* 
 * The reduction step does not remove any location from the nu and sigma environments, it
 * can be observed by simply case analysis. Most of the times the environments are modified
 * with the insert and update operations, so that we get can easilly get such guarantee.
 *)
Lemma expression_list_left : forall e eL e' nu sigma nu' sigma' T TL,
                               nu ; sigma ; Delta0 ; Gamma0 |- e : T ->
                               nu ; sigma ; Delta0 ; Gamma0 |-^ eL : TL ->
                               << nu, sigma, e >> ---> << nu', sigma', e' >> ->
                               (exists SL,
                                  nu' ; sigma' ; Delta0 ; Gamma0 |-^ eL : SL /\
                                  Delta0 |-^ SL <: TL).
Admitted.

(* 
 * The reduction step does not remove any location from the nu and sigma environments, it
 * can be observed by simply case analysis. Most of the times the environments are modified
 * with the insert and update operations, so that we get can easilly get such guarantee.
 *)
Lemma expression_list_right : forall e eL eL' nu sigma nu' sigma' T TL,
                               nu ; sigma ; Delta0 ; Gamma0 |- e : T ->
                               nu ; sigma ; Delta0 ; Gamma0 |-^ eL : TL ->
                               [[ nu, sigma, eL ]] --->^ [[ nu', sigma', eL' ]] ->
                               (exists S,
                                  nu' ; sigma' ; Delta0 ; Gamma0 |- e : S /\
                                  Delta0 |- S <: T).
Admitted.

(* 
 * If l : T, l points to the object o, typeof o = typeof l <: S, considering
 * exp_frame_expression S (exp_location l), than T is subtype of S.
 * This is trivial, since it simply means that if typeof o <: S, then
 * also T <: S, where T is indeed the type of o (which coincides with the
 * type of l).
 *)
Lemma subtype_objects : forall l o nu sigma T S,
  (l, o) __in sigma ->
  nu; sigma; Delta0; Gamma0 |- exp_location l : T ->
  Delta0 |S- typeof o <: S ->
  nu; sigma; Delta0; Gamma0 |- exp_frame_expression S (exp_location l) : S ->
  Delta0 |- T <: S.
Admitted.

(* TYPE SOUNDNESS *)

(* 
 * Error configurations cannot take a further step.
 * This can be proven by simly analysing the error configuration
 * definition and using the operational semantics rules.
 *)
Lemma error_irreducible : forall nu sigma e,
                            well_formed_store nu sigma ->
                            lenv_base_superset sigma ->
                            error nu sigma e ->
                            irred nu sigma e.
Proof.
  intros. induction H1 using error_mut with
                    (P := fun nu sigma e _ =>
                            well_formed_store nu sigma ->
                            lenv_base_superset sigma ->
                            irred nu sigma e)
                    (P0 := fun nu sigma eL _ =>
                             well_formed_store nu sigma ->
                             lenv_base_superset sigma ->
                             forall e, e _in eL -> irred nu sigma e);
          unfold irred; intros; intro F.
  - sinversion F. unfold lenv_base_superset in H. unfold sigma0 in H.
    assert((loc_null, empty_location)
      __in ([(loc_null, empty_location)], first_fresh)). constructor. auto.
    assert(X := H loc_null empty_location).
Admitted.

(* 
 * If l : T with the type system without assignability on premises, then
 * l : T with the type system with assignability on premises too.
 *)
Lemma location_typing_equivalence : forall nu sigma Delta Gamma l T,
                                      nu ; sigma ; Delta ; Gamma |- (exp_location l) : T <->
                                      nu ; sigma ; Delta ; Gamma |-STD (exp_location l) : T.
Proof.
  intros. split;intros.
  - sinversion H. econstructor. eauto. eauto.
  - sinversion H. econstructor. eauto. eauto.
Qed.

(* 
 * The relaxed type system, i.e., the type system without assignability on
 * premises, is less restrictive than the standard type system, i.e, the
 * type system with assiganbility on premises.
 *)
Lemma relaxed_ts_subset_standard_ts : forall nu sigma Delta Gamma e T,
                                        nu ; sigma ; Delta ; Gamma |-STD e : T ->
                                        nu ; sigma ; Delta ; Gamma |- e : T.
Proof.
  intros nu sigma Delta Gamma e T STD_ETyping.
  induction STD_ETyping using exp_typing_standard_mut with
  (P := fun nu sigma Delta Gamma e T _ => nu; sigma; Delta; Gamma |- e : T)
  (P0 := fun nu sigma Delta Gamma eL TL _ =>  nu; sigma; Delta; Gamma |-^ eL : TL).
  - auto.
  - econstructor. eauto. eauto. eauto.
  - econstructor. eauto. eauto. eauto. eauto.
  - econstructor. eauto. eauto.
  - econstructor. eauto. eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
Qed.

End Lemmas.
