(* 
Fletch formalism and message-safety proof
Version: 1.7
Date: 17/06/2016
*)

Require Export fletch_lemmas.
(* Uncomment the following 2 lines if it does not compile *)
(* Require String. Open Scope string_scope. *)
(* Require Import Coq.Program.Equality. *)

Module Properties (C : ClassTable).
Module LemmasImpl := Lemmas(C).
Export LemmasImpl.

(* 
 * The following file contains the message safety theorem and proof. It divides into the following major lemmas
 * - Soundness of the message-safe type system with assignability on premises, which relies on
 * - Soundness of the message-safe type system without assignability on premises, which relies on
 * - Progress and Preservation lemmas for the type system without assignability on premises.
 * Note that all these theorems rely on additiona lemmas. All of them are in "fletch_lemmas.v"
 *)

(* PROGRESS *)

Lemma progress :
  forall sigma nu e T,
    (* IF *)
    (* e is fully annotated *)
    fully_annotated_expression e ->
    (* nu is fully annotated *)
    fully_annotated_subst nu ->
    (* sigma is fully annotated *)
    fully_annotated_store sigma ->
    (* e is well typed with type T *)
    nu ; sigma ; Delta0 ; Gamma0 |- e : T ->
    (* All the environments are well formed *)
    well_formed_envs nu sigma Gamma0 ->
    well_formed_store nu sigma ->
    sigma |--S nu ->
    (* THEN one of the three following statements hold *)
    value e \/ (* a) e is a value *)
    (exists nu' sigma' e',
      << nu, sigma, e >> ---> << nu', sigma', e' >>) \/ (* b) e takes a step *)
      error nu sigma e. (* c) e is a stuck configuration because it raises a subtype violtion or null pointer error *)
Proof.
  (* By induction in the structure of the typing derivation nu ; sigma ; Delta0 ; Gamma0 |- e : T *)
  intros sigma nu e T FA_e FA_nu FA_sigma Exp_typing WFEnvs WFStore WFSigmaNu.
  remember Delta0 as Delta eqn: EQG_Delta; revert EQG_Delta.
  remember Gamma0 as Gamma eqn: EQG_Gamma; revert EQG_Gamma.
  induction Exp_typing using exp_typing_mut with
  (P := fun nu sigma Delta Gamma e T _ => fully_annotated_expression e ->
                                          fully_annotated_subst nu ->
                                          fully_annotated_store sigma ->
                                          well_formed_envs nu sigma Gamma ->
                                          well_formed_store nu sigma ->
                                          sigma |--S nu ->
                                          Gamma = Gamma0 ->
                                          Delta = Delta0 ->
                                          value e \/
                                          (exists (nu' : fletch_rvenv) (sigma' : fletch_lenv) 
                                                  (e' : fletch_exp),
                                             << nu, sigma, e >> ---> << nu', sigma', e' >>) \/ 
                                             error nu sigma e)
   (P0 := fun nu sigma Delta Gamma eL TL _ => fully_annotated_expression_list eL -> 
                                              fully_annotated_subst nu ->
                                              fully_annotated_store sigma ->
                                              well_formed_envs nu sigma Gamma ->
                                              well_formed_store nu sigma ->
                                              sigma |--S nu ->
                                              Gamma = Gamma0 ->
                                              Delta = Delta0 ->
                                              values eL \/
                                              (exists (nu' : fletch_rvenv)
                                                      (sigma' : fletch_lenv) 
                                                      (eL' : list fletch_exp),
                                              [[ nu, sigma, eL ]] --->^
                                              [[ nu', sigma', eL' ]]) \/
                                              error_list nu sigma eL);
    intros; subst Gamma; assert (Trivial_Gamma : Gamma0 = Gamma0);
    assert (Trivial_Delta : Delta0 = Delta0);trivial;trivial.

    - Case "y".
     right. assert(EM := var_compare y var_null). destruct EM as [ YNull | YNNull ].
     * left. exists nu,sigma,(exp_location loc_null). subst. auto.
     * left. exists nu, sigma, (exp_location loc_null). destruct i.
       sinversion H. intuition. intuition. sinversion H.
    - Case "e0.p".
      right. sinversion FA_e. assert(BDN := bound_delta_neutral T T' b). subst. 
      assert(IHe0 := IHExp_typing H1 FA_nu FA_sigma WFEnvs WFStore
                                             WFSigmaNu Trivial_Gamma Trivial_Delta).
      destruct IHe0 as [ Value_e0 | Reduces_e0 ].
      vdestruct Value_e0 l Value.
      * subst. wfdestruct WFEnvs.
        assert(nu; sigma; Delta0; Gamma0 |- exp_getter (exp_location l) p
           : type_nv G). econstructor. apply Exp_typing. apply b. apply g.
        assert(Getter := getter_base_reduction l p nu sigma Gamma0 (type_nv G) 
                                          WFStore WFLenv FA_sigma H).
        destruct Getter as [ l' Reduction ]. left.
        exists nu, sigma, (exp_location l'). auto.
      * destruct Reduces_e0 as [ [ nu' [sigma' [e0' Reduces_e0 ] ] ] | Err ]. left.
      exists nu', sigma', (exp_getter e0' p). constructor. auto. right. constructor. auto.
    - Case "e1.p = e2".
        right. sinversion FA_e.
        assert(IHe1 := IHExp_typing1 H2 FA_nu FA_sigma WFEnvs
                                     WFStore WFSigmaNu Trivial_Gamma Trivial_Delta).
        assert(IHe2 := IHExp_typing2 H4 FA_nu FA_sigma WFEnvs
                                     WFStore WFSigmaNu Trivial_Gamma Trivial_Delta).
        destruct IHe1 as [ Value_e0 | Reduces_e0 ];
        destruct IHe2 as [ Value_e1 | Reduces_e1 ].
      * vdestruct Value_e0 l1 Value_e0. vdestruct Value_e1 l2 Value_e1. subst.
        wfdestruct WFEnvs.
        assert(nu; sigma; Delta0; Gamma0 |-
                              exp_setter (exp_location l1) p (exp_location l2) : R).
        econstructor. eauto. apply b. apply s. apply Exp_typing2.
        assert(Setter := setter_base_reduction l1 l2 p nu sigma Gamma0 R
              WFStore WFLenv FA_sigma H).
        destruct Setter as [ [ sigma' Reduction ] | CheckmodeError ]. left.
        exists nu, sigma', (exp_location l2). auto. right. eauto.
      * destruct Reduces_e1 as [ [ nu' [sigma' e' ] ] | CheckmodeError ].
        vdestruct Value_e0 l Val. subst. destruct e' as [ e' Ex ]. 
        left. exists nu', sigma', (exp_setter (exp_location l) p e'). auto. right.
        apply Ind_Err_setter_right. auto.
      * destruct Reduces_e0 as [ [nu' [sigma' [e' Reduction] ] ] | CheckmodeError ]. 
        destruct Value_e1 as [ l Eq ]. subst. left.
        exists nu', sigma', (exp_setter e' p (exp_location l)).
        constructor. auto. right. constructor. auto.
      * destruct Reduces_e0 as [ [ nu' [sigma' [e' Reduction] ] ] | e0_CheckmodeError ];
        destruct Reduces_e1 as [ [nu'' [sigma'' [e'' Reduction'] ] ] | e1_CheckmodeError ].
        left. exists nu', sigma', (exp_setter e' p e1). constructor. auto. right.
        apply Ind_Err_setter_right. auto. right. apply Ind_Err_setter_left. auto. right. 
        apply Ind_Err_setter_left. auto.
     - Case "x = e0".
        right. inversion i. inversion H. inversion H.
     - Case "e0(el)".
        right. sinversion FA_e.
        assert(IHe := IHExp_typing H2 FA_nu FA_sigma WFEnvs WFStore WFSigmaNu
                      Trivial_Gamma Trivial_Delta).
        assert(IHe0 := IHExp_typing0 H3 FA_nu FA_sigma WFEnvs WFStore WFSigmaNu
                                     Trivial_Gamma Trivial_Delta).
        destruct IHe as [ Value_e0 | Reduces_e0 ];
        destruct IHe0 as [ Value_list | Reduces_list ].
     * vdestruct Value_e0 l V. subst.
       assert(nu; sigma; Delta0; Gamma0 |- exp_call (exp_location l) el : T).
       econstructor. apply Exp_typing. apply e.
       wfdestruct WFEnvs.
       assert(Call := call_base_reduction l nu sigma Gamma0 T el WFStore
             WFLenv FA_sigma H Value_list).
       destruct Call as [ [ l' [ e' [ nu' CallReduction ] ] ] | CheckmodeError ]. left.
       exists nu', sigma, (exp_frame_expression l' e'). auto. right. eauto.
     * destruct Reduces_list as [ [ nu' [ sigma' [ eL' List_Reduction ] ] ] | List_Error ].
       left. exists nu', sigma', (exp_call e0 eL'). vdestruct Value_e0 l ValEq. subst.
       constructor. auto. right. apply Ind_Err_call_args. auto.
     * destruct Reduces_e0 as [ [ nu' [ sigma' [ e0' Red_e0 ] ] ] | Err_e0 ]. left.
       exists nu', sigma', (exp_call e0' el). auto. right. apply Ind_Err_call_function. auto.
     * destruct Reduces_e0 as [ [ nu' [ sigma' [ e' Err ] ] ] | Error ]. left.
       exists nu', sigma', (exp_call e' el). auto. right. apply Ind_Err_call_function. auto.
     - Case "new N()".
       right. wfdestruct WFEnvs.
       assert(nu; sigma; Delta0; Gamma0 |- exp_new N : (type_nv (type_classtype N))).
       constructor. subst. apply t. subst.
       assert(Constructor := constructor_base_reduction nu sigma Gamma0 Delta0 N
              WFStore WFLenv FA_sigma H).
       destruct Constructor as [ l [ nu' [ sigma' Red ] ] ]. left.
       exists nu', sigma', (exp_location l). auto.
     - Case "(G x list) => e".
       assert(nu ; sigma ; Delta0 ; Gamma0 |- (exp_function GxL T e0) : (type_nv ((GL) --> T))).
       econstructor. apply e. rewrite -> H0 in t.
       apply t. intros. apply e1. apply f. rewrite -> H0 in Exp_typing. apply Exp_typing.
       right. wfdestruct WFEnvs. subst.
       assert(Fun := function_base_reduction nu sigma Gamma0 Delta0 e0
                            T GxL (projection GxL) WFStore WFLenv FA_sigma H).
       destruct Fun as [ l [ sigma' Red ] ]. left. exists nu, sigma', (exp_location l). auto.
     - Case "l".
       left. exists l. trivial.
     - Case "tau".
       right. left. exists nu, sigma, (exp_location l). econstructor. apply b.
     - Case "tau = e".
       right. sinversion FA_e.
       assert(Gamma0 = Gamma0). auto. assert(Delta0 = Delta0). auto.
       assert(IHe := IHExp_typing H1 FA_nu FA_sigma WFEnvs WFStore
                     WFSigmaNu H H0).
       destruct IHe as [ Value | Reduction ].
     * vdestruct Value l' Val. subst.
       wfdestruct WFEnvs.
       assert(nu; sigma; Delta0; Gamma0 |- exp_runtime_assignment tau (exp_location l') : T).
       econstructor. apply Exp_typing. apply b.
       assert(RuntimeAssignment := runtime_assignment_base_reduction nu sigma
                                   tau Gamma0 Delta0 T l' WFStore WFLenv FA_nu H2).
       destruct RuntimeAssignment as [ nu' Red ]. left.  exists nu', sigma, (exp_location l').
       auto.
     * destruct Reduction as [ [ nu' [ sigma' [ e' Red ] ] ] | CheckmodeError ]. left.
       exists nu', sigma', (exp_runtime_assignment tau e'). auto. right. constructor. auto.
     - Case "[T,e]". sinversion FA_e. sinversion FA_e. right.
       assert(IHe := IHExp_typing H3 FA_nu FA_sigma WFEnvs WFStore WFSigmaNu
                     Trivial_Gamma Trivial_Delta). destruct IHe as [ Val | Red ].
     * vdestruct Val l V. subst.
       assert(nu; sigma; Delta0; Gamma0
          |- exp_frame_expression T (exp_location l) : T).
       econstructor. apply Exp_typing. auto.
       assert(Frame := frame_expression_base_reduction nu sigma T T l Gamma0 H).
       destruct Frame as [ Red | Err ]. left. exists nu, sigma, (exp_location l).
       auto. right. auto.
     * destruct Red as [ [ nu' [ sigma' [ e' Red ] ] ] | CheckmodeError ]. left.
       exists nu', sigma', (exp_frame_expression T e'). constructor. auto. right. constructor.
       auto.
     - Case "[]".
       left. constructor.
     - Case "[el]".
       inversion FA_e.
       assert(IHe := IHExp_typing H2 FA_nu FA_sigma WFEnvs WFStore WFSigmaNu
                     Trivial_Gamma H0).
       assert(IHeL := IHExp_typing0 H3 FA_nu FA_sigma WFEnvs WFStore WFSigmaNu
                      Trivial_Gamma H0).
       destruct IHe as [ Value_e | Reduces_e ];
       destruct IHeL as [ Value_eL | Reduces_eL ].
     * vdestruct Value_e l Val_e. subst. left. constructor. exists l. auto.
       apply Value_eL.
     * destruct Reduces_eL as [ [ nu' [ sigma' [ eL' RedL ] ] ] | ErrL ].
       right. vdestruct Value_e l Val_e. subst. left.
       exists nu', sigma', (exp_location l :: eL'). constructor. auto. subst. right. right.
       constructor 2. auto.
     * destruct Reduces_e as [ [ nu' [ sigma' [ e' Red ] ] ] | Err ]. right. left.
       exists nu', sigma', (e' :: el). auto. right. right. constructor. auto.
     * destruct Reduces_e as [ [ nu' [ sigma' [ e' Red ] ] ] | Err ]. right. left.
       exists nu', sigma', (e' :: el). constructor. auto. right. right. constructor. auto.
Qed.

(* PRESERVATION *)

Lemma preservation :
  forall e e' sigma sigma' nu nu' T,
    (* IF *)
    (* e is fully annotated *)
    fully_annotated_expression e ->
    (* nu is fully annotated *)
    fully_annotated_subst nu ->
    (* sigma is fully annotated *)
    fully_annotated_store sigma ->
    (* e is well typed *)
    nu ; sigma ; Delta0 ; Gamma0 |- e : T ->
    (* the environments are well formed *)                                        
    well_formed_envs nu sigma Gamma0 ->
    well_formed_store nu sigma ->
    sigma |--S nu ->
    (* e takes a step into e' *)
    << nu, sigma, e >> ---> << nu', sigma', e'>> ->
    lenv_base_superset sigma ->
    (* THEN both of the following two points 1 and 2 hold (2a holds if either 2a or 2b hold) *)
    (error nu' sigma' e' \/ (* 2a) e' is a stuck configuration because of a subtype violation or null pointer error *)
     (exists S, (* 2b) e' well typed and preserves the type *)       
       nu' ; sigma' ; Delta0 ; Gamma0 |- e' : S /\
       Delta0 |- S <: T)) /\                          
    (* 1. e' is fully annoated, all the environemnts are well formed and fully annoated after the execution step *)
    fully_annotated_expression e' /\
    fully_annotated_subst nu' /\
    fully_annotated_store sigma' /\
    well_formed_store nu' sigma' /\       
    sigma' |--S nu' /\
    well_formed_envs nu' sigma' Gamma0 /\
    lenv_base_superset sigma'.
Proof.
  (* By induction in the execution derivation <nu, sigma, e> -> <nu', sigma', e'> *)
  intros e e' sigma sigma' nu nu' T FA_e FA_nu FA_sigma Exp_Typing WFEnvs
         WFsigma WFsigma_nu Exp_Reduction LenvBaseSuperset.
  generalize dependent T.

  induction Exp_Reduction using exp_eval_mut with
    (P := fun nu sigma e nu' sigma' e' _ =>
    fully_annotated_expression e ->
    fully_annotated_subst nu ->
    fully_annotated_store sigma ->
    well_formed_envs nu sigma Gamma0 ->
    well_formed_store nu sigma ->
    sigma |--S nu ->
    lenv_base_superset sigma ->
    forall T : fletch_type,
    nu; sigma; Delta0; Gamma0 |- e : T ->
    (error nu' sigma' e' \/
    exists S : fletch_type,      
      nu'; sigma'; Delta0; Gamma0 |- e' : S /\
      Delta0 |- S <: T) /\
    (fully_annotated_expression e' /\
      fully_annotated_subst nu' /\
      fully_annotated_store sigma' /\     
      well_formed_store nu' sigma' /\
      sigma' |--S nu' /\
      well_formed_envs nu' sigma' Gamma0 /\ lenv_base_superset sigma'))
    (P0 := fun nu sigma eL nu' sigma' eL' _ =>
    fully_annotated_expression_list eL ->
    fully_annotated_subst nu ->
    fully_annotated_store sigma ->
    well_formed_envs nu sigma Gamma0 ->
    well_formed_store nu sigma ->
    sigma |--S nu ->
    lenv_base_superset sigma ->
    forall TL : list fletch_type,
    nu; sigma; Delta0; Gamma0 |-^ eL : TL ->
    (error_list nu' sigma' eL' \/
     (exists SL : list fletch_type,
       nu'; sigma'; Delta0; Gamma0 |-^ eL' : SL /\
       Delta0 |-^ SL <: TL)) /\
    (fully_annotated_expression_list eL' /\
      fully_annotated_subst nu' /\
      fully_annotated_store sigma' /\
      well_formed_store nu' sigma' /\
      sigma' |--S nu' /\
      well_formed_envs nu' sigma' Gamma0 /\ lenv_base_superset sigma')
    );intros R Exp_typing;
      intros;subst.

  - Case "tau -> l".
    sinversion Exp_typing.
    assert(WF := WFsigma_nu tau G l b). wfdestruct WFEnvs. destruct WFRvenv.
    assert(WF' := NoDupKeysLemma tau (G, l) (G0, l0) wenv H b H4). 
    destruct WF as [ o [ LDef Subt ] ]. sinversion WF'. split. right.
    exists (typeof o). split. eauto. auto. split. constructor. split. auto. split. auto.
    split. auto. split. auto. split. auto. split. constructor. auto. auto. auto.
  - Case "null -> l_null".
    wfdestruct WFEnvs. sinversion Exp_typing. sinversion H4. sinversion H. split. right.
    exists (type_nv type_bottom). split. econstructor.
    assert((loc_null, empty_location) __in sigma0). constructor. auto.
    assert(X := LenvBaseSuperset loc_null empty_location H0). apply X.
    split. auto. split. auto. constructor. split. auto. split. auto. split. auto. split. auto.
    split. auto. split. auto. auto. auto. auto. inversion H.
  - Case "tau = l -> l".
    wfdestruct WFEnvs. sinversion Exp_typing. split. right. exists R.
    split. econstructor. destruct sigma. sinversion WFLenv. 
    inversion H5. assert(NDup := NoDupKeysLemma l o o0 f H0 b0 H1). subst. eauto.
    inversion H5. destruct WFLenv.
    assert(NDup := NoDupKeysLemma l o o0 wenv H9 b0 H0). subst. trivial.
    destruct nu. sinversion WFRvenv. destruct sigma.
    assert(NDup' := NoDupKeysLemma tau (G, l0) (G0, l1) f H0 b H7). sinversion NDup'. 
    destruct WFLenv.
    inversion H5. assert(NDup := NoDupKeysLemma l o o0 wenv H b0 H2). subst.
    eauto using subtyping_reflexivity'. split. constructor. split.
    assert(X := FA_nu tau l0 G b).
    assert(Y := rvenv_update_preserves_fully_annotated tau nu nu' G l l0 FA_nu WFRvenv
                b X u). auto. split. auto. split.
    assert(X := rvenv_update_preserves_well_formed_store nu nu' tau sigma G l0 l
                                                         FA_sigma FA_nu WFsigma b u). auto.
    split.
    assert(X := rvenv_update_preserves_WFSigmaRho tau sigma nu G l l0 nu' o WFsigma_nu
                b0 s b u). auto. split.
    assert(X := rvenv_update_preserves_well_formed_store nu nu' tau sigma G l0 l FA_sigma
                FA_nu WFsigma b u).
    assert(Y := update_rvenv_well_formed tau (G, l) nu nu' WFRvenv u). split. auto.
    split. auto. auto. auto.
  - Case "new C<GL> -> l_base".
    sinversion u. sinversion u.
    emdestruct Object NotObject (location_em C builtin_class_object).
    (* destruct o0 as [ NotObject | Object ]. *)
    * SCase "C = Object".
      assert(X := update_correctness l_base (class_instance_object (atype_class C GL) RFL RML)
                env env' H4). split. right.
      exists (type_nv (type_classtype (atype_class C GL))). split. econstructor. apply X.
      simpl. trivial. sinversion Exp_typing. apply subtyping_reflexivity'. split.
      wfdestruct WFEnvs. sinversion FA_e. econstructor. split. wfdestruct WFEnvs. simpl.
      sinversion FA_e. assert(fully_annotated_type (typeof empty_location)). constructor.
      assert(Y := rvenv_insert_preserves_fully_annotated v_base nu nu'
                 (atype_classtype (atype_class builtin_class_object GL)) 
                 l_base FA_nu WFRvenv H1 i0). auto. split.
      wfdestruct WFEnvs. assert(fully_annotated_type (typeof empty_location)). constructor.
      assert(FA_sigma' := lenv_insert_preserves_fully_annotated sigma sigma' l_base
                        empty_location FA_sigma WFLenv H0 i).
      assert(FAMethods : fully_annotated_methods ML).
      subst. unfold fully_annotated_methods. intros. sinversion g0. sinversion H1.
      assert(WFCT := well_formed_CT). destruct WFCT as [ _ [ NOOBJ _ ] ].
      assert(Contradiction := NOOBJ (class builtin_class_object PL N FL0 ML0)). contradiction.
      assert(FASigma'' := bind_methods_preserves_fully_annotated sigma' (env, fresh)
                        v_base (atype_class C GL) ML RML FA_sigma' FAMethods b0).
      assert(WFLenv' := insert_lenv_well_formed empty_location sigma sigma' l_base
                               WFLenv i).
      assert(WFLenv'' := bind_methods_lenv_WF sigma' (env, fresh) v_base ML (atype_class C GL)
                         RML WFLenv' b0).
      assert(WFLenv''' := update_lenv_well_formed l_base
                          (class_instance_object (atype_class C GL) RFL RML)
                          (env, fresh) (env', fresh) WFLenv'' u). sinversion FA_e.
      assert(FASigma''' := lenv_update_preserves_fully_annotated (env, fresh)
                           (env', fresh) l_base
                           (class_instance_object (atype_class builtin_class_object GL) 
                                                  RFL RML) FASigma'' WFLenv'' H2 u). auto.
      split. assert(WFEmpty : well_formed_object nu sigma empty_location). auto.
      assert(WF := lenv_insert_preserves_WF nu sigma sigma' l_base
                   empty_location WFsigma WFEmpty i).
      assert(WF' := bind_methods_preserves_WF nu sigma' (env, fresh) v_base
                    ML (atype_class C GL) RML WF b0).
      assert(WF'' := lenv_update_preserves_WF nu (env, fresh) (env', fresh) l_base
                     (class_instance_object (atype_class C GL) RFL RML) WF' u). sinversion i0.
      assert(UpdateCorrectness := update_fresh_correctness
                                  l_base
                                  (class_instance_object (atype_class builtin_class_object GL) 
                                                         RFL RML)
                                  (env, fresh) (env', fresh) u).
      assert(fully_annotated_type (typeof empty_location)). constructor. wfdestruct WFEnvs.
      assert(FA_sigma' := lenv_insert_preserves_fully_annotated sigma
                        sigma' l_base
                        empty_location FA_sigma WFLenv H0 i).
      assert(WFLenv' := insert_lenv_well_formed empty_location sigma
                        sigma' l_base WFLenv
                        i).
      assert(WFLenv'' := bind_methods_lenv_WF sigma'
                         (env, fresh) (generate_fresh_variable_location k)
                         ML (atype_class builtin_class_object GL) RML WFLenv' b0).
      assert(WFLenv''' := update_lenv_well_formed l_base
                          (class_instance_object (atype_class builtin_class_object GL) RFL RML)
                          (env, fresh) (env', fresh) WFLenv'' u).
      sinversion FA_e.
      assert(FAMethods : fully_annotated_methods ML). 
      assert(WFCT := well_formed_CT). destruct WFCT as [ _ [ NOOBJ _ ] ].
      subst. unfold fully_annotated_methods. intros. sinversion g0. sinversion H1.
      assert(Contradiction := NOOBJ (class builtin_class_object PL N FL0 ML0)). contradiction.
      assert(FASigma'' := bind_methods_preserves_fully_annotated
                          sigma' (env, fresh)
                          (generate_fresh_variable_location k)
                          (atype_class builtin_class_object GL) ML RML FA_sigma' FAMethods b0).
      assert(FA : fully_annotated_type
                 (typeof (class_instance_object (atype_class builtin_class_object GL) 
                                                RFL RML))). simpl. eauto.
      assert(FASigma''' := lenv_update_preserves_fully_annotated (env,fresh) (env', fresh)
                          l_base
                          (class_instance_object (atype_class builtin_class_object GL) 
                                                 RFL RML) FASigma''
                          WFLenv'' FA u).
      assert(WFO := WF'' l_base
                         (class_instance_object (atype_class builtin_class_object GL) RFL RML)
                         UpdateCorrectness).
      assert(S : Delta0
            |- typeof (class_instance_object (atype_class builtin_class_object GL) RFL RML) <:
            type_nv (atype_classtype (atype_class builtin_class_object GL))). simpl.
      eauto using subtyping_reflexivity'.
      assert(WF''' := rvenv_insert_preserves_well_formed_store (env0, k)
                      (env'0, generate_fresh_variable_location k)
                      (env', fresh)
                      (atype_classtype (atype_class builtin_class_object GL)) l_base
                      (class_instance_object (atype_class builtin_class_object GL) RFL RML)
                      (generate_fresh_variable_location k)
                      FASigma''' FA_nu WF'' UpdateCorrectness WFO S i0). auto.
      split. sinversion i0.
      destruct sigma, sigma'. sinversion i.
      assert(LBase := insert_correctness (generate_fresh_location g1) empty_location f f0 H5).
      assert(EmptyST : Delta0 |- typeof empty_location <:
                       type_nv (atype_classtype (atype_class builtin_class_object GL))). eauto.
      assert(WFS'N := lenv_insert_preserves_WFSigmaRho (f, g1)
                      (f0, (generate_fresh_location g1)) (env0, k) (generate_fresh_location g1)
                      empty_location WFsigma_nu i).
      assert(EmptyST' := t_subtyping_relaxed Delta0 (typeof empty_location)
                         (type_nv (atype_classtype (atype_class builtin_class_object GL)))
                         EmptyST).
      assert(WFS'N' := rvenv_insert_preserves_WFSigmaRho (f0, generate_fresh_location g1)
                       (env0, k)
                       (atype_classtype (atype_class builtin_class_object GL)) 
                       (generate_fresh_location g1)
                       (env'0, generate_fresh_variable_location k)
                       empty_location (generate_fresh_variable_location k)
                       WFS'N LBase EmptyST' i0).
      assert(WFS''N' := lenv_bind_preserves_WFSigmaRho (f0, generate_fresh_location g1)
                        (env, fresh) (env'0, generate_fresh_variable_location k)
                        (generate_fresh_variable_location k) 
                        (atype_class builtin_class_object GL) ML RML
                        WFS'N' b0).
      assert(InsAd := bind_methods_add (f0, generate_fresh_location g1) (env, fresh)
                      (generate_fresh_variable_location k) ML RML (generate_fresh_location g1)
                      empty_location (atype_class builtin_class_object GL) LBase b0).
      assert(WFS'''N' := lenv_update_preserves_WFSigmaRho (env, fresh) (env', fresh)
                         (env'0, generate_fresh_variable_location k)
                         (generate_fresh_location g1)
                         empty_location
                         (class_instance_object (atype_class builtin_class_object GL) RFL RML)
                        WFS''N' InsAd u). auto. split. wfdestruct WFEnvs. sinversion i0.
      assert(WFN' := rvenv_insert_preserves_WF (env0, k)
                     (env'0, generate_fresh_variable_location k)
                     (generate_fresh_variable_location k)
                     (atype_classtype (atype_class builtin_class_object GL))
                     l_base WFRvenv i0).
      assert(WFStore' := insert_lenv_well_formed empty_location sigma sigma' l_base WFLenv
                         i).
      assert(WFStore'' := bind_methods_lenv_WF sigma' (env, fresh)
                                               (generate_fresh_variable_location k) ML 
                                               (atype_class builtin_class_object GL) RML
                          WFStore' b0).
      assert(WFStore''' := update_lenv_well_formed l_base
                           (class_instance_object (atype_class builtin_class_object GL) 
                                                  RFL RML)). split.
      auto. split. auto. auto. eauto. eauto.
      sinversion i0.
      assert(LenvBaseSuperset' := lenv_insert_preserves_sigma0 sigma sigma' l_base
                                  empty_location LenvBaseSuperset i).
      assert(LenvBaseSuperset'' := bind_methods_preserves_sigma0 sigma' (env, fresh)
                                   (generate_fresh_variable_location k) ML 
                                   (atype_class builtin_class_object GL)
                                   RML LenvBaseSuperset' b0).
      assert(LenvBaseSuperset''' := lenv_update_preserves_sigma0 (env, fresh) (env', fresh)
                                    l_base (class_instance_object 
                                              (atype_class builtin_class_object GL) RFL RML)
                                    LenvBaseSuperset'' u). auto.
    * SCase "C <> Object".
    assert(X := update_correctness l_base (class_instance_object (atype_class C GL) RFL RML)
                env env' H4). split. right.
    exists (type_nv (type_classtype (atype_class C GL))). split. econstructor. apply X.
    simpl. trivial. sinversion Exp_typing. apply subtyping_reflexivity'. split.
    wfdestruct WFEnvs. sinversion FA_e. econstructor. split. wfdestruct WFEnvs. simpl.
    sinversion FA_e. assert(fully_annotated_type (typeof empty_location)).
    constructor.
    assert(Y := rvenv_insert_preserves_fully_annotated v_base nu nu'
                (atype_classtype (atype_class C GL)) l_base FA_nu WFRvenv H1 i0). auto. split.
    wfdestruct WFEnvs. assert(fully_annotated_type (typeof empty_location)). constructor.
    assert(FA_sigma' := lenv_insert_preserves_fully_annotated sigma sigma' l_base
                        empty_location FA_sigma WFLenv H0 i).
    assert(WFCT := well_formed_CT). destruct WFCT as [ NDup  [ NObj _ ] ]. sinversion g0.
    intuition.
    assert(FAML := fully_annotated_CT C PL N FL0 ML0 H3).
    destruct FAML as [ FAT [ FAP [ FAFields FAMethods ] ] ]. sinversion FA_e. sinversion H2.  
    assert(FAHierarchy := fully_annotated_hierarchy C PL N FL0 ML0 GL
                                                    (method_subst PL GL ML0 ++ ML_Super)
                                                    H3 g0 H5).
    assert(FASigma'' := bind_methods_preserves_fully_annotated sigma' (env, fresh)
                        v_base (atype_class C GL) (method_subst PL GL ML0 ++ ML_Super)
                        RML FA_sigma' FAHierarchy b0).
    assert(WFLenv' := insert_lenv_well_formed empty_location sigma sigma' l_base
                             WFLenv i).
    assert(WFLenv'' := bind_methods_lenv_WF sigma' (env, fresh) v_base 
                       (method_subst PL GL ML0 ++ ML_Super) (atype_class C GL)
                       RML WFLenv' b0).
    assert(WFLenv''' := update_lenv_well_formed l_base
                        (class_instance_object (atype_class C GL) RFL RML)
                        (env, fresh) (env', fresh) WFLenv'' u). sinversion FA_e.
    assert(FASigma''' := lenv_update_preserves_fully_annotated (env, fresh)
                         (env', fresh) l_base
                         (class_instance_object (atype_class C GL) RFL RML) FASigma''
                         WFLenv'' H2 u). auto.
    split. assert(WFEmpty : well_formed_object nu sigma empty_location). auto.
    assert(WF := lenv_insert_preserves_WF nu sigma sigma' l_base
                 empty_location WFsigma WFEmpty i).
    assert(WF' := bind_methods_preserves_WF nu sigma' (env, fresh) v_base
                  ML (atype_class C GL) RML WF b0).
    assert(WF'' := lenv_update_preserves_WF nu (env, fresh) (env', fresh) l_base
                   (class_instance_object (atype_class C GL) RFL RML) WF' u). sinversion i0.
    assert(UpdateCorrectness := update_fresh_correctness
                                l_base
                                (class_instance_object (atype_class C GL) RFL RML)
                                (env, fresh) (env', fresh) u).
    assert(fully_annotated_type (typeof empty_location)). constructor. wfdestruct WFEnvs.
    assert(FA_sigma' := lenv_insert_preserves_fully_annotated sigma
                        sigma' l_base
                        empty_location FA_sigma WFLenv H0 i).
    sinversion g. intuition.
    assert(FAML := fully_annotated_CT C PL N FL0 ML0 H5).
    destruct FAML as [ FAT [ FAP [ FAFields FAMethods ] ] ].
    assert(WFLenv' := insert_lenv_well_formed empty_location sigma
                      sigma' l_base WFLenv
                      i).
    assert(WFLenv'' := bind_methods_lenv_WF sigma'
                       (env, fresh) (generate_fresh_variable_location k)
                       ML (atype_class C GL) RML WFLenv' b0).
    assert(WFLenv''' := update_lenv_well_formed l_base
                        (class_instance_object (atype_class C GL) RFL RML)
                        (env, fresh) (env', fresh) WFLenv'' u).
    sinversion FA_e. sinversion H2.
    assert(FAHierarchy := fully_annotated_hierarchy C PL N FL0 ML0 GL ML H5
                                                    g0 H6).
    assert(FASigma'' := bind_methods_preserves_fully_annotated
                        sigma' (env, fresh)
                        (generate_fresh_variable_location k)
                        (atype_class C GL) ML RML FA_sigma' FAHierarchy b0).
    assert(FA : fully_annotated_type
               (typeof (class_instance_object (atype_class C GL) RFL RML))). simpl. eauto.
    assert(FASigma''' := lenv_update_preserves_fully_annotated (env,fresh) (env', fresh)
                        l_base
                        (class_instance_object (atype_class C GL) RFL RML) FASigma''
                        WFLenv'' FA u).
    assert(WFO := WF'' l_base
                       (class_instance_object (atype_class C GL) RFL RML)
                       UpdateCorrectness).
    assert(S : Delta0
          |- typeof (class_instance_object (atype_class C GL) RFL RML) <:
          type_nv (atype_classtype (atype_class C GL))). simpl.
    eauto using subtyping_reflexivity'.
    assert(WF''' := rvenv_insert_preserves_well_formed_store
                      (env0, k)
                      (env'0, generate_fresh_variable_location k)
                      (env', fresh)
                   (atype_classtype (atype_class C GL)) l_base
                   (class_instance_object (atype_class C GL) RFL RML)
                   (generate_fresh_variable_location k)
                   FASigma''' FA_nu WF'' UpdateCorrectness WFO S i0). auto.
    split. sinversion i0.
    destruct sigma, sigma'. sinversion i.
    assert(LBase := insert_correctness (generate_fresh_location g1) empty_location f f0 H5).
    assert(EmptyST : Delta0 |- typeof empty_location <:
                     type_nv (atype_classtype (atype_class C GL))). eauto.
    assert(WFS'N := lenv_insert_preserves_WFSigmaRho (f, g1)
                    (f0, (generate_fresh_location g1)) (env0, k) (generate_fresh_location g1)
                    empty_location WFsigma_nu i).
   assert(EmptyST' := t_subtyping_relaxed Delta0 (typeof empty_location)
                      (type_nv (atype_classtype (atype_class C GL)))
                      EmptyST).
    assert(WFS'N' := rvenv_insert_preserves_WFSigmaRho (f0, generate_fresh_location g1)
                     (env0, k)
                     (atype_classtype (atype_class C GL)) (generate_fresh_location g1)
                     (env'0, generate_fresh_variable_location k)
                     empty_location (generate_fresh_variable_location k)
                     WFS'N LBase EmptyST' i0).
    assert(WFS''N' := lenv_bind_preserves_WFSigmaRho (f0, generate_fresh_location g1)
                      (env, fresh) (env'0, generate_fresh_variable_location k)
                      (generate_fresh_variable_location k) (atype_class C GL) ML RML
                      WFS'N' b0).
    assert(InsAd := bind_methods_add (f0, generate_fresh_location g1) (env, fresh)
                    (generate_fresh_variable_location k) ML RML (generate_fresh_location g1)
                    empty_location (atype_class C GL) LBase b0).
    assert(WFS'''N' := lenv_update_preserves_WFSigmaRho (env, fresh) (env', fresh)
                       (env'0, generate_fresh_variable_location k)
                       (generate_fresh_location g1)
                       empty_location
                       (class_instance_object (atype_class C GL) RFL RML)
                       WFS''N' InsAd u). auto. split. wfdestruct WFEnvs. sinversion i0.
    assert(WFN' := rvenv_insert_preserves_WF (env0, k)
                   (env'0, generate_fresh_variable_location k)
                   (generate_fresh_variable_location k)
                   (atype_classtype (atype_class C GL))
                   l_base WFRvenv i0).
    assert(WFStore' := insert_lenv_well_formed empty_location sigma sigma' l_base WFLenv
                       i).
    assert(WFStore'' := bind_methods_lenv_WF sigma' (env, fresh)
                        (generate_fresh_variable_location k) ML (atype_class C GL) RML
                        WFStore' b0).
    assert(WFStore''' := update_lenv_well_formed l_base
                         (class_instance_object (atype_class C GL) RFL RML)). split.
    auto. split. auto. auto. eauto. eauto.
    sinversion i0.
    assert(LenvBaseSuperset' := lenv_insert_preserves_sigma0 sigma sigma' l_base
                                empty_location LenvBaseSuperset i).
    assert(LenvBaseSuperset'' := bind_methods_preserves_sigma0 sigma' (env, fresh)
                                 (generate_fresh_variable_location k) ML (atype_class C GL)
                                 RML LenvBaseSuperset' b0).
    assert(LenvBaseSuperset''' := lenv_update_preserves_sigma0 (env, fresh) (env', fresh)
                                  l_base (class_instance_object (atype_class C GL) RFL RML)
                                  LenvBaseSuperset'' u). auto.
  - Case "l1.f -> l2".
    wfdestruct WFEnvs.
    assert(WFObject := WFsigma l1 (class_instance_object N RFL RML) b). sinversion Exp_typing.
    assert(WF := field_lookup_well_formed N RFL RML l1 l2 nu sigma f (type_nv G) T b i
                 Exp_typing WFObject). destruct WF as [ S [ Typ Sub ] ].
    split. right. exists S. eauto. repeat(split;auto). constructor.
  - Case "l1.m -> l2".
    wfdestruct WFEnvs.
    assert(WFObject := WFsigma l1 (class_instance_object N FL ML) b).
    assert(WF := method_lookup_well_formed N FL ML l1 l2 nu sigma m R b i Exp_typing
                 WFObject). destruct WF as [ S [ Typ Sub ] ]. split. right. eauto.
    repeat(split;eauto). constructor.
  - Case "l1.p = l2 -> l2".
    wfdestruct WFEnvs. sinversion Exp_typing. split.
    emdestruct Reduction Error (sem_subtyping_em Delta0 R (type_nv G0)).
    * SCase "l1.p = l2 is correct".
      right. exists R. split. emdestruct L1L2Eq L1L2Diff (location_em l1 l2). subst.
      + SSCase "l1 = l2".
        assert(TUnique := uniqueness_types nu sigma Gamma0 T R (exp_location l2) WFEnv
                                           WFLenv WFRvenv H2 H10). subst.
        sinversion WFLenv. sinversion H2.
        assert(NDup := NoDupKeysLemma l2 o0 (class_instance_object N FL ML) wenv H H1 b).
        subst.
        assert(Sigma'L2 := update_lenv_correctness l2
                                        ((class_instance_object N
                                             (update_runtime_field FL (f, (G, l2))) ML))
                                        (wenv, fresh) sigma' u). econstructor.
        apply Sigma'L2. simpl. auto.
      + SSCase "l1 <> l2".
    assert(UNN := update_lenv_not_noise l1 l2 sigma sigma' o
                  (class_instance_object N (update_runtime_field FL (f, (G, l2))) ML)
                  WFLenv L1L2Diff b0 u). econstructor. eauto.
    inversion H10. destruct sigma. inversion WFLenv.
    assert(NDUp := NoDupKeysLemma l2 o0 o f0 H12 H0 b0). subst. auto.
      + SSCase "R0 <: G0".
        auto. eauto using subtyping_reflexivity'.
    * SCase "l1.p = l2 is an error".
      left. sinversion H10. sinversion WFLenv.
      assert(WF := NoDupKeysLemma l2 o o0 wenv H b0 H0). subst.
      assert(BDN := bound_delta_neutral T T' H7). subst.
      assert(W := WFsigma l1 (class_instance_object N FL ML) b). sinversion W.
      sinversion W. sinversion H6. sinversion H9. sinversion H16.
      assert(T' = type_nv (type_classtype (C <<< convert_anvtype_list GL >>>))).
      sinversion H2. destruct WFLenv. 
      assert(ND := NoDupKeysLemma l1 o
                   (class_instance_object (atype_class C GL) FL ML)
                   wenv0 H3 H4 b). subst. simpl. auto. subst.
      assert(Contradiction := subtyping_well_formed C GL FL ML (wenv, fresh) nu l1 l'
                              f0 G G0 o0 W b i H15 s). contradiction.
    * split. constructor. split. auto. split.
      assert(X := FA_sigma l1 (class_instance_object N FL ML) b). sinversion X.
      assert(F : fully_annotated_type
         (typeof
            (class_instance_object N (update_runtime_field FL (f, (G, l2)))
               ML))). simpl. auto.
      assert(FA := lenv_update_preserves_fully_annotated sigma sigma' l1
                   (class_instance_object N (update_runtime_field FL (f, (G, l2))) ML)
                   FA_sigma WFLenv F u). auto. split.
      assert(WFSigma' := lenv_update_preserves_WF nu sigma sigma' l1
                       (class_instance_object N (update_runtime_field FL (f, (G, l2))) ML)
                       WFsigma u). auto. split.
      assert(WFSigma'Nu := lenv_update_preserves_WFSigmaRho sigma sigma' nu l1
                           (class_instance_object N FL ML)
                           (class_instance_object N (update_runtime_field FL (f, (G, l2))) ML)
                           WFsigma_nu b u). auto. split. split. auto. split.
      assert(WFSigma' := update_lenv_well_formed l1
                         (class_instance_object N (update_runtime_field FL (f, (G, l2))) ML)
                         sigma sigma' WFLenv u). auto. auto.
      assert(WFL := lenv_update_preserves_sigma0 sigma sigma' l1
                    (class_instance_object N (update_runtime_field FL (f, (G, l2))) ML)
                    LenvBaseSuperset u). auto.
  - Case "l(LL) -> [ T, e' ]".
    split. right. exists T.
    assert(FA := FA_sigma l (closure_object GxL T e) b). sinversion FA. split.    
    wfdestruct WFEnvs.
    assert(X := WFsigma l (closure_object GxL T0 e) b). sinversion X. sinversion Exp_typing.
    assert(E'Typing := variable_substitution_preserves_typing nu nu' sigma Gamma0 Gamma' Delta0
                       e GxL tauL e' LLref S RL LL H8 H9 m a l0 l1 H12).
    destruct E'Typing as [ S' [ Exp'Typing Subtyping ] ]. econstructor. eauto.
    sinversion Exp_typing. sinversion H8. wfdestruct WFEnvs.
    sinversion WFLenv. sinversion Exp_typing. sinversion H9.
    assert(X := NoDupKeysLemma l o (closure_object GxL T0 e) wenv H H0 b). subst. simpl in H9.
    sinversion H9.  eauto using subtyping_reflexivity'.
    assert(fully_annotated_expression (exp_location l)). constructor.
    split. constructor.
    assert(X := FA_sigma l (closure_object GxL T e) b). sinversion X. auto.
    assert(X := FA_sigma l (closure_object GxL T e) b). sinversion X.
    assert(FA_e' := variable_substitution_preserves_fully_annotated e GxL e' tauL
                    H5 H4 m). auto. split.
    wfdestruct WFEnvs. assert(X := WFsigma l (closure_object GxL T e) b).
    sinversion X. assert(FA := FA_sigma l (closure_object GxL T e) b). sinversion FA.
    assert(FA_nu' := actual_to_formals_preserves_fully_annotated nu nu' GxL LLref tauL
                     WFRvenv FA_nu H4 a). auto. split. auto. split. sinversion Exp_typing.
    assert(WFStore := actual_to_formals_preserves_well_formed_store sigma nu nu'
                      tauL LLref GxL Delta0 Gamma0 LL RL e e' WFsigma m a l0 l1 H8). auto.
    sinversion FA_e. split. sinversion Exp_typing.
    assert(WFSigmaNu' := substitution_preserves_WFSigmaRho sigma nu nu' tauL LLref GxL
                         Delta0 Gamma0 LL RL e e' WFsigma_nu m a l0 l1 H10). auto.
    split. split. sinversion Exp_typing.
    assert(ATF := actual_to_formals_preserves_well_formed_rvenv sigma nu nu'
                  tauL LLref GxL Delta0 Gamma0 LL RL e e' WFsigma m a l0 l1 H10). auto.
    wfdestruct WFEnvs. repeat(split;auto). auto.
  - Case "(GxL) => e -> l".
    sinversion Exp_typing. assert(ClosureInSigma' :=
                                  (@insert_fresh_correctness _ _ generate_fresh_location)
                                  l (closure_object GxL T e) sigma
                                 sigma' i). split. right.
    exists (type_nv (projection GxL --> T)). split. econstructor. eauto. simpl.
    assert(X := runtime_types_eq GxL). rewrite <- X. auto. eauto using subtyping_reflexivity. 
    split. constructor. split. auto. split.
    wfdestruct WFEnvs. sinversion FA_e. assert(FAsigma' := fully_annotated_ftype GxL T H2 H4).
    assert(fully_annotated_type
                (typeof (closure_object GxL T e))). simpl. auto.
    assert(FA_sigma' := lenv_insert_preserves_fully_annotated sigma sigma'
                        l (closure_object GxL T e) FA_sigma
                        WFLenv H i). auto.
    assert(WFObj : well_formed_object nu sigma (closure_object GxL T e)).
    econstructor. eauto. eauto. eauto. split. 
    assert(WFStore' := lenv_insert_preserves_WF nu sigma sigma' l
                       (closure_object GxL T e) WFsigma
                       WFObj i). auto. split.
    assert(WFSigma'Nu := lenv_insert_preserves_WFSigmaRho sigma sigma' nu l
                         (closure_object GxL T e) WFsigma_nu i). auto. split. wfdestruct WFEnvs.
    assert(WFLenv' := insert_lenv_well_formed (closure_object GxL T e)
                      sigma sigma' l WFLenv i). split. auto. split. auto. auto.
    assert(Lenv'BaseSuperset := lenv_insert_preserves_sigma0 sigma sigma' l
                                (closure_object GxL T e)
                                LenvBaseSuperset i). auto.
  - Case "[ T, l ] -> l".
    sinversion Exp_typing. split.
    emdestruct Subtyping NotSubtyping (subtyping_em Delta0 R0 R).
    (* If the type of l is a subtype of T, then the type is preserved *)
    right. exists R0. split. eauto. eauto.
    (* Otherwise, we have a subtype violation error *)
    left. assert(X := subtype_objects l o nu sigma R0 R b H6 s Exp_typing).
    intuition.
    split. econstructor. repeat(split; eauto).
  - Case "e.p -> e'.p".
    sinversion FA_e.
    sinversion Exp_typing.
    assert(IH := IHExp_Reduction H0 FA_nu FA_sigma WFEnvs WFsigma WFsigma_nu
                 LenvBaseSuperset T H2). destruct IH as [ [ E'Error | IH ] WF ].
    destruct WF as [ FA_e' [ FA_nu' [ FASigma' [ WFStore' [ WFSigma'Nu' [ WFEnvs'
                                                 LenvBaseSuperset' ] ] ] ] ] ]. split.
    left. constructor. auto. split. constructor. auto. repeat(split;auto).
    sinversion FA_e. split. right.  destruct IH as [ S [ E'Typing ST ] ].
    assert(FA_T := fully_annotated_typing nu sigma e T
                   FA_sigma FA_nu WFsigma_nu H1 H2).
    destruct WF as [ FA_e' [ FA_nu' [ FA_sigma' [ WFSigma' [ WFSigma'Nu' [ WFEnvs'
                                      LenvBaseSuperset' ] ] ] ] ] ].
    assert(FA_S := fully_annotated_typing nu' sigma' e' S
                   FA_sigma' FA_nu' WFSigma'Nu' FA_e' E'Typing).
    assert(X := bound_delta_neutral T T' H7). subst.
    assert(Getter := getter_subtyping T' S G p FA_T FA_S ST H9).
    destruct Getter as [ G' [ Get [ Subst BD] ] ].
    exists (type_nv G'). split. econstructor. apply E'Typing.
    apply Subst. auto. auto. split. constructor. 
    destruct WF as [ FA_e' [ FA_nu' [ FA_sigma' [ WFSigma' [ WFSigma'Nu' [ WFEnvs'
                                      LenvBaseSuperset' ] ] ] ] ] ]. auto.
    repeat(split;
    destruct WF as [ FA_e' [ FA_nu' [ FASigma' [ WFStore' [ WFSigma'Nu' [ WFEnvs'
                                                 LenvBaseSuperset' ] ] ] ] ] ]; auto).
  - Case "e1.p = e2 -> e1'.p = e2".
    sinversion FA_e. sinversion Exp_typing.
    assert(IH := IHExp_Reduction H1 FA_nu FA_sigma WFEnvs WFsigma WFsigma_nu
                 LenvBaseSuperset T H4).
    destruct IH as [ [ E1'Error | IH ]
                     [ FA_e' [ FA_nu' [ FAStore' [ WFStore' [ WFSigma'Nu' [ WFEnvs'
                       LenvBaseSuperset' ] ] ] ] ] ] ]. split.
    left. constructor. auto. split. constructor. auto. auto. repeat(split;auto).
    destruct IH as [ S [ E'Typing Subtyping ] ].
    assert(X := bound_delta_neutral T T' H9). subst.
    assert(FA_T' := fully_annotated_typing nu sigma e1 T'
                   FA_sigma FA_nu WFsigma_nu H1 H4).
    assert(FA_S := fully_annotated_typing nu' sigma' e1' S
                   FAStore' FA_nu' WFSigma'Nu' FA_e' E'Typing).
    assert(FA_R := fully_annotated_typing nu sigma e2 R
                   FA_sigma FA_nu WFsigma_nu H3 H12).
    assert(Weak := weakening_setter e1 e1' e2 nu nu' sigma sigma' R WFsigma WFsigma_nu
                   Exp_Reduction H12).
    destruct Weak as [ S' [ Typing Styping ] ].
    assert(FA_S' := fully_annotated_typing nu' sigma' e2 S'
                    FAStore' FA_nu' WFSigma'Nu' H3 Typing).
    assert(Setter := setter_subtyping T' S R S' G p FA_T' FA_S FA_R FA_S' Subtyping
                     H11 Styping).
    destruct Setter as [ G' [ Setr [ Subst ST ] ] ]. split. right.
    exists S'. split. econstructor. eauto. eauto. eauto. eauto. auto.
    split. constructor. auto. auto. repeat(split;auto).
  - Case "l.p = e -> l.p = e'".
    sinversion FA_e. sinversion Exp_typing.
    assert(IH := IHExp_Reduction H3 FA_nu FA_sigma WFEnvs WFsigma WFsigma_nu
                 LenvBaseSuperset R H12).
    destruct IH as [ [ E2'Error | IH ]
                     [ FA_e' [ FA_nu' [ FAStore' [ WFSigma' [ WFSigma'Nu' [ WFEnvs'
                                        LenvBaseSuperset' ] ] ] ] ] ] ]. split.
    left. apply Ind_Err_setter_right. auto. split. constructor. auto. auto.
    repeat(split;auto). split.
    destruct IH as [ S [ E'Typing Subtyping ] ].
    assert(FA_T := fully_annotated_typing nu sigma (exp_location l) T
                   FA_sigma FA_nu WFsigma_nu H1 H4).
    assert(BDN := bound_delta_neutral T T' H9). subst.
    assert(Weak := weakening_setter e e' (exp_location l) nu nu' sigma sigma' T' WFsigma
                   WFsigma_nu Exp_Reduction H4).
    destruct Weak as [ S' [ Typing Subtyping' ] ].
    assert(BDN := bound_delta0_typing l nu' sigma' S' Typing).
    assert(FA_S' := fully_annotated_typing nu' sigma' (exp_location l)
                S' FAStore' FA_nu' WFSigma'Nu' H1 Typing).
    assert(FA_R := fully_annotated_typing nu sigma e
                   R FA_sigma FA_nu WFsigma_nu H3 H12).
    assert(FA_S := fully_annotated_typing nu' sigma' e'
                   S FAStore' FA_nu' WFSigma'Nu' FA_e' E'Typing).
    assert(Setter := setter_subtyping T' S' R S G p FA_T FA_S' FA_R FA_S Subtyping'
                     H11 Subtyping).
    destruct Setter as [ G' [ SetterT [ BD ST ] ] ]. right.
    exists S. split. econstructor. eauto. eauto. eauto. eauto. eauto.
    auto. split. constructor. auto. auto. repeat(split;auto).
  - Case "tau = e -> tau = e'".
    sinversion FA_e. sinversion Exp_typing.
    assert(IH := IHExp_Reduction H0 FA_nu FA_sigma WFEnvs WFsigma WFsigma_nu LenvBaseSuperset
                 R H6).
    destruct IH as [ [ Err | WT] [ FA_e' [ FA_nu' [ FA_sigma' [ WFStore'
                                 [ WFSigma'Nu' [ WFEnvs' LenvBaseSuperset' ] ] ]
                                 ] ] ] ];split. left. constructor. auto. split. econstructor.
    auto. repeat(split;auto).
    destruct WT as [ S [ WT Sub ] ]. right. exists S.
    assert(TauInNu' := subst_not_remove tau G l nu nu' sigma sigma' e e' H8 Exp_Reduction).
    destruct TauInNu' as [ l' I ]. split. econstructor. eauto. eauto.
    eauto using subtyping_reflexivity. split. econstructor. auto. repeat(split;auto).
  - Case "l(eL) -> l(eL')".
    sinversion FA_e. sinversion Exp_typing.
    assert(IH := IHExp_Reduction H2 FA_nu FA_sigma WFEnvs WFsigma WFsigma_nu LenvBaseSuperset
                 RL H9). 
    destruct IH as [ [ Err | WT] [ FA_e' [ FA_nu' [ FA_sigma' [ WFStore'
                                 [ WFSigma'Nu' [ WFEnvs' LenvBaseSuperset' ] ] ]
                                 ] ] ] ];split. left. apply Ind_Err_call_args.
    auto. split. constructor. auto. auto. repeat(split;auto). right. exists R. split.
    destruct WT as [ SL [ Exp'Typing Subtyping ] ].
    assert(Weakening := weakening_list args args' (exp_location l) nu nu' sigma
                                       sigma' GL RL R SL WFsigma WFsigma_nu
                                       e H7 H9 Exp'Typing). econstructor. eauto. eauto.
    eauto using subtyping_reflexivity'. split. constructor. auto. eauto. repeat(split;auto).
  - Case "e(eL) -> e'(eL)".
    sinversion FA_e. sinversion Exp_typing.
    assert(IH := IHExp_Reduction H1 FA_nu FA_sigma WFEnvs WFsigma WFsigma_nu LenvBaseSuperset
                 (type_nv (GL --> R)) H7).
    destruct IH as [ [ Err | Ty ]
                     [ FA_e' [ FA_nu' [ FA_sigma' [
                         WFStore' [ WFSigma'_nu' [ WFEnvs' LenvBaseSuperset' ] ] ] ] ] ] ]. 
    split. left. constructor. auto. split. constructor. auto. auto. repeat(split;auto).
    split. right. destruct Ty as [ S [ Exp'Typing Subtyping ] ].
    assert(FA_S := fully_annotated_typing nu' sigma' e' S FA_sigma' FA_nu'
                   WFSigma'_nu' FA_e' Exp'Typing). 
    assert(FS := function_subtyping GL R S Delta0 Subtyping FA_S).
    destruct FS as [ GL' [ T' Eq ] ]. subst. 
    assert(Weakening := weakening_list_base e e' eL nu nu' sigma sigma' RL R T' GL GL'
                        WFsigma WFsigma_nu Exp_Reduction H9 H7 Exp'Typing).
    destruct Weakening as [ SL [ WT ST ] ]. econstructor. split. econstructor 5. eauto.
    eauto.
    assert(X := function_reduction_preserve_type GL R (type_nv (GL' --> T')) nu sigma nu'
                                                 sigma' e e' H7 Exp_Reduction Exp'Typing).
    sinversion X. eauto using subtyping_reflexivity'. split. constructor. auto. auto.
    repeat(split;auto). 
  - Case "[T, e] -> [T, e']".
    sinversion FA_e. sinversion Exp_typing.
    assert(IH := IHExp_Reduction H2 FA_nu FA_sigma WFEnvs WFsigma WFsigma_nu LenvBaseSuperset
                 R0 H8).
    destruct IH as [ [ Err | Red ]
                     [ FA_e' [ FA_nu' [ FAsigma' [ WFStore'
                     [ WFSigma'Nu' [ WFEnvs' LenvBaseSuperset' ]
                     ] ] ] ] ] ]. split. left. constructor. auto. split. constructor.
    auto. auto. repeat(split;auto).
    destruct Red as [ S [ Exp'Typing Subtyping ] ]. split. right. exists R. split.
    econstructor. apply Exp'Typing.
    assert(FA_S := fully_annotated_typing nu' sigma' e' S FAsigma' FA_nu'
                                          WFSigma'Nu' FA_e' Exp'Typing).
    assert(FA_R0 := fully_annotated_typing nu sigma e R0 FA_sigma FA_nu WFsigma_nu
                    H2 H8). eauto using subtyping_reflexivity'.
    split. econstructor. eauto. eauto. repeat(split;auto).
  - Case "e :: eL -> e' :: eL".
    sinversion FA_e. sinversion Exp_typing.
    assert(IH := IHExp_Reduction H1 FA_nu FA_sigma WFEnvs WFsigma WFsigma_nu LenvBaseSuperset
                 T H7). destruct IH as [ [ Err | Red ]
                                         [ FA_e' [ FA_nu' [ FA_sigma' [ WFStore' [ WFSigma'Nu'
                                         [ WFEnvs' LenvBaseSuperset'] ] ] ] ] ] ]. split.
    left. constructor. auto. split. constructor. auto. auto. repeat(split;auto). split.
    right. assert(X := expression_list_left e eL e' nu sigma nu' sigma'
                                            T TL H7 H9 Exp_Reduction).
    destruct X as [ SL [ ASS Sub ] ]. destruct Red as [ S [ Exp'Typing Subtyping ] ].
    exists (S :: SL). split. constructor. auto. auto. constructor;auto. split.
    constructor. auto. auto. repeat(split;auto).
  - Case "e :: eL -> e :: eL'".
    sinversion FA_e. sinversion Exp_typing.
    assert(IH := IHExp_Reduction H2 FA_nu FA_sigma WFEnvs WFsigma WFsigma_nu LenvBaseSuperset
                 TL H9).
    destruct IH as [ [ Err | Red ]
                     [ FA_eL' [ FA_nu' [ FA_Store' [ WFStore' [ WFSigma'Nu'
                              [ WellFormedEnvs' LenvBaseSuperset' ] ] ] ] ] ] ]. split.
    left. constructor 2. auto. split. auto. repeat(split;auto). split. right.
    assert(X := expression_list_right e eL eL' nu sigma nu' sigma' T TL H7 H9 e0).
    destruct X as [ S [ Typ ST ] ]. destruct Red as [ SL [ Exp'Typing Subtyping ] ].
    exists (S :: SL). split. eauto. auto. split. auto. repeat(split;auto).
Qed.

(* 
 * Type soundness using the type system without assignability on premises, where we assume nu and sigma are
 * generic environments.
 *)

Theorem type_soundness : forall nu nu' sigma sigma' e e' T,
    (* IF *)                          
    (* e is fully annotated *)                           
    fully_annotated_expression e ->
    (* nu is fully annotated *)
    fully_annotated_subst nu ->
    (* sigma is fully annoatated *)
    fully_annotated_store sigma ->
    (* The environments are well formed *)
    well_formed_envs nu sigma Gamma0 ->
    well_formed_store nu sigma ->
    lenv_base_superset sigma ->
    sigma |--S nu ->
    (* The expression e is well typed *)
    nu ; sigma ; Delta0 ; Gamma0 |- e : T ->
    (* The expression e is evaluated into e' in multiple steps *)
    << nu, sigma, e >> || << nu', sigma', e' >> ->
    (* The evaluated expression e' is a normal form *)                                          
    (* THEN one the following hold *)
    irred nu' sigma' e' ->
    (* b) e' is an error *)
    (error nu' sigma' e' \/
     (* a) e' is a well typed value *)
     (value e' /\
      exists S,
        (* e' has type S *)
        nu' ; sigma' ; Delta0 ; Gamma0 |- e' : S)) /\
    (* Moreover, the environments after the execution are well formed and fully annoatated. *)
    fully_annotated_expression e' /\
    fully_annotated_subst nu' /\
    fully_annotated_store sigma' /\
    well_formed_envs nu' sigma' Gamma0 /\
    well_formed_store nu' sigma' /\
    sigma' |--S nu' /\
    lenv_base_superset sigma'.
Proof.
  (* By induction on the derivation sequence <nu, sigma, e> || <nu', sigma', e'> *)
  intros nu nu' sigma sigma' e e' T FA_e FA_nu FASigma WFEnvs WFStore SigmaSSigma0 WFSigmaNu
         Exp_typing RedStar Irred.
  generalize dependent T.
  induction RedStar;intros.
  - Case "e ->* e".
  assert(Progress := progress sigma nu e T FA_e FA_nu FASigma Exp_typing WFEnvs
                     WFStore WFSigmaNu).
  destruct Progress as [ Val | Red ].
    * SCase "e is a value". split. right. split. auto. exists T. auto. repeat(split;auto).
    * SCase "e -> e'". destruct Red as [ [ nu' [ sigma' [ e' R ] ] ] | Err ].
      assert (Contr := Irred e' nu' sigma').
      contradiction. split. left. auto. repeat(split;auto).
  - Case "e -> e' and e' ->* e''".
    assert(Preservation := preservation e e' sigma sigma' nu nu' T FA_e FA_nu
                           FASigma Exp_typing WFEnvs WFStore WFSigmaNu H SigmaSSigma0).
    assert(Progress := progress sigma nu e T FA_e FA_nu FASigma Exp_typing WFEnvs
                     WFStore WFSigmaNu).
    destruct Preservation as [ [ Err | WellTyped ] WF ];
    destruct Progress as [ Val | [ [ nu''' [ sigma''' [ e''' Red ] ] ] | ErrCond ] ];
    destruct WF as [ FA_e' [ FA_nu' [ FA_sigma' [ WFStore'
                   [ WFSigma'Nu' [ WFEnvs' Sigma'SSigma0 ] ] ] ] ] ].
    * SCase "e value". destruct Val as [ l V ]. subst. sinversion H.
    * SCase "e' contains error".
      assert(Irreducible := error_irreducible nu' sigma' e' WFStore' Sigma'SSigma0 Err).
      sinversion RedStar.
      + SSCase "e' ->*  e'". split. left. auto. repeat(split;auto).
      + SSCase "e' -> e'''". 
      unfold irred in Irreducible. assert(Contradiction := Irreducible e'0 nu'0 sigma'0).
      intuition.
    * SCase "e contains error".
      assert(Irreducible := error_irreducible nu sigma e WFStore SigmaSSigma0 ErrCond).
      assert(Contradiction := Irreducible e' nu' sigma'). intuition.
    * SCase "e is a value". destruct Val as [ l V ]. subst. sinversion H.
    * SCase "e' is well typed".
      destruct WellTyped. destruct H0 as [ Exp'Typing ST ].
      assert(IH := IHRedStar FA_e' FA_nu' FA_sigma' WFEnvs' WFStore' Sigma'SSigma0
                   WFSigma'Nu' Irred x Exp'Typing).
      destruct IH as [ [ Error | Value ] WF ]. split. left. eauto. eauto.
      destruct Value as [ Val [ S Typing (* [ Typing S_sub_x ] *) ] ].
      assert(FA_x := fully_annotated_typing nu' sigma' e' x FA_sigma' FA_nu' WFSigma'Nu'
                     FA_e' Exp'Typing).
      assert(FA_T := fully_annotated_typing nu sigma e T FASigma FA_nu WFSigmaNu FA_e
                     Exp_typing).
      destruct WF as [ FA_e'' [ FA_nu'' [ FA_sigma'' [ WFEnvs'' [ WFStore'' [
                       WFSigma''Nu'' LenvBaseSuperset'' ] ] ] ] ] ].
      assert(FA_S := fully_annotated_typing nu'' sigma'' e'' S FA_sigma'' FA_nu''
                     WFSigma''Nu'' FA_e'' Typing).
      split. right. split. eauto. exists S. eauto.
      repeat(split;eauto).
    * SCase "e contains error".
      assert(Irreducible := error_irreducible nu sigma e WFStore SigmaSSigma0 ErrCond).
      assert(Contradiction := Irreducible e' nu' sigma'). intuition.
Qed.

(* Type soundness for closed terms, i.e. with nu0, sigma0, Delta0, Gamma0 *)

Corollary type_soundness_user_programs : forall e e' T nu' sigma',
    (* IF *)                                           
    (* e is fully annotated *)
    fully_annotated_expression e ->
    (* e is well typed *)
    nu0 ; sigma0 ; Delta0 ; Gamma0 |- e : T ->
    (* e reduces into e' *)                                          
    << nu0, sigma0, e >> || << nu', sigma', e' >> ->
    (* e' is a normal form *)
    (* THEN one of the following hold *)
    irred nu' sigma' e' ->    
    (* b) e' is an error *)
    (error nu' sigma' e' \/
     (* a) e' is a well typed value*)
     (value e' /\
      exists S,
        nu' ; sigma' ; Delta0 ; Gamma0 |- e' : S)) /\
    (* Moreover, the environments after the execution are well formed and fully annoatated. *)
    fully_annotated_expression e' /\
    fully_annotated_subst nu' /\
    fully_annotated_store sigma' /\
    well_formed_envs nu' sigma' Gamma0 /\
    well_formed_store nu' sigma' /\
    sigma' |--S nu' /\
    lenv_base_superset sigma'.
Proof.
  intros e e' T nu' sigma' FA_e ExpTyping RedStar Irred.
  assert(FA_nu0 : fully_annotated_subst nu0). unfold fully_annotated_subst. intros.
  sinversion H.
  assert(FA_sigma0 : fully_annotated_store sigma0). unfold fully_annotated_store. intros.
  sinversion H. sinversion H0. constructor. sinversion H0.
  assert(WFEnvs0 : well_formed_envs nu0 sigma0 Gamma0). split. constructor.
  constructor. split. constructor. constructor. auto. auto. constructor. auto. auto.
  assert(WFStore0 : well_formed_store nu0 sigma0). 
  unfold well_formed_store. intros. sinversion H.
  sinversion H0. constructor. sinversion H0.
  assert(LenvBaseSuperset : lenv_base_superset sigma0). unfold lenv_base_superset. intros.
  auto. assert(WFNu0Sigma0 : sigma0 |--S nu0). intros l G l0 A. sinversion A.
  assert(Soundness := type_soundness nu0 nu' sigma0 sigma' e e' T FA_e FA_nu0
                      FA_sigma0 WFEnvs0 WFStore0 LenvBaseSuperset WFNu0Sigma0 ExpTyping
                      RedStar Irred). auto.
Qed.

(* This is the MAIN soundness theorem, i.e. the soundness theorem for the message safe type system with assignability on premises. *)

Corollary type_soundness_complete : forall e e' T nu' sigma',
    (* IF *)                                           
    (* e is fully annotated *)
    fully_annotated_expression e ->
    (* e is well typed with the standard message safe type system *)
    nu0 ; sigma0 ; Delta0 ; Gamma0 |-STD e : T ->
    (* e reduces into e' *)                                          
    << nu0, sigma0, e >> || << nu', sigma', e' >> ->
    (* e' is a normal form *)
    (* THEN one of the following hold *)
    irred nu' sigma' e' ->
    (* b) e' is an error *)
    (error nu' sigma' e' \/
     (* a) e' is a well typed value*)
     (value e' /\
      exists S,
        nu' ; sigma' ; Delta0 ; Gamma0 |-STD e' : S)) /\
    fully_annotated_expression e' /\
    fully_annotated_subst nu' /\
    fully_annotated_store sigma' /\
    well_formed_envs nu' sigma' Gamma0 /\
    well_formed_envs nu' sigma' Gamma0 /\
    well_formed_store nu' sigma' /\
    sigma' |--S nu' /\
    lenv_base_superset sigma'.
Proof.                                         
  intros e e' T nu' sigma' FA_e STDExpTyping Reduction Irred.
  assert(Subset := relaxed_ts_subset_standard_ts nu0 sigma0 Delta0 Gamma0 e T
                   STDExpTyping).
  assert(Soundness := type_soundness_user_programs e e' T nu' sigma' FA_e Subset
                      Reduction Irred). destruct Soundness. split.
  destruct H as [ Error | [ Value [ S (* [ WT ST ] *) ] ] ]. auto. right.
  split. auto. destruct Value as [ l V ]. subst. 
  assert(L := location_typing_equivalence nu' sigma' Delta0 Gamma0 l S). destruct L.
  exists S. auto.
  destruct H0 as [ FA_e' [ FA_nu' [ FA_sigma' [ WFEnvs' [ WFStore' [ WFSigma'Nu'
                                                LenvBaseSuperset' ] ] ] ] ] ].
  repeat(split;auto).
Qed.

(* 
Note: 

PRECONDITIONS
fully_annotated_expression doesn't use the type system
<< nu0, sigma0, e >> || << nu', sigma', e' >> doesn't use the type system
irred nu' sigma' e' doesn't use the type system

POSTCONDITIONS
value e' does not use the type system
error uese the type system only with the location rule, that is equivalent in both TS
Subtyping doesn't use the type system
fully_annotated don't use the type system
well_formed_store contains a typing in well_formed_object in closure object,
                  but a closure never changes during evaluation and it was well
                  part of the original expression, that was well typed in the standard way
*)

End Properties.
