(* ------------------------------  Fletch check mode operational semantics ----------------------------------- *)

Require Export Arith.
Require Export fletch_type_system.

Module StandardCheckModeOS (C : ClassTable).
Module StandardTSImpl := StandardTS(C).
Export StandardTSImpl.

(* SUBTYPING FOR THE OPERATIONAL SEMANTICS *)

Reserved Notation "Delta '|S-' T '<:' R" (at level 40, T at level 40).
Reserved Notation "Delta '|S-^' TL '<:' RL" (at level 40, TL at level 40).
Inductive sem_subtyping : fletch_tenv -> fletch_type -> fletch_type -> Prop :=
  | sem_s_spec : forall Delta T R,
               Delta |- dynamic_sub(T) << R ->
               Delta |S- T <: R
  | sem_s_fun : forall Delta T R GL HL,
              sem_assignable_list Delta (ltype_nv HL) (ltype_nv GL) ->
              ((sem_assignable Delta T R) \/ R = type_void) ->
              Delta |S- (type_nv ((GL) --> T)) <:
                       (type_nv ((HL) --> R))
  | sem_s_obj : forall Delta  GL T,
              Delta |S-
                    type_nv ((GL) --> T) <:
                    (type_nv (type_classtype type_object))
with
 sem_assignable : fletch_tenv -> fletch_type -> fletch_type -> Prop :=
  | sem_a_subtyping : forall Delta T R,
                    Delta |S- T <: R ->
                    sem_assignable Delta T R
  | sem_a_supertyping : forall Delta T R,
                      (~ exists GL S, R = type_nv ((GL) --> S)) \/
                      (~ exists GL S, T = type_nv ((GL) --> S)) ->
                      Delta |S- R <: T ->
                      sem_assignable Delta T R
with
 sem_subtyping_list : fletch_tenv -> list fletch_type -> list fletch_type -> Prop :=
  | sem_subtyping_empty : forall Delta,
              Delta |S-^ nil <: nil
  | sem_subtyping_cons : forall Delta G H GL HL,
             Delta |S- G <: H ->
             Delta |S-^ GL <: HL ->
             Delta |S-^ (G :: GL) <: (H :: HL)
with
 sem_assignable_list : fletch_tenv -> list fletch_type -> list fletch_type -> Prop :=
  | sem_assignable_empty : forall Delta,
              sem_assignable_list Delta nil nil
  | sem_assignable_cons : forall Delta T S TL SL,
             sem_assignable Delta T S ->
             sem_assignable_list Delta TL SL ->
             sem_assignable_list Delta (T :: TL) (S :: SL)
where "Delta '|S-' T '<:' R" := (sem_subtyping Delta T R)
and "Delta '|S-^' TL '<:' RL" := (sem_subtyping_list Delta TL RL).

Hint Constructors sem_subtyping sem_assignable sem_subtyping_list sem_assignable_list.

(* Mutual induction principle for subtyping. *)
Scheme sem_subtyping_mut := Induction for sem_subtyping Sort Prop
with sem_subtyping_list_mut := Induction for sem_subtyping_list Sort Prop
with sem_assignable_mut := Induction for sem_assignable Sort Prop
with sem_assignable_list_mut := Induction for sem_assignable_list Sort Prop.

Fixpoint isbound (GxL : params) (y : fletch_var) := 
  match GxL with
    | [] => false
    | (G,z) :: xl => if var_compare y (var_user z) then true else (isbound xl y)
  end.

(* 
 * variable_vloc_substitution exp y vl returns an expression exp' obtained by exp
 * by replacing every occurrence of the variable y in exp with the variable location
 * vl.
 *)
Fixpoint variable_vloc_substitution (exp : fletch_exp)
                                    (y : fletch_var)
                                    (vl : @generate_id var_location_id) :=
  match exp with
      | exp_var y => exp_var_location vl
      | exp_getter e1 p => exp_getter (variable_vloc_substitution e1 y vl) p
      | exp_setter e1 p e2 => exp_setter
                                         (variable_vloc_substitution e1 y vl)
                                         p
                                         (variable_vloc_substitution e2 y vl)
      | exp_assignment x e => exp_runtime_assignment vl e
      | exp_call e1 eL => exp_call (variable_vloc_substitution e1 y vl)
                                   (map (fun e' => (variable_vloc_substitution e' y vl)) eL)
      | exp_new N => exp_new N
      | exp_function GxL T e => exp_function GxL T (if (isbound GxL y) 
                                                    then e else
                                                    (variable_vloc_substitution e y vl))
      | exp_location l => exp_location l
      | exp_var_location vl => exp_var_location vl
      | exp_runtime_assignment vl e => exp_runtime_assignment vl
                                                          (variable_vloc_substitution e y vl)
      | exp_frame_expression G e => exp_frame_expression G (variable_vloc_substitution e y vl)
  end.

(* 
 * Applies variable_vloc_substitution replacing a list of variables with a list of
 * variable locations.
 *)
Inductive multiple_vloc_subst : fletch_exp ->
                                list fletch_var ->
                                list (@generate_id var_location_id) ->
                                fletch_exp ->
                                Prop :=
  | mvs_base : forall e,
                 multiple_vloc_subst e [] [] e
  | mvs_ind : forall e e' e'' y yL tau tauL,
                multiple_vloc_subst e yL tauL e' ->
                e'' = variable_vloc_substitution e' y tau ->
                multiple_vloc_subst e (y :: yL) (tau :: tauL) e''.

(*
 * Create a list of runtime fields, i.e., mapping from field to location, given
 * a field declaration.
 *)
Inductive bind_fields :  list fletch_anvtype ->
                         list fletch_field ->
                         runtime_fields ->
                         Prop :=
  | bind_field_base : forall GL,
                        bind_fields GL [] []
  | bind_field_indc : forall GL FL RFL T f,
                        bind_fields GL FL RFL ->
                        bind_fields GL ((field T f) :: FL)
                                    ((f, ((fun A : fletch_nvtype => A) T, loc_null)) :: RFL).

Fixpoint field_subst (PL : generics)
                     (GL : list fletch_anvtype)
                     (FL : list fletch_field) : list fletch_field :=
  match FL with
    | [] => []
    | (field G f) :: FL' => (field (amtype_substitution PL GL G) f) :: (field_subst PL GL FL')
  end.

Inductive get_fields : fletch_aclasstype -> list fletch_field -> Prop :=
  | fields_object : get_fields (atype_class builtin_class_object []) []
  | fields_ind : forall C PL N N' FL ML FL_This FL_Super FL_Super_Prep GL,
                   (C, class C PL N FL ML) _in CT ->
                   N' = mctype_substitution PL GL N ->
                   (get_fields N' FL_Super_Prep) ->
                   FL_This = field_subst PL GL FL ->
                   (forall f G, (field f G) _in FL_Super_Prep ->
                                (forall G', (field f G') not _in FL_This) ->
                                (field f G) _in FL_Super) ->
                   (get_fields (atype_class C GL) (FL_This ++ FL_Super)).

Inductive bind_methods : (@generate_id var_location_id) ->
                         fletch_classtype ->
                         fletch_lenv ->
                         list fletch_method ->
                         (list runtime_method * fletch_lenv)%type ->
                         Prop :=
  | bind_method_base: forall N sigma v_base,
                        bind_methods v_base N sigma [] ([], sigma)
  | bind_method_object : forall l_closure N ML RML
                                sigma sigma' sigma'' (T : fletch_atype) e GxL m v_base,
                         insert_lenv sigma
                                     (closure_object GxL T
                                            (variable_vloc_substitution e
                                              var_this
                                              v_base))
                                     (l_closure, sigma') ->
                         bind_methods v_base N sigma' ML (RML, sigma'') ->
                         bind_methods v_base N sigma
                                     ((method T m GxL e) :: ML)
                                     (((m, l_closure) :: RML), sigma'').

Fixpoint param_subst (PL : generics)
                     (GL : list fletch_anvtype)
                     (GxL : params) : params :=
  match GxL with
      | [] => []
      | (G, x) :: GxL' => ((amtype_substitution PL GL G), x) :: (param_subst PL GL GxL')
  end.

Fixpoint method_subst (PL : generics)
                      (GL : list fletch_anvtype)
                      (ML : list fletch_method) : list fletch_method :=
  match ML with
    | [] => []
    | (method T m GxL e) :: ML' =>
      (method (amttype_substitution PL GL T) m (param_subst PL GL GxL) e) :: 
      (method_subst PL GL ML')
  end.

Inductive get_methods : fletch_aclasstype -> list fletch_method -> Prop :=
  | methods_object : get_methods (atype_class builtin_class_object []) []
  | methods_ind : forall C PL N N' FL ML ML_This ML_Super_Prep ML_Super GL,
                    (C, class C PL N FL ML) _in CT ->
                    N' = mctype_substitution PL GL N ->
                    (get_methods N' ML_Super_Prep) ->
                    ML_This = method_subst PL GL ML ->
                    (forall m GxL T e,
                       (method T m GxL e) _in ML_Super_Prep ->
                       (forall T' GxL' e', (method T' m GxL' e') not _in ML_This) ->
                       (method T m GxL e) _in ML_Super) ->
                    get_methods (atype_class C GL) (ML_This ++ ML_Super).

Fixpoint to_location_type_list GxL LL : list (fletch_anvtype * @generate_id location_id)%type :=
  match (GxL, LL) with
    | (Gx :: xs, l :: xl) => (l,Gx) :: (to_location_type_list xs xl)
    | _ => []
  end.

(*
 * actuals_to_formals nu GxL LL tauL nu' holds IFF
 * nu' contains the same variables as nu plus a binding
 * tau -> l_i : G_i for each l_i in LL and (G_i, x_i) in GxL
 *)

Inductive actuals_to_formals : fletch_rvenv ->
                               params ->
                               list (@generate_id location_id) ->
                               (list (@generate_id var_location_id) *
                               fletch_rvenv)%type ->
                               Prop :=
  | atf_base : forall nu,
                actuals_to_formals nu [] [] ([], nu)
  | atf_ind : forall nu nu' nu'' G x GxT L LT tau tauT,
                insert_rvenv nu (G, L) (tau, nu') ->
                actuals_to_formals nu' GxT LT (tauT, nu'') ->
                actuals_to_formals nu ((G,x) :: GxT) (L :: LT) ((tau :: tauT), nu'').

Fixpoint rvar_to_var (RxL : params) : list fletch_var :=
  match RxL with 
    | [] => []
    | (G,x) :: xs => (var_user x) :: rvar_to_var xs
  end.

(* Fixpoint build_initialization_environment GxL : fletch_rvenv := *)
(*   match GxL with *)
(*     | Gx :: xs => (Gx,loc_null) :: (build_initialization_environment xs) *)
(*     | _ => Nu0 *)
(*   end. *)

Inductive locations : list fletch_exp -> list (@generate_id location_id) -> Prop :=
  | loc_base : locations [] []
  | loc_ind : forall l eL lL,
                locations eL lL ->
                locations ((exp_location l)::eL) (l::lL).
Hint Constructors locations.

Lemma location_is_function : forall LL LLRef LLRef',
                            locations LL LLRef ->
                            locations LL LLRef' ->
                            LLRef = LLRef'.
Proof.
  intros LL.
  induction LL;intros. sinversion H. sinversion H0. auto. sinversion H. sinversion H0.
  assert(X := IHLL lL lL0 H4 H5). subst. auto.
Qed.

Inductive location_subtyping : fletch_lenv ->
                               list (@generate_id location_id) ->
                               list fletch_nvtype -> Prop  :=
  | lst_empty : forall Sigma,
                  location_subtyping Sigma [] []
  | lst_symmety : forall Sigma L LL G GL o,
                    (L,o) __in Sigma ->
                    Delta0 |S- typeof o <: type_nv G ->
                    location_subtyping Sigma LL GL ->
                    location_subtyping Sigma (L :: LL) (G :: GL).

Inductive well_formed_parameters : params -> Prop := 
  | wfp_base : well_formed_parameters []
  | wfp_ind : forall (G : fletch_anvtype) x GxL,
                Delta0 |- (atype_nv G) OK ->
                well_formed_parameters GxL ->
                well_formed_parameters ((G,x) :: GxL).

(* CHECKED MODE OPERATIONAL SEMANTICS FOR FLETCH *)

(*
Differences from the paper: a runtime field is a pair (f,loc) in the paper and (f,G,loc) in Coq where G is a type. Therefore the new constructor (new N) semantics assign (f,G,loc) where G is the type of f in the N class with the N type argument replaced. In the paper, this substitution is performed in the field assignment (setter) where the type of f is looked up from the class table.
*)

Reserved Notation "'<<' nu ',' sigma ',' e '>>' '--->' '<<' nu' ',' sigma' ',' e' '>>'" (at level 39, sigma,nu,e,sigma',nu' at level 38).
Reserved Notation "'[[' nu ',' sigma ',' eL ']]' '--->^' '[[' nu' ',' sigma' ',' eL' ']]'" (at level 39, sigma,nu,e,sigma',nu' at level 38). 
Inductive exp_eval : fletch_rvenv -> fletch_lenv -> fletch_exp ->
                     fletch_rvenv -> fletch_lenv -> fletch_exp -> Prop :=

  (* Computational rules *)

  | E_exp_compute_var_location : forall nu sigma l tau G,
              (tau, (G, l)) __in nu ->
              << nu, sigma, exp_var_location tau >> ---> << nu, sigma, exp_location l >>
  | E_exp_compute_null : forall nu sigma,
                           << nu, sigma, exp_var var_null >> --->
                           << nu, sigma, exp_location loc_null >>
  | E_exp_compute_runtime_assignment : forall nu sigma l nu' tau G l0 o,
                     (tau, (G, l0)) __in nu ->
                     update_rvenv nu tau (G, l) nu' ->
                     (l,o) __in sigma ->
                     Delta0 |S- typeof o <: type_nv G ->
                     << nu, sigma, exp_runtime_assignment tau (exp_location l) >> 
                       --->  
                     << nu', sigma, (exp_location l) >>
  | E_exp_compute_new : forall (nu : fletch_rvenv) nu' sigma sigma''' GL l_base
                               C FL ML sigma' sigma'' o RFL RML v_base,
                  
                  get_fields (atype_class C GL) FL -> (* Class + superclass field lookup *)
                  get_methods (atype_class C GL) ML -> (* Class + superclass method lookup *)
                  insert_lenv sigma empty_location (l_base, sigma') ->
                  insert_rvenv nu (atype_classtype (atype_class C GL), l_base)
                                   (v_base, nu') ->
                  bind_fields GL FL RFL ->
                  bind_methods v_base (atype_class C GL) sigma' ML (RML, sigma'') ->
                  o = class_instance_object (atype_class C GL) RFL RML ->
                  update_lenv sigma'' l_base o sigma''' ->
                  << nu, sigma, exp_new (atype_class C GL) >>
                  ---> 
                  << nu', sigma''', exp_location l_base >>
  | E_exp_compute_field_getter : forall nu sigma l1 l2 f N RFL RML T,
                     (l1, class_instance_object N RFL RML) __in sigma ->
                     (f, (T, l2)) _in RFL ->
                     << nu, sigma, exp_getter (exp_location l1) (field_property f) >> --->
                     << nu, sigma, exp_location l2 >>
  | E_exp_compute_method_getter : forall nu sigma
                                         l1 l2 m N FL ML,
                     (l1, class_instance_object N FL ML) __in sigma ->
                     (m, l2) _in ML ->
                     << nu, sigma, exp_getter (exp_location l1) (method_property m) >> 
                     --->
                     << nu, sigma, exp_location l2 >>
  | E_exp_compute_setter : forall nu sigma sigma' l1 f l2 l' N FL ML G o,
                 (l1, class_instance_object N FL ML) __in sigma ->
                 (f, (G, l')) _in FL ->
                 update_lenv sigma l1 (class_instance_object N
                                       (update_runtime_field FL (f, (G, l2)))
                                       ML)
                                      sigma' ->
                 (l2,o) __in sigma ->
                 Delta0 |S- typeof o <: type_nv G ->
                 << nu, sigma,
                 exp_setter (exp_location l1) (field_property f) (exp_location l2) >> --->
                 << nu, sigma', exp_location l2 >>
  | E_exp_compute_function_call : forall nu nu' sigma l T
                                         LL e' GxL e LLref tauL,
                (l, closure_object GxL T e) __in sigma ->
                Delta0 |- T OK ->
                actuals_to_formals nu GxL LLref (tauL, nu') ->
                locations LL LLref ->
                multiple_vloc_subst e (rvar_to_var GxL) tauL e' ->
                location_subtyping sigma LLref (runtime_types GxL) ->
                       << nu, sigma, exp_call (exp_location l)
                                                LL >> --->
                       << nu', sigma, exp_frame_expression T e' >>
  | E_exp_compute_closure_creation : forall nu sigma sigma' e GxL l T,
                                       well_formed_parameters GxL ->
                                       insert_lenv sigma
                                                   (closure_object
                                                      GxL
                                                      T
                                                      e)
                                                   (l,sigma') ->
                           << nu, sigma, exp_function GxL T e >>
                           --->
                           << nu, sigma', exp_location l >>
  | E_exp_compute_unfolded_call : forall nu sigma T l o,
                                    (l,o) __in sigma ->
                                    Delta0 |S- typeof o <: T ->
                                    << nu,
                                       sigma,
                                       exp_frame_expression T (exp_location l) >>
                                    --->
                                     << nu, sigma, (exp_location l) >>

  (* Congruence rules *)


  | E_exp_congruence_getter : forall nu nu' sigma sigma' e e' p,
                                << nu, sigma, e >> ---> << nu', sigma', e' >>
                                ->
                                << nu, sigma, exp_getter e p >> --->
                                << nu', sigma', exp_getter e' p >>
  | E_exp_congruence_LHS_setter : forall nu nu' sigma sigma' e1 e2 p e1',
                          << nu, sigma, e1 >> ---> << nu', sigma', e1' >> ->
                          << nu, sigma, exp_setter e1 p e2 >> --->
                          << nu', sigma', exp_setter e1' p e2 >>
  | E_exp_congruence_RHS_setter : forall nu nu' sigma sigma' e e' p l,
                            << nu, sigma, e >> ---> << nu', sigma', e' >> ->
                            << nu, sigma, exp_setter (exp_location l) p e >>
                            --->
                            << nu', sigma', exp_setter (exp_location l) p e' >>
  | E_exp_congruence_runtime_assignment : forall nu nu' sigma sigma' tau e e',
                               << nu, sigma, e >> ---> <<nu', sigma', e' >> ->
                               << nu, sigma, exp_runtime_assignment tau  e >> --->
                                << nu', sigma', exp_runtime_assignment tau e' >>
  | E_exp_congruence_function_call_args : forall nu nu' sigma sigma' args args' l,
                        [[ nu, sigma, args ]] --->^ [[ nu', sigma', args' ]] ->
                        << nu, sigma, exp_call (exp_location l) args >> --->
                        << nu', sigma', exp_call (exp_location l) args' >>
  | E_exp_congruence_function_call_callsite : forall nu nu' sigma sigma' e e' eL,
                        << nu, sigma, e >> ---> << nu', sigma', e' >> ->
                        << nu, sigma, exp_call e eL >> --->
                        << nu', sigma', exp_call e' eL >>
  | E_exp_congruence_unfolded_call : forall nu sigma nu' sigma' e e' T,
                        << nu, sigma, e >> ---> << nu', sigma', e' >> ->
                        << nu, sigma, exp_frame_expression T e >> --->
                        << nu', sigma', exp_frame_expression T e' >>
with
 exp_eval_list : fletch_rvenv -> fletch_lenv -> list fletch_exp ->
                 fletch_rvenv -> fletch_lenv -> list fletch_exp -> Prop :=
 | E_base : forall nu sigma nu' sigma' e e' eL,
             << nu, sigma, e >> ---> << nu', sigma', e' >> ->
             [[ nu, sigma, e :: eL ]] --->^ [[ nu', sigma', e' :: eL ]]
 | E_ind : forall nu sigma nu' sigma' e eL eL',
             [[ nu, sigma, eL ]] --->^ [[ nu', sigma', eL' ]] ->             
             [[ nu, sigma, e :: eL ]] --->^ [[ nu', sigma', e :: eL' ]]
where "'<<' nu ',' sigma ',' e '>>' '--->' '<<' nu' ',' sigma' ',' e' '>>'" :=
  (exp_eval nu sigma e nu' sigma' e')
and "'[[' nu ',' sigma ',' eL ']]' '--->^' '[[' nu' ',' sigma' ',' eL' ']]'" :=
  (exp_eval_list nu sigma eL nu' sigma' eL').

Hint Constructors exp_eval.
Hint Constructors exp_eval_list.

(* Mutual induction principle for the operational semantics. *)
Scheme exp_eval_mut := Induction for exp_eval Sort Prop
with exp_eval_list_mut := Induction for exp_eval_list Sort Prop.

(* 
 * Reflexive and transitive closure of the one step expression evaluation exp_eval.
 * exp_eval_star nu sigma e nu' sigma' e' IFF e' is a normal form.
 * This relation is used in the soundness theorem.
 *)
Reserved Notation "'<<' nu ',' sigma ',' e '>>' '||' '<<' nu' ',' sigma' ',' e' '>>'" (at level 39, sigma,nu,e,sigma',nu' at level 38).
Inductive exp_eval_star : fletch_rvenv -> fletch_lenv -> fletch_exp ->
                          fletch_rvenv -> fletch_lenv -> fletch_exp -> Prop :=
  | E_step_base : forall nu sigma e,
                    << nu, sigma, e >> || << nu, sigma, e >>
  | E_step_ind : forall nu sigma nu' nu'' sigma' sigma'' e e' e'',
                    << nu, sigma, e >> ---> << nu', sigma', e' >> ->
                    << nu', sigma', e' >> || << nu'', sigma'', e'' >> ->
                    << nu, sigma, e >> || << nu'', sigma'', e'' >>
where "'<<' nu ',' sigma ',' e '>>' '||' '<<' nu' ',' sigma' ',' e' '>>'" :=
  (exp_eval_star nu sigma e nu' sigma' e').

Definition irred (nu : fletch_rvenv) (sigma : fletch_lenv) (e : fletch_exp) :=
    forall e' nu' sigma', ~ << nu, sigma, e >> ---> << nu', sigma', e' >>.

(* 
 * Acceptable error configuration.
 * An error configuration represents a "null-pointer error" or a "subtype violation" error,
 * but it cannot be a "message-not-understood" error.
 *)
Inductive error : fletch_rvenv -> fletch_lenv -> fletch_exp -> Prop :=
  | Base_Err_null_getter : forall nu sigma p,
                        error nu sigma (exp_getter (exp_location loc_null) p)
  | Base_Err_null_setter : forall nu sigma p l,
                        error nu sigma (exp_setter (exp_location loc_null) p
                                                   (exp_location l))
  | Base_Err_Subtyping : forall nu sigma l1 l2 o1 N FL ML T G l' f,
                                      (l1, o1) __in sigma ->
                                      o1 = class_instance_object N FL ML ->
                                      (f, (G, l')) _in FL ->
                                      nu ; sigma ; Delta0 ; Gamma0 |- (exp_location l2) : T ->
                                      ~ Delta0 |S- T <: (type_nv G) ->
                                      error nu
                                            sigma
                                            (exp_setter (exp_location l1)
                                                        (field_property f)
                                                        (exp_location l2))
  | Base_Err_call_subtyping : forall nu sigma l lL LL GL o T,
                           locations lL LL ->
                           (l,o) __in sigma ->
                           typeof o = (type_nv ((GL) --> T)) ->                           
                           ~ (location_subtyping sigma LL GL) ->
                           error nu sigma (exp_call (exp_location l) lL)
  | Base_Err_assignment_subtyping : forall nu sigma tau T G l,
                                  nu ; sigma ; Delta0 ; Gamma0 |- (exp_location l) : T ->
                                  (tau, (G, l)) __in nu ->
                                  ~ Delta0 |S- T <: atype_nv G ->
                                  error nu sigma (exp_runtime_assignment tau (exp_location l))
  | Base_Err_unfolded_subtyping : forall nu sigma T l o,
                                       (l,o) __in sigma ->
                                       ~ Delta0 |S- typeof o <: T ->
                                       error nu sigma (exp_frame_expression T (exp_location l))
  | Ind_Err_getter : forall nu sigma p e,
                       error nu sigma e ->
                       error nu sigma (exp_getter e p)
  | Ind_Err_setter_left : forall nu sigma p e1 e2,
                       error nu sigma e1 ->
                       error nu sigma (exp_setter e1 p e2)
  | Ind_Err_setter_right : forall nu sigma p e1 e2,
                       error nu sigma e2 ->
                       error nu sigma (exp_setter e1 p e2)
  | Ind_Err_call_function : forall nu sigma e eL,
                     error nu sigma e ->
                     error nu sigma (exp_call e eL)
  | Ind_Err_call_args : forall nu sigma e eL,
                     error_list nu sigma eL ->
                     error nu sigma (exp_call e eL)
  | Ind_Err_assignment : forall nu sigma tau e,
                           error nu sigma e ->
                           error nu sigma (exp_runtime_assignment tau e)
  | Ind_Err_frame : forall nu sigma e T,
                      error nu sigma e ->
                      error nu sigma (exp_frame_expression T e)
with
 error_list : fletch_rvenv -> fletch_lenv -> list fletch_exp -> Prop :=
  | EL_base : forall nu sigma e eL,
               error nu sigma e ->
               error_list nu sigma (e :: eL)
  | EL_ind : forall nu sigma e eL,
              error_list nu sigma eL ->
              error_list nu sigma (e :: eL).

(* Mutual induction principle for error configurations. *)
Scheme error_mut := Induction for error Sort Prop
with error_list_mut := Induction for error_list Sort Prop.

(*It is always possible to update a store environment with a fresh location. *)
Lemma lenv_update_holds : forall sigma l o,
                            (exists o, (l,o) __in sigma) ->
                            exists sigma', update_lenv sigma l o sigma'.
Proof.
  intros. destruct H. destruct sigma. induction f. inversion H.
  assert(X := in_inv H). destruct X. exists ((l,o) :: f, g). rewrite -> H0. constructor.
  constructor. assert(Y := IHf H0). destruct Y. destruct a.
  emdestruct EQ NEQ (location_em l g0). destruct x0.
  subst. exists ((g0, o) :: f, g). constructor. constructor. destruct x0.
  exists ((g0,f0) :: f1, g).  constructor. constructor. trivial. inversion H1. auto.
Qed.

(* Minor lemmas on store environments and objects *)

Lemma lenv_in : forall l (sigma : fletch_lenv),
                  (exists o, (l,o) __in sigma)
                  \/ (forall o, (l,o) not __in sigma).
Proof.
  intros. assert(EM := excluded_middle (exists o, (l,o) __in sigma)). destruct EM.
  auto. right. intros o F. intuition. destruct H. exists o. trivial.
Qed.

Lemma obj_structure : forall o,
                        (exists N FL ML, o = class_instance_object N FL ML) \/
                        (forall N FL ML, o <> class_instance_object N FL ML).
Proof.
  intros. assert(EM := excluded_middle (exists N FL ML, o = class_instance_object N FL ML)).
  destruct EM. auto. right. intros. intuition. destruct H. exists N,FL,ML. trivial.
Qed.

Lemma closure_stucture : forall o,
                           (exists GxL T e,
                              o = closure_object GxL T e) \/
                           (forall GxL T e,
                              o <> closure_object GxL T e).
Proof.
  intros. assert(EM := excluded_middle (exists GxL T e, o = closure_object GxL T e)).
  destruct EM. auto. right. intros. intuition. destruct H. exists GxL,T,e. trivial.
Qed.

Lemma field_el : forall f (FL : list runtime_field),
                  (exists G l, (f,(G,l)) _in FL) \/
                  (forall G l, (f,(G,l)) not _in FL).
Proof.
  intros. assert(EM := excluded_middle (exists G l, ((f,(G,l)) _in FL))).
  destruct EM. auto. right. intros. intuition. destruct H. exists G,l. trivial.
Qed.

Lemma method_el : forall m (ML : list runtime_method),
                  (exists l, (m,l) _in ML) \/
                  (forall l, (m,l) not _in ML).
Proof.
  intros. assert(EM := excluded_middle (exists l : generate_id, (m, l) _in ML)).
  destruct EM. auto. right. intros. intuition. destruct H. exists l. trivial.
Qed.

Lemma lenv_in_obj : forall l (sigma : fletch_lenv),
                      (exists o, (l,o) __in sigma) \/
                      (forall o, (l,o) not __in sigma).
Proof.
  intros. assert(EM := excluded_middle (exists o : fletch_runtime_object, (l, o) __in sigma)).
  destruct EM. auto. right. intros. intuition. destruct H. exists o. trivial.
Qed.

Lemma object_subtyping : forall o G,
                           Delta0 |S- typeof o <: type_nv G \/
                           ~ Delta0 |S- typeof o <: type_nv G.
Proof.
  intros. assert(EM := excluded_middle (Delta0 |S- typeof o <: type_nv G)). eauto.
Qed.

Lemma sem_subtyping_em : forall Delta T S,
                       Delta |S- T <: S \/ ~ Delta |S- T <: S.
Proof.
  intros. assert(EM := excluded_middle (Delta |S- T <: S)). auto.
Qed.

(* Excluded middle lemmas. *)

Lemma locationb_em : forall (l : @generate_id location_id) l',
                       l = l' \/ l <> l'.
Proof.
  intros. assert(EM := excluded_middle (l = l')). auto.
Qed.

Lemma location_subtyping_em : forall sigma LL TL,
                                location_subtyping sigma LL TL \/
                              ~ location_subtyping sigma LL TL.
Proof.
  intros. assert(EM := excluded_middle (location_subtyping sigma LL TL)). auto.            
Qed.

Lemma exp_location_em : forall e,
                       (exists l, e = exp_location l) \/
                       (forall l, e <> exp_location l).
Proof.
  intros. assert(X := excluded_middle (exists l, e = exp_location l)). destruct X.
  auto. right. intuition. destruct H. exists l. auto.
Qed.

Lemma object_em : forall (o1 : fletch_runtime_object) o2,
                    o1 = o2 \/ o1 <> o2.
Proof. 
  intros. assert(EM := excluded_middle (o1 = o2)). auto.
Qed.

End StandardCheckModeOS.