Require Export fletch_well_formedness.

Module SubLemmas (C : ClassTable).
Module WellFormednessImpl := WellFormedness(C).
Export WellFormednessImpl.

(* A simple notion of well-formed environments without cyclic dependencies *)

Reserved Notation "'|-t' Delta 'OK'" (at level 40, Delta at level 40).
Inductive fletch_tenv_ok : fletch_tenv -> Prop :=
| fto_nil : fletch_tenv_ok []
| fto_cons : forall Delta N X,
  |-t Delta OK ->
  Delta |- type_nv (type_classtype N) OK ->
  |-t ((X,N) :: Delta) OK
where "'|-t' Delta 'OK'" := (fletch_tenv_ok Delta).

Reserved Notation "Delta '|-v' Gamma 'OK'" (at level 40, Gamma at level 40).
Inductive fletch_venv_ok : fletch_tenv -> fletch_venv -> Prop :=
| fvo_nil : forall Delta,
  fletch_venv_ok Delta []
| fvo_cons : forall Delta Gamma G x,
  Delta |-v Gamma OK ->
  Delta |- (type_nv G) OK ->
  Delta |-v ((x,G) :: Gamma) OK
where "Delta '|-v' Gamma 'OK'" := (fletch_venv_ok Delta Gamma).

(* 
 * If Delta is well formed, then each (X,G) in Delta, where G is
 * a type (defining the upper bound for the type parameter X)
 * is well formed. It trivially follows from the definition
 * of well formedness for Delta.
 *)
Fact type_in_wellformed_venv_ok : forall Delta Gamma X G,
  Delta |-v Gamma OK ->
  (X,G) _in Gamma ->
  Delta |- (type_nv G) OK.
Admitted.

(*
 * If the environment Delta is well formed and the type T is well formed,
 * the type S obtainied by replacing all the type parameters occurring
 * in T with their bound in Delta is well formed. As above, it trivially
 * follows from Delta and type well formedness definition.
 *)
Fact bound_delta_preserves_wellformedness : forall Delta T S,
  |-t Delta OK ->
  Delta |- T OK ->
  bound_delta Delta T S ->
  Delta |- S OK.
Proof.
Admitted.

(* 
 * Similar to the lemma above.
 *)
Fact getter_preserves_wellformedness : forall Delta T p G,
  |-t Delta OK ->
  Delta |- T OK ->
  getter T p G ->
  Delta |- (type_nv G) OK.
Proof.
Admitted.

(* 
 * Given a list of types GL, if any type G in GL is well formed, then
 * the type list GL is well formed too.
 *)
Fact type_ok_in_list_implies_type_ok_list : forall Delta GL,
  (forall G, G _in GL -> Delta |- (type_nv G) OK) ->
  Delta |-^ GL OK.
Proof.
  intros Delta GL. induction GL; intro H; constructor.
  apply H. left. reflexivity.
  apply IHGL. intros G Hin. apply H. right. assumption.
Qed.

(* 
 * For any (l, o) in sigma, where l is a location and o is an object,
 * if the type of the object o is well formed.
 *)
Fact heap_object_type_ok : forall l o (sigma : fletch_lenv) T Delta,
  (* EE: we will need to know something like '.. |-heap sigma OK' *)
  (l, o) __in sigma ->
  typeof o = T ->
  Delta |- T OK.
Proof.
Admitted.

(* 
 * Similar to the above lemma but with variable locations.
 *)
Fact variable_type_ok : forall Delta tau G l (nu : fletch_rvenv),
  (* EE: will need to know something like '.. |-venv nu OK *)
  |-t Delta OK ->
  (tau, (G, l)) __in nu ->
  Delta |- (type_nv G) OK.
Proof.
Admitted.

(* 
 * If Delta is well formed, Gamma is well formed under Delta, and the
 * expression e is well typed with typed T under the environments
 * nu, sigma, Delta, and Gamma, then T is well formed.
 *)
Lemma well_typed_well_formed : forall nu sigma Delta Gamma e T,
  |-t Delta OK ->
  Delta |-v Gamma OK ->
  nu ; sigma ; Delta ; Gamma |- e : T ->
  Delta |- T OK.
Proof.
  intros nu sigma Delta Gamma e T HtenvOK HvenvOK Htyp. revert HtenvOK HvenvOK.
  induction Htyp using exp_typing_mut with
  (P := fun nu sigma Delta Gamma e T (Htyp : nu;sigma;Delta;Gamma |- e : T) =>
          |-t Delta OK ->
          Delta |-v Gamma OK ->
          Delta |- T OK)
  (P0 := fun nu sigma Delta Gamma el TL (Htyps : nu;sigma;Delta;Gamma |-^ el : TL) =>
           True);
  try ((simpl; discriminate) || auto); intros HtenvOK HvenvOK.
  (* EE: trying to recognize the cases -- not yet correct *)
  - Case "T_exp_var".
    apply (type_in_wellformed_venv_ok _ Gamma y); assumption.
  - Case "T_exp_getter".
    assert (HTok := IHHtyp HtenvOK HvenvOK).
    assert (HT'ok: Delta |- T' OK).
      apply bound_delta_preserves_wellformedness with (T:=T); auto.
    cut (Delta |- type_nv ([] --> type_nv G) OK).
      intros Hclok. inversion Hclok. assumption.
    apply getter_preserves_wellformedness with (T:=T') (p:=p); auto.
  (* Case "T_exp_setter" *)
  (* Case "T_exp_assign" *)
  - Case "T_exp_call".
    cut (Delta |- type_nv (GL --> T) OK).
      intro Hok. inversion Hok. assumption.
    apply IHHtyp; auto.
  (* Case "T_exp_dynamic_call" *)
  (* Case "T_exp_new" *)
  - Case "T_exp_function".
    constructor.
      destruct GL; constructor. admit.
        (* apply t. left. reflexivity. *)
      apply type_ok_in_list_implies_type_ok_list.
      intros G Hin. admit. (* apply t. right. assumption. *)
    apply IHHtyp.
      assumption.
    admit. (* TODO: study fold_pred, see what we can know about Gamma'' *)
  - Case "T_exp_location".
    apply (heap_object_type_ok l o sigma); assumption.
  - Case "T_exp_var_location".
    apply (variable_type_ok _ tau _ l nu); assumption.
  - Case "T_exp_unfolded_call".
  (* TODO: 'exp_frame_expression T e' admits an arbitrary type T, *)
  (*          so we need to introduce something that enforces 'type_ok T', *)
  (*          maybe by requiring it at each step.. *)
     admit.
Qed.

(* 
 * Just Coq machinery to show that if a list of expression is a list of
 * values, then every value is a location.
 *)
Lemma values_are_locations : forall e,
                               values e ->
                               exists LL, locations e LL.
Proof.
  intros.
  induction H. exists []. auto. destruct IHvalues. destruct H. subst. exists (x0 :: x).
  eauto.
Qed.

(* Quick, but mainly obvious lemmas *)

(* 
 * Just Coq machinery to show that for all parameters
 * (G, x) in the set of parameters GL, a set of locations L in the set of 
 * locations LL, it is always possible to insert (tau, L) of type G
 * in the variable environment nu.
 *)
Lemma actuals_to_formals_exists : forall nu GL LL,
                                    exists nu' tauL,
                                      actuals_to_formals nu GL LL (tauL, nu').
Admitted.

(*
 * Just Coq machinery to show that it is always possible to replace any 
 * occurrence of a variable x with a variable location in an expression e.
 *)
Lemma variable_substitution_exists : forall e GL tauL,
  exists e', multiple_vloc_subst e (rvar_to_var GL) tauL e'.
Admitted.

(* FULLY ANNOTATED TYPING DERIVATION *)

(* 
 * If T is a fully annoatated type, then the type of the property p in T
 * is fully annotated too. It trivially hold since we consider only
 * fully programs, therefore with classes declaring fields and methods
 * with fully annotated types only.
 *)
Lemma fully_annotated_getter : forall T G p,
                                 fully_annotated_type T ->
                                 getter T p ([] --> type_nv G) ->
                                 fully_annotated_type (type_nv ([] --> type_nv G)).
Admitted.

(* 
 * Just Coq machinery.
 *)
Lemma fully_annotated_closure_type : forall GxL,
   (forall G l, (G, l) _in GxL -> fully_annotated_type (atype_nv G)) ->
   fully_annotated_type_list (ltype_nv (projection GxL)).
Admitted.

(* 
 * Just Coq machinery.
 *)
Lemma fully_annotated_closure_rtype : forall l0,
  (forall (G : fletch_anvtype) (x : generate_id),
    (G, x) _in l0 -> fully_annotated_type (type_nv G)) ->
   fully_annotated_type_list (ltype_nv (runtime_types l0)).
Admitted.

(* WELL FORMEDNESS *)

(* 
 * If an object is well formed under nu and sigma, inserting such object
 * in sigma, generating the new environment sigma', still keeps
 * the object well formed in nu and sigma'.
 *)
Lemma lenv_insert_preserves_well_formed_object :
  forall nu sigma o sigma' l,
          well_formed_object nu sigma o ->
          insert_lenv sigma o (l, sigma') ->
          well_formed_object nu sigma' o.
Admitted.

(* SMALL INVOLVING ENVIRONMENTS *)

(* For each environment env defined in fletch, if (k, v) is not in env, then
 * let env' be env updated with the mapping (k', v'), where k' is not k, then
 * (k, v) is not in env'.
 *)
Lemma update_fresh_not_add {KEY:Set}{VALUE:Set} :
  forall k v env env' k' v',
    (k, v) not __in env ->
    (@update_fresh KEY VALUE) env k' v' env' ->
    k' <> k ->
    (k, v) not __in env'.
Proof.
  intros.
  inversion H0.
  subst. unfold BLIn in H.
  assert(X := (@update_not_add KEY VALUE) k v env0 env'0 k' v' H H2 H1).
  unfold BLIn. exact X.
Qed.

(* 
 * Given a well formed environment, inserting a mapping (k, v) in such environment,
 * where k is fresh, preserves environment well formedness.
 *)
Lemma insert_fresh_well_formed {KEY:Set}{VALUE:Set}{NFRESH : KEY -> KEY} :
  forall (v : VALUE) env env' fresh,
         (@well_formed_fresh_env KEY VALUE) env ->
         (@insert_fresh KEY VALUE NFRESH) env v (fresh, env') ->
         well_formed_fresh_env env'.
Proof.
    intros v env env' k WFenv Insert.
    inversion Insert. econstructor. inversion WFenv. subst. inversion H5.
    rewrite -> H0 in H4.
    exact((@insert_well_formed KEY VALUE) (NFRESH k0) v env0 env'0 H4 H2).
Qed.

(* 
 * Given a well formed environment env, updating env with the mapping (k, v),
 * where k is in the domain of env, preserves environment well formedness.
 *)     
Lemma update_fresh_well_formed {KEY:Set}{VALUE:Set} :
  forall (k : KEY) (v : VALUE) env env',
         well_formed_fresh_env env ->
         update_fresh env k v env' ->
         well_formed_fresh_env env'.
Proof.
  intros k v env env' WFenv Update.
  inversion Update. econstructor. inversion WFenv. subst. inversion H5. rewrite -> H1 in H4.
  inversion Update. subst.
  exact((@update_well_formed KEY VALUE) k v env0 env'0 H4 H8).
Qed.

(* Simple correctness lemma. *)
Lemma insert_fresh_correctness {KEY : Set}{VALUE : Set}{NFRESH : KEY -> KEY} :
  forall (k : KEY) (v : VALUE) env env',
    (@insert_fresh KEY VALUE NFRESH) env v (k,env') ->
    (k,v) __in env'.
Proof.
  intros k v env env' Insert.
  inversion Insert. subst.
  unfold BLIn.
  exact((@insert_correctness KEY VALUE) (NFRESH k0) v env0 env'0 H2).
Qed.

(* Simple correctness lemma. *)
Lemma update_fresh_correctness {KEY : Set}{VALUE : Set} :
  forall (k : KEY) (v : VALUE) env env',
    update_fresh env k v env' ->
    (k,v) __in env'.
Proof.
  intros k v env env' Update.
  inversion Update. subst.
  simpl.
  exact((@update_correctness KEY VALUE) k v env0 env'0 H).
Qed.

(* It is possible to update environments with fresh values. *)
Lemma updatable_fresh {KEY : Set}{VALUE : Set} :
  forall (k : KEY) (v1 v2 : VALUE) env,
    (k, v1) __in env ->
    exists env', update_fresh env k v2 env'.
Proof.
   intros k v1 v2 env Exists. destruct env.
   assert(X := updatable k v1 v2 f Exists). destruct X. exists (x, k0). constructor.
   auto.
Qed.

(* It is possible to insert fresh values into any environment. *)
Lemma insert_fresh_add {KEY : Set}{VALUE : Set}{NFRESH : KEY -> KEY} :
  forall (k : KEY) k' (v : VALUE) v' env env',
    (k,v) __in env ->
    (@insert_fresh KEY VALUE NFRESH) env v' (k',env') ->
    (k,v) __in env'.
Proof.
  intros.
    sinversion H0. assert(X := (@insert_add KEY VALUE) k (NFRESH k0) v v' env0 env'0
                               H H4). 
    auto.
Qed.

(* Insert preserves variable environment well formedness. *)
Lemma insert_renv_well_formed :
  forall nu nu' n l,
         well_formed_rvenv nu ->
         insert_rvenv nu l (n,nu') ->
         well_formed_rvenv nu'.
Proof.
  intros nu nu' n l WFRvenv Insert.
  unfold well_formed_rvenv.
  exact(insert_fresh_well_formed l nu nu' n WFRvenv Insert).
Qed.

(* Update preserves variable environment well formedness. *)
Lemma update_rvenv_well_formed :
  forall tau l nu nu',
         well_formed_rvenv nu ->
         update_rvenv nu tau l nu' ->
         well_formed_rvenv nu'.
Proof.
  intros tau l nu nu' WFRvenv Update.
  exact(update_fresh_well_formed tau l nu nu' WFRvenv Update).
Qed.

(* Insert preserves store environment well formedness. *)
Lemma insert_lenv_well_formed :
  forall o sigma sigma' n,
         well_formed_lenv sigma ->
         insert_lenv sigma o (n,sigma') ->
         well_formed_lenv sigma'.
Proof.
  intros o sigma sigma' n WFLenv Insert.
  unfold well_formed_lenv.
  exact(insert_fresh_well_formed o sigma sigma' n WFLenv Insert).
Qed.

(* Update preserves store environment well formedness. *)
Lemma update_lenv_well_formed :
  forall l o sigma sigma',
         well_formed_lenv sigma ->
         update_lenv sigma l o sigma' ->
         well_formed_lenv sigma'.
Proof.
  intros l o sigma sigma' WFLenv Update.
  exact(update_fresh_well_formed l o sigma sigma' WFLenv Update).
Qed.

(* Update preserves variable environment well formedness. *)
Lemma update_rvenv_correctness :
  forall tau o nu nu',
    update_rvenv nu tau o nu' ->
    (tau, o) __in nu'.
Proof.
  intros. assert(X := update_fresh_correctness tau o nu nu' H). auto.
Qed.

(* Store environment correctness lemma.  *)
Lemma update_lenv_correctness :
  forall (l : generate_id) (o : fletch_runtime_object) sigma sigma',
    update_lenv sigma l o sigma' ->
    (l,o) __in sigma'.
Proof.
  intros. assert(X := update_fresh_correctness l o sigma sigma' H). auto.
Qed.

(* 
 * Just Coq machinery to say that it is always possible to insert a new object in
 * a store environment.
 *)
Lemma insertable_lenv : forall o sigma,
                         exists sigma', insert_lenv sigma o sigma'.
Proof.
  intros. destruct sigma.
  assert(X := insertable (generate_fresh_location g) o f).
  (* We assume that we always bind a new object into fresh locations *)
  assert(Assumption : (forall v : fletch_runtime_object,
       (generate_fresh_location g, v) not _in f)).
  admit.
  assert(Y := X Assumption). destruct Y as [ env' Insert ]. 
  exists ((generate_fresh_location g), (env', (generate_fresh_location g))). constructor.
  auto.
Qed.

(* 
 * Just Coq machinery to say that it is always possible to insert a new binding from
 * a fresh variable location to a location in a variable environment. 
 *)
Lemma insertable_rvenv : forall l G nu,
                         exists nu' tau, insert_rvenv nu (G, l) (tau, nu').
Proof.
  intros. destruct nu.
  assert(X := insertable (generate_fresh_variable_location g) (G, l) f).
  (* We assume that we always generate fresh locations *)
  assert(Assumption : (forall v,
       (generate_fresh_variable_location g, v) not _in f)).
  admit.
  assert(Y := X Assumption). destruct Y as [ env' Insert ].
  exists (env', (generate_fresh_variable_location g)). econstructor.
  constructor.
  auto.
Qed.

(* Just Coq machinery to say that it is always possible to update a store environment. *)
Lemma updatable_lenv : forall l o1 o2 sigma,
                         (l, o1) __in sigma ->
                         exists sigma', update_lenv sigma l o2 sigma'.
Proof.
  intros. assert(X := updatable_fresh l o1 o2 sigma H). auto.
Qed.

(* Just Coq machinery to say that  *)
Lemma update_fresh_not_noise {KEY : Set}{VALUE : Set} :
  forall (k1 : KEY) k2 v1 (v2 : VALUE) env env',
                           well_formed_fresh_env env ->
                           k1 <> k2 ->
                           (k2, v1) __in env ->
                           update_fresh env k1 v2 env' ->
                           (k2, v1) __in env'.
Proof.
  intros. sinversion H2. sinversion H.
  assert(X := update_not_noise k1 k2 v1 v2 env0 env'0 H5 H0 H1 H3). auto.
Qed.

(* Store environmenht insert is monotonic.  *)
Lemma update_lenv_not_noise : forall l1 l2 sigma sigma' o o',
                           well_formed_lenv sigma ->
                           l1 <> l2 ->
                           (l2, o) __in sigma ->
                           update_lenv sigma l1 o' sigma' ->
                           (l2, o) __in sigma'.
Proof.
  intros. eauto using update_fresh_not_noise.
Qed.

(* Store environmenht update is monotonic.  *)
Lemma update_rvenv_not_noise : forall tau1 tau2 nu nu' o o',
                           well_formed_rvenv nu ->
                           tau1 <> tau2 ->
                           (tau2, o) __in nu ->
                           update_rvenv nu tau1 o' nu' ->
                           (tau2, o) __in nu'.
Proof.
  intros. eauto using update_fresh_not_noise.
Qed.

(* Excluded middle: (k,v) is in the environment env, or it is not. *)
Lemma fresh_env_excluded_middle {KEY : Set}{VALUE : Set} : forall
                                                             (env : @fletch_fresh_env KEY VALUE)
                                                             k v,
                                                       (forall x, (k,v) = x \/ ~ (k,v) = x) ->
                                                       (k, v) __in env \/
                                                       (k, v) not __in env.
Proof.
  intros. destruct env. assert(X := @env_excluded_middle KEY VALUE f k v).
  destruct X. left. simpl. trivial. right. simpl. trivial.
Qed.

(* Just Coq machinery to say that the fresh key generator actually generates fresh keys. *)
Lemma generate_fresh_key_correctness {K : id_kind}{VALUE : Set} :
  forall (k : (@generate_id K)) (v : VALUE) senv env',
    (@insert_fresh _ _ generate_fresh_key) (senv, k) v ((generate_fresh_key k), env') ->
    (forall v',
      ((generate_fresh_key k), v') not __in (senv, k)).
Admitted.

End SubLemmas.
