(* --------------------------------------------- Fletch syntax --------------------------------------------- *)

Require Export Arith.
Require Export fletch_types.

Module FletchSyntax.
Export FletchTypes.

(* Method/function parameter: a pair (type annotations, parameter id) *)
Definition param := (fletch_anvtype * @generate_id var_id)%type.
(* Parameter list definition *)
Definition params := list param.

(* 
 * Fletch variable definition. A fletch variable can be a declared variable identifier,
 * the this literal or the null literal.
 *)
Inductive fletch_var : Set :=
  | var_user : @generate_id var_id -> fletch_var
  | var_this : fletch_var
  | var_null : fletch_var.

(* The following functions allow comparison between variables *)

Definition var_id_compare (v1 v2 : (@generate_id var_id)) : {v1 = v2} + {v1 <> v2}. 
  destruct v1,v2.
  destruct (eq_nat_dec n n0).
  rewrite e.
  left. reflexivity.
  right. intro H. injection H. assumption.
Defined.

Definition var_compare (v1 v2 : fletch_var) : {v1 = v2} + {v1 <> v2}. 
  destruct v1,v2.
  destruct g. destruct g0.
  destruct (eq_nat_dec n n0).
  rewrite e.
  left. reflexivity.
  right. intro H. injection H. assumption.
  right. intro H. inversion H.
  right. intro H. inversion H.
  right. intro H. inversion H.
  left. reflexivity.
  right. intro H. inversion H.
  right. intro H. inversion H.
  right. intro H. inversion H.
  left. reflexivity.
Defined.

Definition type_var_id_compare (v1 v2 : (@generate_id type_var_id)) : {v1 = v2} + {v1 <> v2}. 
  destruct v1,v2.
  destruct (eq_nat_dec n n0).
  rewrite e.
  left. reflexivity.
  right. intro H. injection H. assumption.
Defined.

Definition field_id_compare (f1 f2 : (@generate_id field_id)) : {f1 = f2} + {f1 <> f2}. 
  destruct f1,f2.
  destruct (eq_nat_dec n n0).
  rewrite e.
  left. reflexivity.
  right. intro H. injection H. assumption.
Defined. 

Definition location_id_compare (l1 l2 : (@generate_id location_id)) : {l1 = l2} + {l1 <> l2}. 
  destruct l1,l2.
  destruct (eq_nat_dec n n0).
  rewrite e.
  left. reflexivity.
  right. intro H. injection H. assumption.
Defined.

Definition type_var_compare (v1 v2 : fletch_var) : {v1 = v2} + {v1 <> v2}.
  destruct v1 ; destruct v2 ; try (right ; intro ; inversion H ; fail).
  destruct g, g0.
  destruct (eq_nat_dec n n0).
    left ; rewrite e ; auto.
    right ; intro H ; injection H ; assumption.
  left ; reflexivity.
  left ; reflexivity.
Defined.

(* 
 * A property is either a field or a method name. 
 * Note that the namespaces for field names and method names
 * is disjoing.
 *)
Inductive property : Set :=
  | field_property : (@generate_id field_id) -> property
  | method_property : (@generate_id method_id) -> property.

(* Fletch epxressions. *)
Inductive fletch_exp : Set :=
  | exp_var : fletch_var -> fletch_exp
  | exp_getter : fletch_exp -> property -> fletch_exp
  | exp_setter : fletch_exp -> property -> fletch_exp -> fletch_exp
  | exp_assignment : (@generate_id var_id) -> fletch_exp -> fletch_exp
  | exp_call : fletch_exp -> list fletch_exp -> fletch_exp
  | exp_new : fletch_aclasstype -> fletch_exp
  | exp_function : params -> fletch_type -> fletch_exp -> fletch_exp
  (* Runtime semantics syntax *)
  | exp_location : (@generate_id location_id) -> fletch_exp
  | exp_var_location : (@generate_id var_location_id) -> fletch_exp
  | exp_runtime_assignment : (@generate_id var_location_id) -> fletch_exp -> fletch_exp
  | exp_frame_expression : fletch_type -> fletch_exp ->
                           fletch_exp.

(* 
 * Induction principle for fletch expressions. Coq is not able to automatically generate 
 * the following more powerful induction principle.
 *)
Lemma fletch_exp_induction
     : forall P : fletch_exp -> Prop,
       (* var *)
       (forall x : fletch_var, P (exp_var x)) ->
       (* getter *)
       (forall e : fletch_exp,
        P e -> forall p : property, P (exp_getter e p)) ->
       (* setter *)
       (forall e1: fletch_exp,
        P e1 ->
        forall (p : property) (e2 : fletch_exp),
        P e2 -> P (exp_setter e1 p e2)) ->
       (* assignment *)
       (forall (x : generate_id) (e : fletch_exp),
        P e -> P (exp_assignment x e)) ->
       (* call *)
       (forall (e : fletch_exp) (args : list fletch_exp),
        P e -> (forall arg, arg _in args -> P arg) ->
        P (exp_call e args)) ->
       (* new *)
       (forall N : fletch_aclasstype, P (exp_new N)) ->
       (* function *)
       (forall (p : params) (T : fletch_type) (e : fletch_exp),
        P e -> P (exp_function p T e)) ->
       (* location  *)
       (forall l : generate_id, P (exp_location l)) ->
       (* var location  *)
       (forall tau : generate_id, P (exp_var_location tau)) ->
       (* runtime assignment  *)
       (forall (tau : generate_id) (e : fletch_exp),
        P e -> P (exp_runtime_assignment tau e)) ->
       (* unfolded call  *)
       (forall (T : fletch_type) (e : fletch_exp),
        P e -> P (exp_frame_expression T e)) ->
       (* error *)
       forall e : fletch_exp, P e.
Proof.
  intros;revert e.
  refine (fix IHe (e : fletch_exp) : P e := _).
  destruct e.
    apply H.
    apply H0, IHe.
    apply H1; apply IHe.
    apply H2; apply IHe.
    apply H3; [apply IHe |].
      induction l.
        contradiction.
        intros arg [HEq | HIn].
          rewrite <- HEq. apply IHe.
          apply IHl, HIn.
    apply H4.
    apply H5, IHe.
    apply H6.
    apply H7.
    apply H8, IHe.
    apply H9, IHe.
Qed.

(* Fletch value definitions. Values can only be locations in fletch. *)
Definition value (e : fletch_exp) :=
  exists l, e = exp_location l.

(* Let LL be a list of locations, values LL holds IFF each expression l in LL is a location *)
Inductive values : list fletch_exp -> Prop :=
  | vals_base : values []
  | vals_ind : forall e eL,
                   value e ->
                   values eL ->
                   values (e :: eL).

(* Method declaration in Fletch. *)
Inductive fletch_method : Set :=
  | method : fletch_atype -> (@generate_id method_id) -> params -> fletch_exp -> fletch_method.

(* Field declaration in Fletch. *)
Inductive fletch_field : Set :=
  | field : fletch_anvtype -> (@generate_id field_id) -> fletch_field.

(* Generic prameter definition, i.e., a list of pairs (type parameter, annotated type) *)
Definition generic := ((@generate_id type_var_id) * fletch_aclasstype)%type.
(* List of generic parameters. *)
Definition generics := list generic.

(* Class declaration in Fletch. *)
Inductive fletch_class : Set :=
  | class : (@generate_id class_id) -> generics -> fletch_aclasstype -> list fletch_field -> list fletch_method -> fletch_class.

(* Just Coq machinery to provide more additional support for lists than in the standard Coq list library. *)
Lemma list_containement {A : Type} : forall x (y : A) L,
                            x <> y ->
                            x _in (y :: L) ->
                            x _in L.
Proof.
  intros. induction H0. intuition. symmetry in H0.  assert(X := H H0). contradiction.
  auto.
Qed.

(* Excluded middle lemmas *)

Lemma expression_em : forall (e : fletch_exp) e',
                       e = e' \/ e <> e'.
Proof.
  intros. assert(EM := excluded_middle (e = e')). auto.
Qed.

Lemma location_em {Id : id_kind} : forall (l : @generate_id Id) l',
                       l = l' \/ l <> l'.
Proof.
  intros. assert(EM := excluded_middle (l = l')). auto.
Qed.

End FletchSyntax.