(* ------------------------------------------  Fletch type system ----------------------------------------- *)

Require Export fletch_syntax.
Require Export fletch_types.
Require Export fletch_ct.
Require Export fletch_env.

Module StandardTS (C : ClassTable).
Export C.
Export FletchEnv.

(* Coercions *) 

Definition convert_id {A:id_kind}{B:id_kind} (x : @generate_id A) : @generate_id B :=
  match x with
    | id y => id y
  end.

Definition convert_anvtype_list (GL : list fletch_anvtype) : list fletch_nvtype :=
  map (fun G : fletch_anvtype => G : fletch_nvtype) GL.

(* SUBTYPING  *)

(* The following functions allow to replace occurrences of a type parameter X with a type G in a type T. *)
Fixpoint type_substitution (X : (@generate_id type_var_id)) (G : fletch_nvtype) (T : fletch_type) : fletch_type :=
  match T with
    | type_nv G' => type_nv (nvtype_substitution X G G')
    | type_void => type_void
  end
with
 nvtype_substitution (X : (@generate_id type_var_id)) (G : fletch_nvtype) (G' : fletch_nvtype)  : fletch_nvtype :=
  match G' with
    | type_var Y =>
        if type_var_id_compare X Y then G
        else G'
    | (GL) --> T => (map (nvtype_substitution X G) GL) -->
                                           type_substitution X G T
    | type_classtype (C <<<GL>>>) => (type_classtype (C<<<(map (fun A => (nvtype_substitution X G A)) GL)>>>))
    | type_dynamic => type_dynamic
    | type_bottom => type_bottom
  end.

Fixpoint classtype_substitution (X : (@generate_id type_var_id))
                                (G : fletch_nvtype)
                                (N : fletch_classtype) (*{struct N}*) : fletch_classtype :=
  match N with
    | C<<<GL>>> => C<<<(map (fun A => nvtype_substitution X G A) GL)>>>
  end.

Fixpoint atype_substitution (X : (@generate_id type_var_id)) (G : fletch_anvtype)
                            (T : fletch_atype) : fletch_atype :=
  match T with
    | atype_nv G' => atype_nv (anvtype_substitution X G G')
    | atype_void => atype_void
  end
with
 anvtype_substitution (X : (@generate_id type_var_id)) (G : fletch_anvtype)
                      (G' : fletch_anvtype)  : fletch_anvtype :=
  match G' with
    | atype_var Y =>
        if type_var_id_compare X Y then G
        else G'
    | atype_closure GL T => (atype_closure (map (anvtype_substitution X G) GL)
                                           (atype_substitution X G T))
    | atype_classtype (atype_class C GL) => 
      (atype_classtype (atype_class C (map (fun A => (anvtype_substitution X G A)) GL)))
    | atype_dynamic => atype_dynamic
  end.

Fixpoint aclasstype_substitution (X : (@generate_id type_var_id))
                                (G : fletch_anvtype)
                                (N : fletch_aclasstype) (*{struct N}*) : fletch_aclasstype :=
  match N with
    | (atype_class C GL) => (atype_class C (map (fun A => anvtype_substitution X G A) GL))
  end.

(* If PL = [(X1,G1),...,(Xn,Gn)] then mtype_substitution PL GL' T = [GL'/XL]T *)
Fixpoint mtype_substitution (PL : generics) (GL : list fletch_nvtype) (T : fletch_nvtype) : fletch_nvtype :=
  match (PL, GL) with
    | (nil, _) => T
    | (_, nil) => T
    | ((X,N) :: PL', G :: GL') =>
        mtype_substitution PL' GL' (nvtype_substitution X G T)
  end.

Fixpoint mctype_substitution (PL : generics)
                             (GL : list fletch_anvtype)
                             (N : fletch_aclasstype) : fletch_aclasstype :=
  match (PL, GL) with
    | (nil, _) => N
    | (_, nil) => N
    | ((X,N) :: PL', G :: GL') =>
        mctype_substitution PL' GL' (aclasstype_substitution X G N)
  end.

Fixpoint mtype_substitution_list (PL : generics) (GL : list fletch_nvtype) : list fletch_nvtype :=
  match PL with
    | nil => nil
    | (X,N) :: PL' => (mtype_substitution PL
                                          GL
                                          (type_classtype N)) ::
                                          (mtype_substitution_list PL'
                                                                   GL)
  end.

(* If PL = [(X1,G1),...,(Xn,Gn)] then amtype_substitution PL GL' T = [GL'/XL]T *)

Fixpoint amtype_substitution (PL : generics) (GL : list fletch_anvtype) (T : fletch_anvtype) : fletch_anvtype :=
  match (PL, GL) with
    | (nil, _) => T
    | (_, nil) => T
    | ((X,N) :: PL', G :: GL') =>
        amtype_substitution PL' GL' (anvtype_substitution X G T)
  end.

Fixpoint amttype_substitution (PL : generics)
                              (GL : list fletch_anvtype)
                              (T : fletch_atype) : fletch_atype :=
  match (PL, GL) with
    | (nil, _) => T
    | (_, nil) => T
    | ((X,N) :: PL', G :: GL') =>
        amttype_substitution PL' GL' (atype_substitution X G T)
  end.
  

(* Coertion from type parameters (G, x) to variable environment entry x : G. *)
Fixpoint params_to_venv (GxL : params) : fletch_venv := 
  match GxL with
      | [] => []
      | (G,x) :: xs => ((var_user x), (convert_nvtype G)) :: (params_to_venv xs)
  end.

(* 
 * Definition of the dynamic_sub predicate.
 * dynamic_sub_nv G = G[bottom/dynamic], for any non function type G.
 * We could propagate dynamic_sub into function types too, but it is would not
 * change our formalism.
 *)
Fixpoint dynamic_sub_nv (G : fletch_nvtype) : fletch_nvtype :=
  match G with
    | type_dynamic => type_bottom
    | type_classtype N => (type_classtype (dynamic_sub_class N))
    | G' => G'
  end
with
 dynamic_sub_class (N : fletch_classtype) : fletch_classtype :=
  match N with
    | C<<<GL>>> => type_class C (map dynamic_sub_nv GL)
  end.

Definition dynamic_sub (T : fletch_type) : fletch_type :=
  match T with
    | type_nv G => type_nv (dynamic_sub_nv G)                                                 
    | T' => T'
  end.

(* Definition of the specific typing relation *)
Reserved Notation "Delta '|-' T '<<' R" (at level 40, T at level 40).
Reserved Notation "Delta '|-^' TL '<<' RL" (at level 40, TL at level 40).
Inductive specific_typing : fletch_tenv -> fletch_type -> fletch_type -> Prop :=
  | st_reflexivity : forall Delta T, Delta |- T << T
  | st_transitivity : forall Delta T R R', Delta |- T << R' ->
                                           Delta |- R'<< R ->
                                           Delta |- T << R
  | st_bottom : forall Delta T, Delta |- (type_nv type_bottom) << T
  | st_dynamic : forall Delta T, Delta |-
                                 T << (type_nv type_dynamic)
  | st_class_covariance : forall Delta GL HL C,
                            Delta |-^ GL << HL ->
                            Delta |-
                (type_nv (type_classtype (C<<<GL>>>))) <<
                (type_nv (type_classtype (C<<<HL>>>)))
  | st_subclassing : forall Delta C PL FL ML GL N' HL IL D,
                (C, class C PL (atype_class D HL) FL ML) _in CT ->
                N' = (mtype_substitution PL GL (type_classtype (D<<<IL>>>))) ->
                (map dynamic_sub_nv (map convert_nvtype HL)) = IL ->
                        Delta |- (type_nv (type_classtype (C<<<GL>>>))) << type_nv N'
  | st_type_variable : forall X Delta Delta_X,
                         (X, Delta_X) _in Delta ->
                         Delta |-
                         (type_nv (type_var X))
                         <<
                         (type_nv (type_classtype (Delta_X)))
with
 specific_typing_list : fletch_tenv -> list fletch_nvtype -> list fletch_nvtype -> Prop :=
  | st_empty : forall Delta,
              Delta |-^ nil << nil
  | st_cons : forall Delta G H GL HL,
             Delta |- (type_nv G) << (type_nv H) ->
             Delta |-^ GL << HL ->
             Delta |-^ (G :: GL) << (H :: HL)
where "Delta '|-' T '<<' R" := (specific_typing Delta T R)
and "Delta '|-^' TL '<<' RL" := (specific_typing_list Delta TL RL).

(* Mutual induction principle for specific typing. *)
Scheme specific_typing_mut := Induction for specific_typing Sort Prop
with specific_typing_list_mut := Induction for specific_typing_list Sort Prop.

Hint Constructors specific_typing.

(* Definition of the Fletch subtyping relation. *)
Reserved Notation "Delta '|-' T '<:' R" (at level 40, T at level 40).
Reserved Notation "Delta '|-^' TL '<:' RL" (at level 40, TL at level 40).
Inductive subtyping : fletch_tenv -> fletch_type -> fletch_type -> Prop :=
  | s_spec : forall Delta T R,
               Delta |- dynamic_sub(T) << R ->
               Delta |- T <: R
  | s_fun : forall Delta T R GL HL,
              Delta |-^ (ltype_nv HL) <: (ltype_nv GL) ->
              ((Delta |- T <: R) \/ R = type_void) ->
              Delta |- (type_nv ((GL) --> T)) <:
                       (type_nv ((HL) --> R))
  | s_obj : forall Delta  GL T,
              Delta |-
                    type_nv ((GL) --> T) <:
                    (type_nv (type_classtype type_object))
with
 assignable : fletch_tenv -> fletch_type -> fletch_type -> Prop :=
  | a_subtyping : forall Delta T R,
                    Delta |- T <: R ->
                    assignable Delta T R
  | a_supertyping : forall Delta T R,
                      (~ exists GL S, R = type_nv ((GL) --> S)) \/
                      (~ exists GL S, T = type_nv ((GL) --> S)) ->
                      Delta |- R <: T ->
                      assignable Delta T R
with
 subtyping_list : fletch_tenv -> list fletch_type -> list fletch_type -> Prop :=
  | subtyping_empty : forall Delta,
              Delta |-^ nil <: nil
  | subtyping_cons : forall Delta G H GL HL,
             Delta |- G <: H ->
             Delta |-^ GL <: HL ->
             Delta |-^ (G :: GL) <: (H :: HL)
with
 assignable_list : fletch_tenv -> list fletch_type -> list fletch_type -> Prop :=
  | assignable_empty : forall Delta,
              assignable_list Delta nil nil
  | assignable_cons : forall Delta T S TL SL,
             assignable Delta T S ->
             assignable_list Delta TL SL ->
             assignable_list Delta (T :: TL) (S :: SL)
where "Delta '|-' T '<:' R" := (subtyping Delta T R)
and "Delta '|-^' TL '<:' RL" := (subtyping_list Delta TL RL).

Hint Constructors subtyping assignable subtyping_list assignable_list.

(* Mutual induction principle for subtyping. *)
Scheme subtyping_mut := Induction for subtyping Sort Prop
with subtyping_list_mut := Induction for subtyping_list Sort Prop
with assignable_mut := Induction for assignable Sort Prop
with assignable_list_mut := Induction for assignable_list Sort Prop.

Lemma subtyping_em : forall Delta T S,
                       Delta |- T <: S \/ ~ Delta |- T <: S.
Proof.
  intros. assert(EM := excluded_middle (Delta |- T <: S)). auto.
Qed.

(* Definition of well formed types. *)
Reserved Notation "Delta '|-' T 'OK'" (at level 40, T at level 40).
Reserved Notation "Delta '|-^' TL 'OK'" (at level 40, TL at level 40).
Inductive type_ok : fletch_tenv -> fletch_type -> Prop :=
  | dynamic_ok : forall Delta, Delta |- type_nv type_dynamic OK
  | bottom_ok : forall Delta, Delta |- type_nv type_bottom OK
  | void_ok : forall Delta, Delta |- type_void OK
  | variable_ok : forall X XL Delta,
                     (X, XL) _in Delta ->
                     Delta |- type_nv (type_var X) OK
  | class_ok : forall Delta C PL N FL ML GL,
                       (C, class C PL N FL ML) _in CT ->
                       Delta |-^ GL OK ->
                       Delta |-^  (map nvtype_type GL) <:
                             (map nvtype_type
                              (mtype_substitution_list
                                PL
                                GL)) ->
                       Delta |- type_nv (type_classtype (C<<<GL>>>)) OK
  | object_ok : forall Delta, Delta  |- 
                              type_nv (type_classtype (builtin_class_object<<<nil>>>)) OK
  | closure_ok : forall Delta GL T,
                   Delta |-^ GL OK ->
                   Delta |- T OK ->
                   Delta |- type_nv ((GL) --> T) OK
with
 type_ok_list : fletch_tenv -> list fletch_nvtype -> Prop :=
  | typeok_empty : forall Delta,
                     Delta |-^ nil OK
  | typeok_cons : forall Delta G GL,
                     Delta |- type_nv G OK ->
                     Delta |-^ GL OK ->
                     Delta |-^ (G :: GL) OK
where "Delta '|-' T 'OK'" := (type_ok Delta T)
and "Delta '|-^' TL 'OK'" := (type_ok_list Delta TL).

Hint Constructors type_ok type_ok_list.

(* Auxiliary predicates *)

(* nonsubtype (G_1 ... G_n) (H1 ... Hn) holds iff G_i is not a subtype of H_i for some i in (1...n)  *)
Inductive nonsubtype : list fletch_nvtype ->
                       list fletch_nvtype -> Prop :=
  | ns_base : forall G GL H HL,
                ~ Delta0 |- (type_nv G) <: (type_nv H) ->
                nonsubtype (G :: GL) (H :: HL)
  | ns_ind : forall G GL H HL,
               nonsubtype GL HL ->
               nonsubtype (G :: GL) (H :: HL).

Inductive location_typing : fletch_lenv ->
                               list (@generate_id location_id) ->
                               list fletch_nvtype -> Prop :=
  | lst_empty : forall sigma,
                  location_typing sigma [] []
  | lst_symmety : forall sigma L LL G GL o,
                    (L,o) __in sigma ->
                    typeof o = type_nv G ->
                    location_typing sigma LL GL ->
                    location_typing sigma (L :: LL) (G :: GL).
(* 
 * Field type lookup from class types.
 * ftype C<GL> f G holds IFF the type of f in the class C is G (after
 * binding the type arguments GL with the corresponding type parameters
 * accordingly)
 *)
Inductive ftype : fletch_type ->
                         (@generate_id field_id) ->
                         fletch_nvtype -> Prop :=
  | base_class_field : forall C PL N FL ML GL H f G,
                       (C, class C PL N FL ML) _in CT ->
                       (field G f) _in FL ->
                       H = mtype_substitution PL GL G ->
                       ftype (type_nv (type_classtype (C<<<GL>>>))) f H
  | super_class_field : forall C PL N FL ML GL f H N' H',
                       (C, class C PL N FL ML) _in CT ->
                       N' = mtype_substitution
                          PL
                          GL
                          (type_classtype (convert_classtype N)) ->
                       (forall H, (field H f) not _in FL) ->
                       ftype (type_nv N') f H ->
                       (field H' f) _in FL ->
                       ftype
                         (type_nv (type_classtype (C<<<GL>>>)))
                         f
                         H
  | dynamic_field_field : forall f, ftype
                                      (type_nv type_dynamic)
                                      f
                                      type_dynamic.

Hint Constructors ftype.

Fixpoint input_types (pl : params) : list fletch_nvtype :=
  match pl with
    | nil => nil
    | (G,x) :: pl' => (convert_nvtype G) :: (input_types pl')
  end.

(* 
 * Method type lookup from class types.
 * mtype C<GL> m G holds IFF the type of m in the class C is G (after
 * binding the type arguments GL with the corresponding type parameters
 * accordingly)
 *)
Inductive mtype : fletch_type ->
                         (@generate_id method_id) ->
                         fletch_nvtype -> Prop :=
  | base_class_method : forall C PL N FL ML GL m sm re T MT,
                       (C, class C PL N FL ML) _in CT ->
                       (method T m sm re) _in ML ->
                       MT = (mtype_substitution
                          PL
                          GL
                          ((input_types sm) --> convert_type T)) ->
                       mtype (type_nv (type_classtype (C<<<GL>>>)))
                                    m
                                    MT
  | super_class_method : forall C PL N FL ML GL m N' T,
                       (C, class C PL N FL ML) _in CT ->
                       N' = mtype_substitution
                           PL
                           GL
                           (type_classtype (convert_classtype N)) ->
                       mtype (type_nv N') m T ->
                       (forall T sm re, (method T m sm re) not _in ML) ->
                       mtype (type_nv (type_classtype (C<<<GL>>>))) m T
  | dynamic_method : forall m, mtype
                                 (type_nv type_dynamic)
                                 m
                                 type_dynamic.

Hint Constructors mtype.

(* Getter type lookup definition. *)
Inductive getter : fletch_type ->
                   property ->
                   fletch_nvtype -> Prop :=
  | get_field : forall G T f p,
                  ftype G f T ->
                  p = field_property f ->
                  getter G p ([] --> (type_nv T))
  | get_method : forall G m T p,
                   mtype G m T ->
                   p = method_property m ->
                   getter G p ([] --> (type_nv T)).
(* Setter type lookup definition. *)
Inductive setter : fletch_type ->
                   property ->
                   fletch_nvtype -> Prop :=
  | set_field : forall G H f p,
                  ftype G f H ->
                  p = field_property f ->
                  setter G p ([H] --> (type_nv H)).

Hint Constructors getter setter.

(* 
 * Given an environment Delta, bound_delta determines the type G' obtained
 * by a type G replacing all the type parameters in the domain of Delta, with
 * their repsective bound in Delta.
 *) 
Inductive bound_delta : fletch_tenv -> fletch_type -> fletch_type -> Prop :=
 | bound_void : forall Delta, bound_delta Delta type_void type_void
 | bound_nv : forall Delta G G',
                bound_delta_nv Delta G G' ->
                bound_delta Delta (type_nv G) (type_nv G')
with
 bound_delta_nv : fletch_tenv -> fletch_nvtype -> fletch_nvtype -> Prop :=
  | bound_bottom : forall Delta, bound_delta_nv Delta type_bottom type_bottom
  | bound_closure : forall Delta GL GL' T T',
                      bound_delta_nv_list Delta GL GL' ->
                      bound_delta Delta T T' ->
                      bound_delta_nv Delta ((GL) --> T) ((GL') --> T')
  | bound_variable_bind : forall Delta X T,
                            (X,T) _in Delta ->
                            bound_delta_nv Delta (type_var X) (type_classtype T)
  | bound_dynamic : forall Delta, bound_delta_nv Delta type_dynamic type_dynamic
  | bound_class : forall Delta C GL GL',
                    bound_delta_nv_list Delta GL GL' ->
                    bound_delta_nv Delta
                                   (type_classtype (C<<<GL>>>))
                                   (type_classtype (C<<<GL'>>>))
with
 bound_delta_nv_list : fletch_tenv -> list fletch_nvtype -> list fletch_nvtype -> Prop :=
  | boundnv_nil : forall Delta, bound_delta_nv_list Delta nil nil
  | boundnv_cons : forall Delta G GL G' GL' L1 L2,
                   bound_delta_nv Delta G G' ->
                   bound_delta_nv_list Delta GL GL' ->
                   L1 = G :: GL ->
                   L2 = G' :: GL' ->
                   bound_delta_nv_list Delta L1 L2
with
 bound_delta_list : fletch_tenv -> list fletch_type -> list fletch_type -> Prop :=
  | bound_nil : forall Delta, bound_delta_list Delta nil nil
  | bound_cons : forall Delta T TL T' TL' L1 L2,
                   bound_delta Delta T T' ->
                   bound_delta_list Delta TL TL' ->
                   L1 = T :: TL ->
                   L2 = T' :: TL' ->
                   bound_delta_list Delta L1 L2.

(* Mutual induction for bound_delta. *)
Scheme bound_delta_mut := Induction for bound_delta Sort Prop
with bound_delta_nv_mut := Induction for bound_delta_nv Sort Prop
with bound_delta_list_mut := Induction for bound_delta_list Sort Prop
with bound_delta_nv_list_mut := Induction for bound_delta_nv_list Sort Prop.

(* 
 * Definition the overriding predicate for fields, which specifies if a field can 
 * be overrided with a given type.
 *)
Inductive foverride : fletch_tenv ->
                      (@generate_id field_id) ->
                      fletch_classtype ->
                      fletch_type -> Prop :=
  | foverride_base : forall Delta f (N : fletch_aclasstype) (G : fletch_nvtype)
                            H (N' : fletch_aclasstype),
     bound_delta Delta (type_nv (type_classtype (convert_classtype N)))
                       (type_nv (type_classtype (convert_classtype N'))) ->
     (ftype (type_nv (convert_nvtype (atype_classtype N'))) f H) ->
     (* assignable Delta G (type_nv H) -- in original Fletch TS -- *)
     Delta |- (type_nv G) <: (type_nv H) ->
     foverride Delta f (convert_classtype N) (type_nv G).

(* 
 * Definition the overriding predicate for methods, which specifies if a method can 
 * be overrided with a given type.
 *)
Inductive moverride : fletch_tenv ->
                      (@generate_id method_id) ->
                      fletch_classtype ->
                      fletch_type -> Prop :=
  | moverride_base : forall Delta m (N : fletch_aclasstype) GL HL T R (N' : fletch_aclasstype),
     bound_delta Delta (type_nv (type_classtype (convert_classtype N)))
                       (type_nv (type_classtype (convert_classtype N'))) ->
     (mtype (type_nv (convert_nvtype (atype_classtype N'))) m (HL --> R)) ->
      (* assignable_list Delta (ltype_nv GL) (ltype_nv HL) -> *)
      (* Delta |- T <: R -> *)
     GL = HL ->
     T = R ->
     moverride Delta m (convert_classtype N) (type_nv (GL --> T)).

Fixpoint projection (GxL : params) : list fletch_nvtype :=
  (map
     (fun L => match L with | (G,x) => (convert_nvtype G) end)
     GxL).

Inductive arg_venv : fletch_venv -> param -> fletch_venv -> Prop :=
  | arg_venv_case : forall Gamma (G : fletch_anvtype) x Gamma',
                      insert_venv Gamma (var_user x) (convert_nvtype G) Gamma' ->
                      arg_venv Gamma (G,x) Gamma'.

Lemma arg_venv_well_formed :
  forall Gx (Gamma : fletch_venv) (Gamma' : fletch_venv),
         well_formed_env Gamma ->
         arg_venv Gamma Gx Gamma' ->
         well_formed_env Gamma'.
Proof.
  intros Gx Gamma Gamma' WFenv ArgVenv. destruct Gx.
  inversion ArgVenv.
  assert(wfLemma := (@insert_well_formed fletch_var fletch_nvtype)
                    (var_user g) f Gamma Gamma' WFenv H3).
  exact wfLemma.
Qed.

Reserved Notation "nu ';' sigma ';' Delta ';' Gamma '|-STD' e ':' T" (at level 39, sigma, Delta, Gamma,e,T at level 38).
Reserved Notation "nu ';' sigma ';' Delta ';' Gamma '|-STD^' el ':' TL" (at level 39, sigma, Delta, Gamma,el,TL at level 38).

(* MESSAGE SAFE TYPE SYSTEM *)

(* 
 * The getter and setter predicates have been collapsed into the accessor predicate in the
 * paper formal definition. Here we keep the getter and setter predicates for simplicity and more clarity in the Coq
 * proof, but they are equivalent to the accessor predicate defined in the paper.
 *)
Inductive exp_typing_standard : fletch_rvenv -> fletch_lenv -> fletch_tenv -> fletch_venv ->
                                fletch_exp -> fletch_type -> Prop :=
  | TS_exp_var : forall nu sigma Delta Gamma y G,
              (y, G) _in Gamma ->
              nu ; sigma ; Delta ; Gamma |-STD (exp_var y) : (type_nv G)
  | TS_exp_getter : forall nu sigma Delta Gamma e0 p G T T',
                     nu ; sigma ; Delta ; Gamma |-STD e0 : T ->
                     bound_delta Delta T T' ->
                     getter T' p (nil --> (type_nv G)) ->
              nu ; sigma ; Delta ; Gamma |-STD (exp_getter e0 p) : type_nv G
  | TS_exp_setter : forall nu sigma Delta Gamma e0 e1 p R T T' G,
                     nu ; sigma ; Delta ; Gamma |-STD e0 : T ->
                     bound_delta Delta T T' ->
                     setter T' p ((G :: nil) --> type_nv G) ->
                     nu ; sigma ; Delta ; Gamma |-STD e1 : R ->
                     assignable Delta R (type_nv G) ->
              nu ; sigma ; Delta ; Gamma |-STD (exp_setter e0 p e1) : R
  | TS_exp_assign : forall nu sigma Delta Gamma e0 T x R,
                      nu ; sigma ; Delta ; Gamma |-STD e0 : T ->
                      ((var_user (id x)), R) _in Gamma ->
                      assignable Delta T (type_nv R) ->
              nu ; sigma ; Delta ; Gamma |-STD (exp_assignment (id x) e0) : T
  | TS_exp_call : forall nu sigma Delta Gamma e0 T el GL RL,
                      nu ; sigma ; Delta ; Gamma |-STD e0 : type_nv ((GL) --> T) ->
                      nu ; sigma ; Delta ; Gamma |-STD^ el : RL ->
                      assignable_list Delta (ltype_nv GL) RL ->
                      nu ; sigma ; Delta ; Gamma |-STD (exp_call e0 el) : T
  | TS_exp_new : forall nu sigma Delta Gamma (N : fletch_aclasstype),
              Delta |- (type_nv (type_classtype (convert_classtype N))) OK ->
              nu ; sigma ; Delta ; Gamma |-STD (exp_new N) :
                                             (type_nv (type_classtype (convert_classtype N)))
  | TS_exp_function : forall nu sigma Delta GL Gamma Gamma'' GxL e0 T Gamma',
                     GL = projection GxL ->
                     Delta |-^ GL OK ->
                     Gamma' = params_to_venv GxL ->
                     fold_pred update_venv Gamma Gamma' Gamma'' ->
                     nu ; sigma ; Delta ; Gamma'' |-STD e0 : T ->
                     nu ; sigma ; Delta ; Gamma |-STD
                              (exp_function GxL T e0) : (type_nv ((GL) --> T))
  | TS_exp_location : forall nu sigma Delta Gamma l T o,
                         (l,o) __in sigma ->
                         typeof o = T ->
                         nu ; sigma ; Delta ; Gamma |-STD (exp_location l) : T
  | TS_exp_var_location : forall nu sigma Delta Gamma tau l G,
                           (tau, (G, l)) __in nu ->
                           nu ; sigma ; Delta ; Gamma |-STD (exp_var_location tau) : (type_nv G)
  | TS_exp_runtime_assignment : forall nu sigma Delta Gamma tau e T l G,
                   nu ; sigma ; Delta ; Gamma |-STD e : T ->
                   (tau, (G, l)) __in nu ->
                   assignable Delta (type_nv G) T ->
                   nu ; sigma ; Delta ; Gamma |-STD (exp_runtime_assignment tau e) : T
  | TS_exp_unfolded_call : forall nu sigma Delta Gamma T R e,
                            nu ; sigma ; Delta ; Gamma |-STD e : R ->
                            Delta |- R <: T ->
                            nu ; sigma ; Delta ; Gamma |-STD (exp_frame_expression T e) : T
with
 exp_typing_standard_list : fletch_rvenv -> fletch_lenv -> fletch_tenv -> fletch_venv ->
                   list fletch_exp -> list fletch_type -> Prop :=
  | TLS_nil : forall nu sigma Delta Gamma, nu ; sigma ; Delta ; Gamma |-STD^ nil : nil
  | TLS_cons : forall nu sigma Delta Gamma e T el TL,
               nu ; sigma ; Delta ; Gamma |-STD e : T ->
               nu ; sigma ; Delta ; Gamma |-STD^ el : TL ->
               nu ; sigma ; Delta ; Gamma |-STD^ (e :: el) : (T :: TL)
where "nu ';' sigma ';' Delta ';' Gamma '|-STD' e ':' T"  :=
  (exp_typing_standard nu sigma Delta Gamma e T)
and "nu ';' sigma ';' Delta ';' Gamma '|-STD^' el ':' TL" :=
  (exp_typing_standard_list nu sigma Delta Gamma el TL).

Scheme exp_typing_standard_mut := Induction for exp_typing_standard Sort Prop
with exp_typing_standard_list_mut := Induction for exp_typing_standard_list Sort Prop.

(* RELAXED TYPE SYSTEM *)

(* 
 * The following typing rules are obtained by the typing rules for the message safe type system, by removing the assignability and
 * subtyping requirements in the premises.
 * We will prove progress and preservation for the relaxed type system.
 *)
Reserved Notation "nu ';' sigma ';' Delta ';' Gamma '|-' e ':' T" (at level 39, sigma, Delta, Gamma,e,T at level 38).
Reserved Notation "nu ';' sigma ';' Delta ';' Gamma '|-^' el ':' TL" (at level 39, sigma, Delta, Gamma,el,TL at level 38).

Inductive exp_typing : fletch_rvenv -> fletch_lenv -> fletch_tenv -> fletch_venv ->
                       fletch_exp -> fletch_type -> Prop :=
  | T_exp_var : forall nu sigma Delta Gamma y G,
              (y, G) _in Gamma ->
              nu ; sigma ; Delta ; Gamma |- (exp_var y) : (type_nv G)
  | T_exp_getter : forall nu sigma Delta Gamma e0 p G T T',
                     nu ; sigma ; Delta ; Gamma |- e0 : T ->
                     bound_delta Delta T T' ->
                     getter T' p (nil --> (type_nv G)) ->
                 nu ; sigma ; Delta ; Gamma |- (exp_getter e0 p) : type_nv G
  | T_exp_setter : forall nu sigma Delta Gamma e0 e1 p R T T' G,
                     nu ; sigma ; Delta ; Gamma |- e0 : T ->
                     bound_delta Delta T T' ->
                     setter T' p ((G :: nil) --> type_nv G) ->
                     nu ; sigma ; Delta ; Gamma |- e1 : R -> (* No assignability check *)
                 nu ; sigma ; Delta ; Gamma |- (exp_setter e0 p e1) : R
  | T_exp_assign : forall nu sigma Delta Gamma e0 T tau R,
                      nu ; sigma ; Delta ; Gamma |- e0 : T ->
                      ((var_user tau), R) _in Gamma -> (* No assignability check *)
                 nu ; sigma ; Delta ; Gamma |- (exp_assignment tau e0) : T
  | T_exp_call : forall nu sigma Delta Gamma e0 T el GL RL,
                      nu ; sigma ; Delta ; Gamma |- e0 : type_nv ((GL) --> T) ->
                      nu ; sigma ; Delta ; Gamma |-^ el : RL -> (* No assignability check *)
                      nu ; sigma ; Delta ; Gamma |- (exp_call e0 el) : T
  | T_exp_new : forall nu sigma Delta Gamma (N : fletch_aclasstype),
              Delta |- (type_nv (type_classtype (convert_classtype N))) OK ->
              nu ; sigma ; Delta ; Gamma |- (exp_new N) :
                                            (type_nv (type_classtype (convert_classtype N)))
  | T_exp_function : forall nu sigma Delta GL Gamma Gamma'' GxL e0 T Gamma',
                     GL = projection GxL ->
                     Delta |-^ GL OK ->
                     Gamma' = params_to_venv GxL ->
                     fold_pred update_venv Gamma Gamma' Gamma'' ->
                     nu ; sigma ; Delta ; Gamma'' |- e0 : T ->
                     nu ; sigma ; Delta ; Gamma |-
                                          (exp_function GxL T e0) :
                                          (type_nv ((GL) --> T))
  | T_exp_location : forall nu sigma Delta Gamma l T o,
                         (l,o) __in sigma ->
                         typeof o = T ->
                         nu ; sigma ; Delta ; Gamma |- (exp_location l) : T
  | T_exp_var_location : forall nu sigma Delta Gamma tau l G,
                           (tau, (G, l)) __in nu ->
                           nu ; sigma ; Delta ; Gamma |- (exp_var_location tau) : (type_nv G)
  | T_exp_runtime_assignment : forall nu sigma Delta Gamma tau e T l G,
                   nu ; sigma ; Delta ; Gamma |- e : T ->
                   (tau, (G, l)) __in nu -> (* No assignability check *)
                   nu ; sigma ; Delta ; Gamma |- (exp_runtime_assignment tau e) : T
  | T_exp_unfolded_call : forall nu sigma Delta Gamma T e R,
                            nu ; sigma ; Delta ; Gamma |- e : R ->
                            (* Delta |- R <: T -> *) (* No subtyping check *)
                            nu ; sigma ; Delta ; Gamma |- (exp_frame_expression T e) : T
with
 exp_typing_list : fletch_rvenv -> fletch_lenv -> fletch_tenv -> fletch_venv ->
                   list fletch_exp -> list fletch_type -> Prop :=
  | TL_nil : forall nu sigma Delta Gamma, nu ; sigma ; Delta ; Gamma |-^ nil : nil
  | TL_cons : forall nu sigma Delta Gamma e T el TL,
               nu ; sigma ; Delta ; Gamma |- e : T ->
               nu ; sigma ; Delta ; Gamma |-^ el : TL ->
               nu ; sigma ; Delta ; Gamma |-^ (e :: el) : (T :: TL)
where "nu ';' sigma ';' Delta ';' Gamma '|-' e ':' T"  := (exp_typing nu sigma Delta Gamma e T)
and "nu ';' sigma ';' Delta ';' Gamma '|-^' el ':' TL" := (exp_typing_list nu sigma Delta Gamma el TL).

Scheme exp_typing_mut := Induction for exp_typing Sort Prop
with exp_typing_list_mut := Induction for exp_typing_list Sort Prop.

Hint Constructors exp_typing exp_typing_list.

Inductive add_generic_toenv : fletch_tenv -> generic -> fletch_tenv -> Prop :=
  | add_generic_toenv_case : forall Delta X (N : fletch_aclasstype) Delta',
                               insert_tenv Delta X N Delta' ->
                               add_generic_toenv Delta (X,N) Delta'.

(* Typing rule for field declarations. *)
Reserved Notation "'field_dec' F 'OK' '_in' C" (at level 40, F at level 40).
Inductive field_declaration_typing : fletch_field ->
                                     (@generate_id class_id) -> Prop :=
  | field_ok_base : forall Delta (PL : generics) (G : fletch_anvtype) C N FL ML f,
                      fold_pred (fun Delta X N Delta' => add_generic_toenv Delta (X,N) Delta')
                                Delta0
                                PL
                                Delta ->
                      Delta |- (type_nv (convert_nvtype G)) OK ->
                      (C, class C PL N FL ML) _in CT ->
                      foverride Delta f (convert_classtype N) (type_nv (convert_nvtype G)) ->
                      field_dec (field G f) OK _in C
where "'field_dec' F 'OK' '_in' C" := (field_declaration_typing F C).

Fixpoint extract_vars (PL : generics) : list fletch_nvtype :=
  match PL with
    | nil => nil
    | (X,N) :: PL' => (type_var X) :: (extract_vars PL')
  end.

Fixpoint extract_types (PL : generics) : list fletch_nvtype :=
  match PL with
    | nil => nil
    | (X,N) :: PL' => (type_classtype (convert_classtype N)) :: (extract_types PL')
  end.

(* Typing rule for method declarations. *)
Reserved Notation "'method_dec' M 'OK' '_in' C" (at level 40, M,C at level 40).
Inductive method_declaration_typing : fletch_method ->
                                      (@generate_id class_id) -> Prop :=
| method_ok_base : forall C e0 GxL m T Delta PL Gamma Gamma' Gamma'' GL R N FL ML,
                     fold_pred (fun Delta X N Delta' => add_generic_toenv Delta (X, N) Delta')
                                Delta0
                                PL
                                Delta ->
                     Delta |- convert_type T OK ->
                     Delta |-^ GL OK ->
                     fold_pred
                      (fun (Gamma : fletch_venv)
                      (G:fletch_anvtype)
                      (x:@generate_id var_id)
                      (Gamma' : fletch_venv) =>
                      arg_venv Gamma (G,x) Gamma') Gamma0 GxL Gamma ->
                      insert_venv
                         Gamma'
                         var_this
                         (type_classtype (C<<<(extract_vars PL)>>>))
                         Gamma'' ->
                      nu0 ; sigma0 ; Delta ; Gamma'' |- e0 : R ->
                      (C, class C PL N FL ML) _in CT ->
                      assignable Delta R (convert_type T) ->
                      moverride Delta
                                m
                                (convert_classtype N)
                                (type_nv ((GL) --> (convert_type T))) ->
                      method_dec (method T m GxL e0) OK _in C
where "'method_dec' M 'OK' '_in' C" := (method_declaration_typing M C).

(* Minor definitions regarding duplicate field and method declarations. *)

Inductive NoDupFields : list fletch_field -> Prop :=
  | NoDupKeys_nil : NoDupFields []
  | NoDupKeys_cons : forall f FL T,
                       (forall T, (field T f) not _in FL) ->
                       NoDupFields FL ->
                       NoDupFields ((field T f) :: FL).

Inductive NoDupMethods : list fletch_method -> Prop :=
  | NoDupMethods_nil : NoDupMethods []
  | NoDupMethods_cons : forall m ML T GxL e,
                          (forall T GxL e, (method T m GxL e) not _in ML) ->
                          NoDupMethods ML ->
                          NoDupMethods ((method T m GxL e) :: ML).

(* Typing rule for class declarations. *)

Reserved Notation "'class_dec' C 'OK'" (at level 200).
Inductive class_declaration_typing : fletch_class -> Prop :=
  | class_ok_base : forall PL C (N : fletch_aclasstype) FL ML D PL1 Delta,
                       fold_pred (fun Delta X N Delta' => add_generic_toenv Delta (X,N) Delta')
                                 Delta0
                                 PL
                                 Delta /\
                       Delta |-^ (extract_types PL) OK ->
                       Delta |- type_nv (type_classtype (convert_classtype N)) OK ->
                       (N = (atype_class D PL1) -> C <> D) ->
                       (forall F, F _in FL -> field_dec F OK _in C) ->
                       (forall M, M _in ML -> method_dec M OK _in C) ->
                       NoDupFields FL /\
                       NoDupMethods ML ->
                       class_dec (class C PL N FL ML) OK
where "'class_dec' C 'OK'" := (class_declaration_typing C).

Hint Constructors field_declaration_typing method_declaration_typing class_declaration_typing.

Definition function_type (T : fletch_type) :=
  exists GL T', T = (type_nv ((GL) --> T')).

Lemma ftype_em : forall T,
                   function_type T \/ ~ function_type T.
Proof.
  intros. assert(EM := excluded_middle (function_type T)). auto.
Qed.

End StandardTS.
