(* ------------------------------------------  Fletch type syntax ----------------------------------------- *)

Require Export fletch_commons.
Require Export List.

Module FletchTypes.
Export Commons.

(* 
 * Type syntax for type annotations in fletch programs.
 * Instances of this type are only used to provide type annotations to
 * program components such as fields and parameters.
 *)

Inductive fletch_atype : Set :=
  | atype_nv : fletch_anvtype -> fletch_atype
  | atype_void : fletch_atype
with
 fletch_anvtype : Set :=
  | atype_var : @generate_id type_var_id -> fletch_anvtype
  | atype_classtype : fletch_aclasstype -> fletch_anvtype
  | atype_closure : list fletch_anvtype -> fletch_atype -> fletch_anvtype
  | atype_dynamic : fletch_anvtype
with
 fletch_aclasstype : Set :=
  | atype_class : @generate_id class_id -> list fletch_anvtype -> fletch_aclasstype.

(* Notations for types. E.g. "atype_object" represents the type "Object". *) 
Notation atype_object := (atype_class builtin_class_object nil).
Definition anv_atype_class (C : @generate_id class_id) (GL : list fletch_anvtype) := (atype_classtype (atype_class C GL)).
Definition a_atype_class (C : @generate_id class_id) (GL : list fletch_anvtype) :=
  atype_nv (anv_atype_class C GL).
Definition a_atype_var (x : @generate_id type_var_id) := atype_nv (atype_var x).
Definition a_atype_dynamic := atype_nv atype_dynamic.

(* 
 * Type syntax for the type system.
 * Instances of this type represents type assigned to expressions. The type system
 * uses this type sysntax to define typing judgments.
 * fletch_type is a proper superset of fletch_atype, because it contains type_bottom, i.e., the bottom type. 
 * It is not possible to annotated program components with type bottom, but the type system will assign, for example,
 * the type bottom to the null expression.
 *)
Inductive fletch_type : Set :=
  | type_nv : fletch_nvtype -> fletch_type
  | type_void : fletch_type
with
 fletch_nvtype : Set :=
  | type_var : @generate_id type_var_id -> fletch_nvtype
  | type_classtype : fletch_classtype -> fletch_nvtype
  | type_closure : list fletch_nvtype -> fletch_type -> fletch_nvtype
  | type_dynamic : fletch_nvtype
  | type_bottom : fletch_nvtype
with
 fletch_classtype : Set :=
  | type_class : @generate_id class_id -> list fletch_nvtype -> fletch_classtype.

(* Mutual induction principle for types. This induction principle be automatically generated by Coq. *)
Lemma fletch_type_mut :
  forall (P : fletch_nvtype -> Prop) (P0 : fletch_type -> Prop)
         (P1 : fletch_classtype -> Prop),
       (forall X : generate_id, P (type_var X)) ->
       (forall N : fletch_classtype, P1 N -> P (type_classtype N)) ->
       (forall (GL : list fletch_nvtype) (T : fletch_type),
        P0 T -> (forall G, G _in GL -> P G) -> P (type_closure GL T)) ->
       P type_dynamic ->
       P type_bottom ->
       (forall G : fletch_nvtype, P G -> P0 (type_nv G)) ->
       P0 type_void ->
       (forall (C : generate_id) (GL : list fletch_nvtype),
        (forall G, G _in GL -> P G) -> P1 (type_class C GL)) ->
       forall T : fletch_type, P0 T.
Proof.
  intros;revert T.
  refine (fix IHT (T : fletch_type) : P0 T := _
          with IHG (G : fletch_nvtype) : P G := _
          with IHN (N : fletch_classtype) : P1 N := _
          for IHT).
  destruct T.
    apply H4, IHG.
    apply H5.
  destruct G.
    apply H.
    apply H0, IHN.
    apply H1. apply IHT.
        induction l.
          contradiction.
          intros GL [HEq | HIn].
          rewrite <- HEq. apply IHG.
          apply IHl, HIn.
    apply H2.
    apply H3.
  destruct N.
    apply H6.
         induction l.
          contradiction.
          intros GL [HEq | HIn].
          rewrite <- HEq. apply IHG.
          apply IHl, HIn.
Qed.

Fixpoint convert_type (T : fletch_atype) : fletch_type :=
  match T with
    | atype_nv G => type_nv (convert_nvtype G)
    | atype_void => type_void
  end
with
 convert_nvtype (G : fletch_anvtype) : fletch_nvtype :=
  match G with
    | atype_var X => type_var X
    | atype_classtype N => type_classtype (convert_classtype N)
    | atype_closure CL T => type_closure (map convert_nvtype CL) (convert_type T)
    | atype_dynamic => type_dynamic
  end
with
 convert_classtype (N : fletch_aclasstype) : fletch_classtype :=
  match N with
    | atype_class C GL => type_class C (map convert_nvtype GL)
  end.

(* Coertions to convert annotated types into types generated by the type system *)
Coercion convert_type : fletch_atype >-> fletch_type.
Coercion convert_nvtype : fletch_anvtype >-> fletch_nvtype.
Coercion convert_classtype : fletch_aclasstype >-> fletch_classtype.

Definition nvtype_type (G : fletch_nvtype) : fletch_type := type_nv G.

(* Notations for types. *)
Notation type_object := (atype_class builtin_class_object nil).
Notation "C '<<<' GL '>>>'" := (type_class C GL) (at level 37).
Notation "GL '-->' T" := (type_closure GL T) (at level 37).

End FletchTypes.
