Require Export fletch_semantics.

(* ------------------------------  Fletch environment well formedness definitions ----------------------------------- *)

Module WellFormedness (C : ClassTable).
Module StandardCheckModeOSImpl := StandardCheckModeOS(C).
Export StandardCheckModeOSImpl.

(* Well formed environments *)

(* 
 * well_formed_envs nu sigma Gamma hold IFF nu, sigma, and Gamma do not contain duplicate
 * keys.
 *)
Definition well_formed_envs (nu:fletch_rvenv)
                            (sigma:fletch_lenv)
                            (Gamma:fletch_venv) : Prop :=
  well_formed_rvenv nu /\
  well_formed_lenv sigma /\
  well_formed_env Gamma.

(* 
 * Store and variable environment well formedness. Intuitively, it enforces that
 * any variable location of type G must point to an heap location whose type is a subtype of G.
 *
 * Let sigma be a store environment and nu a variable environment,
 * sigma |- nu IFF for each (tau, (G, l)) in nu (where tau is a variable
 * location, l a location, and G a type), the type of l is a subtype of G.
 *)
Definition well_formed_substenv_lenv (sigma:fletch_lenv) (nu:fletch_rvenv) :=
  forall tau G l,
    (tau,(G,l)) __in nu ->
    (exists o,
       (l,o) __in sigma /\
       Delta0 |- typeof o <: type_nv G).

Notation "sigma '|--S' nu" := (well_formed_substenv_lenv sigma nu)
                               (at level 45).

(* 
 * Field well formendess. Similarly to the definition above, a mapping (field, location)
 * is well formed IFF the type of "location" is a subtype of the type of "field".
 *)
Definition well_formed_fields (sigma : fletch_lenv)
                              (RFL : list runtime_field)
                              (GL : list fletch_anvtype)
                              (C : @generate_id class_id) :=
    (exists FL,
       get_fields (atype_class C GL) FL /\
       (forall f G G', (field G f) _in FL ->
          exists l o,
          (f, ((convert_nvtype G'), l)) _in RFL /\
          (l, o) __in sigma /\
          G = G' /\
          Delta0 |- typeof o <: (type_nv G)) /\
          NoDupKeys RFL).

(* Similar to the definition above but for methods. *)
Definition well_formed_methods (sigma : fletch_lenv)
                               (RML : list runtime_method)
                               (GL : list fletch_anvtype)
                               (C : @generate_id class_id) :=
    (exists ML,
       get_methods (atype_class C GL) ML /\
       (forall m GxL T e, (method T m GxL e) _in ML ->
       exists l o,
         (m, l) _in RML /\
         (l, o) __in sigma /\
         typeof o = type_nv ((runtime_types GxL) --> T))) /\
       NoDupKeys RML.

(* 
 * An object o is well formed if
 * - o is a class instance, and all its runtime field and method values are well formed
 * - o is a well typed closure
 * - o is the null object
 *)
Inductive well_formed_object : fletch_rvenv -> fletch_lenv -> fletch_runtime_object -> Prop :=
  | class_instance_wf : 
      forall C (GL : list fletch_anvtype) RFL RML nu sigma,
          well_formed_fields sigma RFL GL C ->
          well_formed_methods sigma RML GL C ->
        well_formed_object nu sigma (class_instance_object
                                    (atype_class C GL)
                                     RFL RML)
  | closure_object_wf :
      forall GxL S T e nu sigma Gamma Gamma',
            Gamma = params_to_venv GxL ->
            fold_pred update_venv Gamma0 Gamma Gamma' ->
            nu ; sigma ; Delta0 ; Gamma' |- e : S ->
        well_formed_object nu sigma (closure_object GxL T e)
  | empty_location_wf : forall nu sigma, well_formed_object nu sigma empty_location.

(* A store is well formed IFF all its objects are well formed. *)
Definition well_formed_store (nu : fletch_rvenv) (sigma : fletch_lenv) :=
  forall l o,
    (l, o) __in sigma ->
    well_formed_object nu sigma o.

Hint Constructors well_formed_object.

(* 
 * A variable environment nu is fully annotated IFF for each (tau, (G, l)) in nu,
 * the type G is fully annotated.
 *)
Definition fully_annotated_subst (nu : fletch_rvenv) :=
  forall tau l G,
    (tau,(G,l)) __in nu ->
    fully_annotated_type (type_nv G).

(* 
 * Similarly to the definition above, an object is fully annotated if all its
 * declared types are fully annotated.
 *)
Inductive fully_annotated_object : fletch_runtime_object -> Prop :=
  | fa_class_object : forall N RF RM G l f,
                        fully_annotated_type (atype_nv (atype_classtype N)) ->
                        (f,(G, l)) _in RF ->
                        fully_annotated_type (type_nv G) ->
                        fully_annotated_object (class_instance_object N RF RM)
  | fa_closure_object : forall (GxL : params) (T:fletch_atype) e,
                          fully_annotated_type T ->
                          (forall G x, (G,x) _in GxL ->
                                       fully_annotated_type (type_nv G)) ->
                          fully_annotated_expression e ->
                          fully_annotated_object (closure_object GxL T e)
  | fa_empty_location : fully_annotated_object empty_location.

(* A store is fully annotated if all its object are fully annotated. *)
Definition fully_annotated_store (store : fletch_lenv) : Prop:=
  forall l o,
    (l,o) __in store ->
    fully_annotated_object o.

(* 
 * nosubtype (L_1 ... L_n) (L_1' ... L_n') holds IFF the type of L_i 
 * is not a subtype of the type of L_i', for some i in (1..n).
 *)
Record nonsubtype : list fletch_type -> list fletch_type -> Prop :=
  | ns_ind : forall G H GL HL,
             (~Delta0 |- G <: H
             \/ nonsubtype GL HL) ->
             nonsubtype (G::GL) (H::HL).

(* 
 * location_typing (L_1 ... L_n) (G_1 ... G_n) holds IFF the type of L_i is G_i,
 * forall i in (1..n)
 *)
Inductive location_typing : fletch_lenv ->
                               list (@generate_id location_id) ->
                               list fletch_nvtype -> Prop  :=
  | lst_empty : forall sigma,
                  location_typing sigma [] []
  | lst_symmety : forall sigma L LL G GL o,
                    (L,o) __in sigma ->
                    typeof o = type_nv G ->
                    location_typing sigma LL GL ->
                    location_typing sigma (L :: LL) (G :: GL).

Hint Constructors location_typing fully_annotated_expression_list.

End WellFormedness.